#include "api/df_fingerprint.h"
#include "datafox_library_iv.h"

DF_API_EXTERN_DECORATE int DF_API_CALL DFCFingerprintAppendRecord( int connectionNumber, int deviceAddress, int templateType, const unsigned char *buffer, int bufferSize, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t%i", templateType );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0
        || templateType < 0 || templateType > 3
        || (templateType == 0 && bufferSize > 161) 
        || (templateType == 1 && bufferSize > 216)
        || (templateType == 2 && bufferSize > 561)
        || (templateType == 3 && bufferSize != (11 + 624))
        || buffer == nullptr || bufferSize < 20 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t0x%08X\t%i\t%i\t(%i, %i)", buffer, bufferSize, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fingertemplate pruefen (Gleichen sich in den ersten 11 Byte)
    StructFingertemplateDIN *fp = (StructFingertemplateDIN *)buffer;
    if ( fp->pidLo != 0 || fp->fid > 10 || (templateType != 3 && (fp->quality == 0 || fp->count == 0)) )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t0x%08X\t%i\t%i\t(%i, %i)", buffer, bufferSize, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );
        channelBridge->logEnd( "\t0x%08X\t%i\t%i\t(%i, %i)", buffer, bufferSize, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    // Fingertemplate und size ausgeben.
    // Daten fuer Log in Puffer uebergeben
    unsigned char ucTmp[(11 + 624) * 3 + 1];
    memset( ucTmp, 0, sizeof(ucTmp) );
    if ( channelBridge->logEnabled() )
    {
        int i = 0;
        for ( ; i < bufferSize && i < static_cast<int>(sizeof( ucTmp ) - 3); ++i )
        {
            sprintf( (char*)&ucTmp[i * 3], "%02X ", (unsigned char)buffer[i] );
        }
        ucTmp[i * 3] = 0;
    }
    else
    {
        sprintf( (char*)ucTmp, "0x%p", buffer );
    }

    channelBridge->logContinue( "\t%s\t%i", ucTmp, bufferSize );

    int iRes = channelBridge->fingerprintAppendRecord( l_device_address, templateType, buffer, bufferSize );
    if ( iRes == IV_ERROR_SUCCESS )
    {
        // Kein Fehler
        iRes = 1;
    }
    else if ( iRes > IV_ERROR_SUCCESS )
    {
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( IV_RESULT_MODULE( iRes ) )
        {
            
            if ( IV_RESULT_MODULE( iRes ) == 2 )
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_UNAVAILABLE, errorNumber );
            }
            else
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        }
        else
        {
            // Funktionsfehler auswerten
            if ( IV_RESULT_FUNCTION( iRes ) ==  129 )
            {
                // Das Fingerprintmodul bringt einen nicht näher spezifizierten Fehler.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FINGERPRINT, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  132 )
            {
                // Die Datenbank des Fingerprintmoduls ist voll.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FINGERPRINT_MEMORY_FULL, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  133 )
            {
                // Das Fingerprintmodul enthält keine Fingertemplates.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FINGERPRINT_TEMPLATES_NONE, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  134 )
            {
                // Das Fingertemplate existiert bereits.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FINGERPRINT_TEMPLATE_EXISTS, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  135 )
            {
                // Das angeforderte Fingertemplate existiert nicht.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FINGERPRINT_TEMPLATE_NOT_EXISTS, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) == 28 )
            {
                // Passwort wird benoetigt?
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_NEED_PASSWORD, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) == 29 )
            {
                // Nach 3 maliger Passwortfalscheingabe ist Geraet fuer 60 Sekunden gesperrt.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PASSWORD_LOCK, errorNumber );
            }
            else
            {
                // Funktionsfehler : Fehler in Verarbeitung.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        }

        iRes = 0;
    } 
    else
    {
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes, errorNumber );
        iRes = 0;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->resetWithError(channelBridge->getLastErrorNumber( l_device_address ));
    }

    channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCFingerprintGetRecord( int connectionNumber, int deviceAddress, int templateType, int pid, int fid, unsigned char *buffer, int *bufferSize, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t%i\t%i\t%i", templateType, pid, fid );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0
        || templateType < 0 || templateType > 3
        || pid == 0 || fid < 1  || fid > 10 || buffer == nullptr || bufferSize == nullptr || *bufferSize < 20
        || (templateType == 0 && *bufferSize < 161) 
        || (templateType == 1 && *bufferSize < 216)
        || (templateType == 2 && *bufferSize < 561)
        || (templateType == 3 && *bufferSize < (11 + 624)) )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t0x%08X\t%i\t%i\t(%i, %i)", buffer, (bufferSize == nullptr) ? 0 : *bufferSize, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );
        channelBridge->logEnd( "\t0x%08X\t%i\t%i\t(%i, %i)", buffer, *bufferSize, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    uint16_t l_buffer_size = (*bufferSize > 0xFFFF) ? 0xFFFF : static_cast<uint16_t>( *bufferSize );
    int iRes = channelBridge->fingerprintGetRecord( l_device_address, (unsigned char&)templateType, pid, fid, buffer, l_buffer_size );
    if ( iRes == IV_ERROR_SUCCESS )
    {
        // Kein Fehler
        iRes = 1;

        *bufferSize = l_buffer_size;
    }
    else if ( iRes > IV_ERROR_SUCCESS )
    {
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( IV_RESULT_MODULE( iRes ) )
        {
            if ( IV_RESULT_MODULE( iRes ) == 2 )
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_UNAVAILABLE, errorNumber );
            }
            else
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        }
        else
        {
            // Funktionsfehler auswerten
            if ( IV_RESULT_FUNCTION( iRes ) ==  129 )
            {
                // Das Fingerprintmodul bringt einen nicht näher spezifizierten Fehler.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FINGERPRINT, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  132 )
            {
                // Die Datenbank des Fingerprintmoduls ist voll.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FINGERPRINT_MEMORY_FULL, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  133 )
            {
                // Das Fingerprintmodul enthält keine Fingertemplates.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FINGERPRINT_TEMPLATES_NONE, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  134 )
            {
                // Das Fingertemplate existiert bereits.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FINGERPRINT_TEMPLATE_EXISTS, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  135 )
            {
                // Das angeforderte Fingertemplate existiert nicht.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FINGERPRINT_TEMPLATE_NOT_EXISTS, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) == 28 )
            {
                // Passwort wird benoetigt?
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_NEED_PASSWORD, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) == 29 )
            {
                // Nach 3 maliger Passwortfalscheingabe ist Geraet fuer 60 Sekunden gesperrt.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PASSWORD_LOCK, errorNumber );
            }
            else
            {
                // Funktionsfehler : Fehler in Verarbeitung.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        }

        iRes = 0;
    }
    else
    {
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes, errorNumber );
        iRes = 0;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    // Fingertemplate und size ausgeben.
    // Daten fuer Log in Puffer uebergeben
    unsigned char ucTmp[(11 + 624) * 3 + 1];
    memset( ucTmp, 0, sizeof( ucTmp ) );
    if ( iRes == 1 && channelBridge->logEnabled() )
    {
        int i = 0;
        for ( ; i < *bufferSize && i < static_cast<int>(sizeof( ucTmp ) - 3); ++i )
        {
            sprintf((char*)&ucTmp[i * 3], "%02X ", (unsigned char)buffer[i]);
        }
        ucTmp[i * 3] = 0;
    }
    else
    {
        sprintf( (char*)ucTmp, "0x%p", buffer );
    }

    channelBridge->logContinue( "\t%s\t%i", ucTmp, *bufferSize );

    channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCFingerprintDeleteRecord( int connectionNumber, int deviceAddress, int pid, int fid, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t%i\t%i", pid, fid );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 || (pid == 0 && fid != 0) )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );
        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    int iRes = channelBridge->fingerprintDeleteRecord( l_device_address, pid, fid );
    if ( iRes == IV_ERROR_SUCCESS )
    {
        // Kein Fehler
        iRes = 1;
    }
    else if ( iRes > IV_ERROR_SUCCESS )
    {
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( IV_RESULT_MODULE( iRes ) )
        {
            if ( IV_RESULT_MODULE( iRes ) == 2 )
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_UNAVAILABLE, errorNumber );
            }
            else
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        }
        else
        {
            // Funktionsfehler auswerten
            if ( IV_RESULT_FUNCTION( iRes ) ==  129 )
            {
                // Das Fingerprintmodul bringt einen nicht näher spezifizierten Fehler.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FINGERPRINT, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  132 )
            {
                // Die Datenbank des Fingerprintmoduls ist voll.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FINGERPRINT_MEMORY_FULL, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  133 )
            {
                // Das Fingerprintmodul enthält keine Fingertemplates.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FINGERPRINT_TEMPLATES_NONE, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  134 )
            {
                // Das Fingertemplate existiert bereits.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FINGERPRINT_TEMPLATE_EXISTS, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  135 )
            {
                // Das angeforderte Fingertemplate existiert nicht.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FINGERPRINT_TEMPLATE_NOT_EXISTS, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) == 28 )
            {
                // Passwort wird benoetigt?
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_NEED_PASSWORD, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) == 29 )
            {
                // Nach 3 maliger Passwortfalscheingabe ist Geraet fuer 60 Sekunden gesperrt.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PASSWORD_LOCK, errorNumber );
            }
            else
            {
                // Funktionsfehler : Fehler in Verarbeitung.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        }

        iRes = 0;
    }
    else
    {
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes, errorNumber );
        iRes = 0;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCFingerprintList( int connectionNumber, int deviceAddress, int flags, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t0x%08X", flags );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 || (flags != 0 && flags != 0x01) )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );
        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    unsigned int recordCount;
    int iRes = channelBridge->fingerprintList( l_device_address, flags, recordCount );
    if ( iRes == IV_ERROR_SUCCESS )
    {
        // Kein Fehler
        iRes = 1;
        // Anschliessende Ermittlung der Datensatzanzahl erwuenscht?
        if ( flags & 0x01 )
        {
            iRes = recordCount;
        }
    }
    else if ( iRes > IV_ERROR_SUCCESS )
    {
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( IV_RESULT_MODULE( iRes ) )
        {
            if ( IV_RESULT_MODULE( iRes ) == 2 )
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_UNAVAILABLE, errorNumber );
            }
            else
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        }
        else
        {
            // Funktionsfehler auswerten
            if ( IV_RESULT_FUNCTION( iRes ) ==  129 )
            {
                // Das Fingerprintmodul bringt einen nicht näher spezifizierten Fehler.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FINGERPRINT, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  132 )
            {
                // Die Datenbank des Fingerprintmoduls ist voll.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FINGERPRINT_MEMORY_FULL, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  133 )
            {
                // Das Fingerprintmodul enthält keine Fingertemplates.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FINGERPRINT_TEMPLATES_NONE, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  134 )
            {
                // Das Fingertemplate existiert bereits.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FINGERPRINT_TEMPLATE_EXISTS, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  135 )
            {
                // Das angeforderte Fingertemplate existiert nicht.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FINGERPRINT_TEMPLATE_NOT_EXISTS, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) == 28 )
            {
                // Passwort wird benoetigt?
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_NEED_PASSWORD, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) == 29 )
            {
                // Nach 3 maliger Passwortfalscheingabe ist Geraet fuer 60 Sekunden gesperrt.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PASSWORD_LOCK, errorNumber );
            }
            else
            {
                // Funktionsfehler : Fehler in Verarbeitung.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        }

        iRes = 0;
    }
    else
    {
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
         channelBridge->setLastErrorNumber( l_device_address, -iRes, errorNumber );
        iRes = 0;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCFingerprintBackup( int connectionNumber, int deviceAddress, const char *filePath, int flags, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }


    channelBridge->logBegin( __func__, deviceAddress, "" );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 || filePath == 0 || flags < 0 || flags > 1 ) 
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t%s\t0x%08X\t%i\t(%i, %i)", (filePath == nullptr) ? "" : filePath, flags, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );
        channelBridge->logEnd( "\t%s\t0x%08X\t%i\t(%i, %i)", filePath, flags, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    int iRes = channelBridge->fingerprintBackup( l_device_address, filePath, flags );
    if ( iRes == IV_ERROR_SUCCESS )
    {
        // Kein Fehler
        iRes = 1;
    }
    else if ( iRes > IV_ERROR_SUCCESS )
    {
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( IV_RESULT_MODULE( iRes ) )
        {
            if ( IV_RESULT_MODULE( iRes ) == 2 )
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_UNAVAILABLE, errorNumber );
            }
            else
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        }
        else
        {
            // Funktionsfehler auswerten
            if ( IV_RESULT_FUNCTION( iRes ) ==  129 )
            {
                // Das Fingerprintmodul bringt einen nicht näher spezifizierten Fehler.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FINGERPRINT, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  132 )
            {
                // Die Datenbank des Fingerprintmoduls ist voll.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FINGERPRINT_MEMORY_FULL, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  133 )
            {
                // Das Fingerprintmodul enthält keine Fingertemplates.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FINGERPRINT_TEMPLATES_NONE, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  134 )
            {
                // Das Fingertemplate existiert bereits.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FINGERPRINT_TEMPLATE_EXISTS, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  135 )
            {
                // Das angeforderte Fingertemplate existiert nicht.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FINGERPRINT_TEMPLATE_NOT_EXISTS, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) == 28 )
            {
                // Passwort wird benoetigt?
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_NEED_PASSWORD, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) == 29 )
            {
                // Nach 3 maliger Passwortfalscheingabe ist Geraet fuer 60 Sekunden gesperrt.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PASSWORD_LOCK, errorNumber );
            }
            else
            {
                // Funktionsfehler : Fehler in Verarbeitung.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        }

        iRes = 0;
    }
    else
    {
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes, errorNumber );
        iRes = 0;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%s\t0x%08X\t%i\t(%i, %i)", filePath, flags, (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCFingerprintRestore( int connectionNumber, int deviceAddress, const char *filePath, int flags, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "" );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 || filePath == 0 || flags < 0 || flags > 1 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t%s\t0x%08X\t%i\t(%i, %i)", (filePath == nullptr) ? "" : filePath, flags, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );
        channelBridge->logEnd( "\t%s\t0x%08X\t%i\t(%i, %i)", filePath, flags, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    int iRes = channelBridge->fingerprintRestore( l_device_address, filePath, flags );
    if ( iRes == IV_ERROR_SUCCESS )
    {
        // Kein Fehler
        iRes = 1;
    }
    else if ( iRes > IV_ERROR_SUCCESS )
    {
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( IV_RESULT_MODULE( iRes ) )
        {
            if ( IV_RESULT_MODULE( iRes ) == 2 )
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_UNAVAILABLE, errorNumber );
            }
            else
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        }
        else
        {
            // Funktionsfehler auswerten
            if ( IV_RESULT_FUNCTION( iRes ) ==  129 )
            {
                // Das Fingerprintmodul bringt einen nicht näher spezifizierten Fehler.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FINGERPRINT, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  132 )
            {
                // Die Datenbank des Fingerprintmoduls ist voll.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FINGERPRINT_MEMORY_FULL, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  133 )
            {
                // Das Fingerprintmodul enthält keine Fingertemplates.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FINGERPRINT_TEMPLATES_NONE, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  134 )
            {
                // Das Fingertemplate existiert bereits.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FINGERPRINT_TEMPLATE_EXISTS, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) ==  135 )
            {
                // Das angeforderte Fingertemplate existiert nicht.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FINGERPRINT_TEMPLATE_NOT_EXISTS, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) == 28 )
            {
                // Passwort wird benoetigt?
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_NEED_PASSWORD, errorNumber );
            }
            else if ( IV_RESULT_FUNCTION( iRes ) == 29 )
            {
                // Nach 3 maliger Passwortfalscheingabe ist Geraet fuer 60 Sekunden gesperrt.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PASSWORD_LOCK, errorNumber );
            }
            else
            {
                // Funktionsfehler : Fehler in Verarbeitung.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        }

        iRes = 0;
    }
    else
    {
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes, errorNumber );
        iRes = 0;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%s\t0x%08X\t%i\t(%i, %i)", filePath, flags, (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}
