/****************************************************************************
**
** Copyright (C) Datafox GmbH.
** All rights reserved.
** Contact: Datafox GmbH (develop@datafox.de)
**
** This file is part of the DFComDLL module of the Datafox SDK.
**
****************************************************************************/

#include "datafox/communication/connection_base.h"

// Format-Lib
#include "fmt/core.h"
#include "fmt/format.h"

#if !defined(_MSC_VER) && !defined(__MINGW32__)
#include <sys/ioctl.h>
#include <sys/time.h>
#endif

#if defined(_MSC_VER) && !defined(snprintf)
#define snprintf _snprintf
#endif

#include <string>
#include <cstring>
#include <cstdio>
#include <cstdarg>
#include <ctime>
#include <memory>
#include <thread>
#include <sstream>

Datafox::Communication::TraceBase::TraceBase( int32_t number )
{
    m_number = number;
    // Tracemeldungen
    m_mode  = TraceMode::off;
    // Tracefilename
    m_filePath = fmt::format( "DFComDLL_{}.log", m_number );

#ifdef _DEBUG
    // Nur im Debugmodus standardmaessig aktiv.
    setMode();
#endif
}

// ##################################################################
/**
* @brief Hauptlogfunktion der Verbindungsklassen.
*/
void Datafox::Communication::TraceBase::print( const char *format, va_list args )
{
    if ( mode() >= TraceMode::on )
    {
        if (/*std::strstr(format, "INFO(1)") != nullptr ||
            std::strstr(format, "INFO(2)") != nullptr ||*/
            std::strstr(format, "INFO(3)") != nullptr || /*
            std::strstr(format, "WARNING(1)") != nullptr ||
            std::strstr(format, "WARNING(2)") != nullptr ||*/
            std::strstr(format, "WARNING(3)") != nullptr /*||
            std::strstr(format, "ERROR(1)") != nullptr ||
            std::strstr(format, "ERROR(2)") != nullptr ||
            std::strstr(format, "ERROR(3)") != nullptr*/ )
        {
            return;
        }

        va_list args2;
        va_copy( args2, args );
        std::string buffer( vsnprintf( nullptr, 0, format, args ) + 1, '\0' );
        vsnprintf( &buffer[0], buffer.size(), format, args2 );
        va_end( args2 );

        // Eine ggf. mitberechnete Nullterminierung entfernen.
        if ( buffer[buffer.size() - 1] == '\0' )
        {
            buffer.resize( buffer.size() - 1 );
        }

        std::size_t pos;
        if ( (pos = buffer.find( ":INFO(" )) != std::string::npos || 
            (pos = buffer.find( ":WARNING(" )) != std::string::npos || 
            (pos = buffer.find( ":ERROR(" )) != std::string::npos )
        {
            // Uhrzeit des Logeintrages
            time_t t = time( nullptr );
            tm *ltm = localtime( &t );
            unsigned short msec = 0;
#if defined(_MSC_VER) || defined(__MINGW32__)
            SYSTEMTIME st;
            GetLocalTime( &st );
            msec = st.wMilliseconds;
#endif
            // Uhrzeit einfuegen
            std::string temp( 100, '\0' );
            snprintf( &temp[0], temp.size(), ":%02i:%02i:%02i.%03i", ltm->tm_hour, ltm->tm_min, ltm->tm_sec, msec );
            buffer.insert( pos, temp.c_str() );

            // ThreadID einfuegen
            std::stringstream ss_ids;
            ss_ids << ":(" << number() << ", " << std::this_thread::get_id() << ")";
            buffer.insert( pos, ss_ids.str().c_str() );
        }

#if !defined(_MSC_VER) && !defined(__MINGW32__)
        if ( buffer[strlen(buffer.c_str()) - 1] == '\n' )
            buffer.append( "\r\n" );
#endif

        FILE* fp = fopen( m_filePath.data(), "a" );
        if ( fp == nullptr )
            return;

        fwrite( buffer.data(), 1, buffer.size(), fp );
        fclose( fp );

#ifdef _DEBUG
#if defined(_MSC_VER) || defined(__MINGW32__)
        OutputDebugStringA( buffer.data() );
#endif
#endif
    }
}

// ##################################################################
/**
* @brief Initialisieren der Membervariablen.
*
* @param connection_id Nummer der Verbindungsklasse.
*/
Datafox::Communication::ConnectionBase::ConnectionBase( int32_t number ) 
     : m_trace( std::make_shared<TraceBase>( number ) )
{
    // Verbindungsnummer zuweisen
    m_number = number;

    // Membervars initialisieren
    m_type    = ConnectionType::undefined;
    m_serialPort  = INVALID_HANDLE_VALUE;
    m_socket  = INVALID_SOCKET;
    m_value   = 0;
    m_readTimeout = 0;
#if !defined(_MSC_VER) && !defined(__MINGW32__)
    std::fill_n( (uint8_t *)&m_flock, sizeof( struct flock ), 0 );
    std::fill_n( (uint8_t *)&m_termio, sizeof( struct termios ), 0 );
#endif

    // Statistik
    m_sizeWritten = 0;
    m_sizeRead = 0;
}

// ##################################################################
/**
* @brief Destruktor schliesst die eventuell noch geoeffnete Schnittstelle. 
*
*/
Datafox::Communication::ConnectionBase::~ConnectionBase() 
{
    // Schliessen der Verbindung, wenn nicht bereits geschehen.
    close();
}


//////////////////////////////////////////////////////////////////////
// Members
//////////////////////////////////////////////////////////////////////

// ##################################################################
/**
* @brief Verwirft anstehende Zeichen der Schnittstelle.
*/
void Datafox::Communication::ConnectionBase::flush()
{
    trace( "%s(%d):INFO(3): %s\n", __FILE__, __LINE__, __func__ );

    uint8_t buffer[0x100];
    uint32_t readCount = 0xFF;

    // eventuell anstehende Datenpakete verwerfen
    for ( int32_t i = 0; (i < 5) && readCount; ++i )
    {
        if ( readData( buffer, sizeof( buffer ) - 1, readCount ) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%d):INFO(2): %s() [ Abbruch nach %d durchläufen. ]\n", __FILE__, __LINE__, __func__, i );
#ifdef _DEBUG
            for ( uint32_t z = 0; z < readCount; ++z ) 
            {
                trace( "%02X ", buffer[z] );
            }

            trace( "\n" );
#endif
            break;
        }

        trace( "%s(%d):INFO(2): %s() [ Verworfene Anzahl Zeichen (%u Byte) ]\n", __FILE__, __LINE__, __func__, readCount );
#ifdef _DEBUG
        for ( uint32_t z = 0; z < readCount; ++z ) 
        {
            trace( "%02X ", buffer[z] );
        }

        trace( "\n" );
#endif
    }
}

// ##################################################################
/**
* @brief Oeffnen der Kommunikationsschnittstelle.
*/
int32_t Datafox::Communication::ConnectionBase::open( ConnectionType connection_type, const std::string &connectionString, uint32_t connectionValue, uint16_t communicationReadTimeout )
{
    trace( "%s(%d):INFO(3): %s( %d, %s, %u, %u )\n", __FILE__, __LINE__, __func__, connection_type, connectionString.c_str(), connectionValue, communicationReadTimeout );

    int32_t res = IV_ERROR_FAIL;

    switch ( connection_type )
    {
    case ConnectionType::serial:

        if ( communicationReadTimeout == 0 )
        {
            // Von einem Standardtimeout 800ms ausgehen.
            if ( m_readTimeout == 0 )
            {
                communicationReadTimeout = 800;
            }
            else
            {
                communicationReadTimeout = m_readTimeout; // Zuvor gewaehlte.
            }

        }
        else if ( communicationReadTimeout < 500 )
        { 
            // Mindesttimeout sicherstellen.
            trace( "%s(%d):WARNING(1): %s() [ Zu kleiner Timeout (%u), wurde auf 500 gesetzt. ]\n", __FILE__, __LINE__, __func__, communicationReadTimeout );
            communicationReadTimeout = 500;
        }
        else if ( communicationReadTimeout > 60000 )
        {
            // Maximal 1 Minute als Timeout zulassen.
            trace( "%s(%d):WARNING(1): %s() [ Zu großer Timeout (%u), wurde auf 60000 gesetzt. ]\n", __FILE__, __LINE__, __func__, communicationReadTimeout );
            communicationReadTimeout = 60000;
        }

        if ( connectionValue == 0 )
        {
            // Von einer Standardbaudrate 38400bd ausgehen.
            if ( m_value == 0 )
            {
                connectionValue = 38400;
            }
            else 
            {
                connectionValue = m_value; // Zuvor gewaehlte.
            }
        }

        if ( openSerial( connectionString, connectionValue, communicationReadTimeout ) == IV_ERROR_SUCCESS )
        {
            m_type = ConnectionType::serial;
            m_value = connectionValue;
            m_readTimeout = communicationReadTimeout;
            m_string = connectionString;

            res = IV_ERROR_SUCCESS;
        }
        break;

    case ConnectionType::socket:
        // Mindesttimeout sicherstellen.
        if ( communicationReadTimeout < 2000 )
        {
            trace( "%s(%d):WARNING(1): %s() [ Zu kleiner Timeout (%u), wurde auf 2000 gesetzt. ]\n", __FILE__, __LINE__, __func__, communicationReadTimeout );
            communicationReadTimeout = 2000;
        }
        else if ( communicationReadTimeout > 60000 )
        {
            trace( "%s(%d):WARNING(1): %s() [ Zu großer Timeout (%u), wurde auf 60000 gesetzt. ]\n", __FILE__, __LINE__, __func__, communicationReadTimeout );
            communicationReadTimeout = 60000;
        }

        if ( openSocket( connectionString, (uint16_t)connectionValue, communicationReadTimeout ) == IV_ERROR_SUCCESS )
        {
            m_type = ConnectionType::socket;
            m_value = connectionValue;
            m_readTimeout = communicationReadTimeout;
            m_string = connectionString;

            res = IV_ERROR_SUCCESS;
        }
        break;

    default:
        trace( "%s(%d):ERROR(1): %s() [ Unbekannter Type (%d) ]\n", __FILE__, __LINE__, __func__, connection_type );
        return res;
    }

    return res;
}

// ##################################################################
/**
* @brief Ermitteln ob die Kommunikationsschnittstelle geoeffnet ist.
*/
bool Datafox::Communication::ConnectionBase::isOpen()
{
    trace( "%s(%d):INFO(3): %s()\n", __FILE__, __LINE__, __func__ );

    // Bei Socketverbindung sicherstellen, dass diese aktuell noch besteht.
    if ( type() == ConnectionType::socket )
    {
#if defined(_MSC_VER)
#if !defined(MYCONST)
#define MYCONST const
#endif
#else
#define MYCONST
#endif
        struct timeval seltime;
        fd_set readfds;
        char buffer = 0;

        int32_t res;

        FD_ZERO( (fd_set FAR *) &readfds );
        FD_SET( m_socket, (fd_set FAR *) &readfds );
        seltime.tv_sec  = 0;
        seltime.tv_usec = 500;

        if ( (res = select( FD_SETSIZE,
            (fd_set FAR *)&readfds, nullptr, nullptr,
            (MYCONST struct timeval FAR *)&seltime )) < 0 )
        {
            trace( "%s(%d):ERROR(1): %s() [ id: %d, error: %d ]\n", __FILE__, __LINE__, __func__, number(), WSAGetLastError() );
            close();
        }
        else if ( res ) // Liegen Daten an?
        {
            if ( (res = recv( m_socket, (char*)&buffer, 1, MSG_PEEK )) == SOCKET_ERROR )
            {
                trace( "%s(%d):ERROR(1): %s() [ id: %d, error: %d ]\n", __FILE__, __LINE__, __func__, number(), WSAGetLastError() );
                close();
            }
            else if ( res == 0 )
            {
                // Wenn Daten gemeldet werden,
                // die Anzahl aber 0 ist, so wurde die Verbindung beendet.

                // Verbindung wurde von Gegenseite geschlossen.
                trace( "%s(%d):ERROR(1): %s() [ id: %d, error: %d ]\n", __FILE__, __LINE__, __func__, number(), WSAGetLastError() );
                close();
            }
        }
        /*
        int32_t res;
        char buffer = '?';
        // Daten senden
        if ( (res = send( _socket, "?", 1, MSG_OOB )) == SOCKET_ERROR )
        {
        trace( "%s(%d):ERROR(1): %s() [ id: %d, error: %d ]\n", __FILE__, __LINE__, __func__, getId(), WSAGetLastError() );
        close();
        }

        // Wurden alle Daten versendet?
        if ( res == 0 )
        {
        trace( "%s(%d):WARNING(1): %s() [ Es wurden keine Daten gesendet. id: %d ]\n", __FILE__, __LINE__, __func__, getId() );
        close();
        }
        */
    }

    return (type() != ConnectionType::undefined);
}

// ##################################################################
/**
* @brief Schliessen der Kommunikationsschnittstelle.
*/
int32_t Datafox::Communication::ConnectionBase::close()
{
    trace( "%s(%d):INFO(3): %s()\n", __FILE__, __LINE__, __func__ );

    switch ( type() )
    {
    case ConnectionType::undefined:
        return IV_ERROR_SUCCESS;

    case ConnectionType::socket:
        closeSocket();
        break;

    case ConnectionType::serial:
        closeSerial();
        break;

    default:
        trace( "%s(%d):ERROR(1): %s() [ Unbekannter Type (%d) ]\n", __FILE__, __LINE__, __func__, type() );
        return IV_ERROR_FAIL;
    }

    m_type = ConnectionType::undefined;
    m_string = "";

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
* @brief Zuweisen einer vorhandenen Kommunikationsschnittstelle.
* 
*/
int32_t Datafox::Communication::ConnectionBase::attach( ConnectionType connectionType, HANDLE connectionHandle, uint16_t communicationReadTimeout )
{
    trace("%s(%d):INFO(3): %s( %d, %p, %u )\n", __FILE__, __LINE__, __func__, connectionType, connectionHandle, communicationReadTimeout);

    bool was_undefined = (type() == ConnectionType::undefined);

    if ( type() == ConnectionType::undefined
        || (type() == connectionType && connectionType == ConnectionType::serial && connectionHandle == m_serialPort)
        || (type() == connectionType && connectionType == ConnectionType::socket && connectionHandle == (HANDLE)m_socket) )
    {
        switch ( connectionType )
        {
        case ConnectionType::serial:
            m_type = ConnectionType::serial;
            m_serialPort = (HANDLE)connectionHandle;
            m_readTimeout   = communicationReadTimeout;
            break;

        case ConnectionType::socket:
            if ( communicationReadTimeout < 2000 )
            {
                trace( "%s(%d):WARNING(1): %s() [ Zu kleiner Timeout (%u), wurde auf 2000 gesetzt. ]\n", __FILE__, __LINE__, __func__, communicationReadTimeout );
                communicationReadTimeout = 2000;
            }
            else if ( communicationReadTimeout > 60000 )
            {
                trace( "%s(%d):WARNING(1): %s() [ Zu großer Timeout (%u), wurde auf 60000 gesetzt. ]\n", __FILE__, __LINE__, __func__, communicationReadTimeout );
                communicationReadTimeout = 60000;
            }

            m_type   = ConnectionType::socket;
            m_socket = (SOCKET)connectionHandle;
            m_readTimeout     = communicationReadTimeout;
            break;

        default:
            trace( "%s(%d):ERROR(1): %s() [ Unbekannter Type (%d) ]\n", __FILE__, __LINE__, __func__, connectionType );
            return IV_ERROR_FAIL;
        }

        if ( was_undefined ) // Kanal Zugewiesen
            m_string = "ATTACH";

        return IV_ERROR_SUCCESS;
    }

    trace( "%s(%d):ERROR(1): %s() [ Zuweisen des Kanals nicht moeglich, da nicht Undefiniert. ]\n", __FILE__, __LINE__, __func__ );

    return IV_ERROR_FAIL;

}

// ##################################################################
/**
* @brief Freigeben der aktuellen Kommunikationsschnittstelle.
*/
int32_t Datafox::Communication::ConnectionBase::detach()
{
    trace( "%s(%d):INFO(3): %s()\n", __FILE__, __LINE__, __func__ );

    if ( type() == ConnectionType::serial || m_string == "ATTACH" )
    {
        m_type = ConnectionType::undefined;
        m_value = 0;
        m_readTimeout = 0;
        m_serialPort = INVALID_HANDLE_VALUE;
        m_socket = INVALID_SOCKET;

        // Kommunikationsstring loeschen
        m_string = "";

        return IV_ERROR_SUCCESS;
    }

    trace( "%s(%d):ERROR(1): %s() [ Freigabe des Handle nicht moeglich, da kein serial-port. ]\n", __FILE__, __LINE__, __func__ );

    return IV_ERROR_FAIL;
}

// ##################################################################
/**
* @brief Oeffnen einer seriellen Kommunikationsschnittstelle.
*/
int32_t Datafox::Communication::ConnectionBase::openSerial( const std::string &serialPortName, uint32_t baudRate, uint16_t communicationReadTimeout )
{
    trace( "%s(%d):INFO(3): %s( %s, %u, %u )\n", __FILE__, __LINE__, __func__, serialPortName.c_str(), baudRate, communicationReadTimeout );

    int32_t rc;

    if ( type() != ConnectionType::undefined )
    {
        // Verbindungs-Id bereits geoeffnet
        trace( "%s(%d):ERROR(1): %s() [ Verbindungs-Id bereits geoeffnet. ]\n", __FILE__, __LINE__, __func__ );
        return IV_ERROR_FAIL;
    }

#if defined(_MSC_VER) || defined(__MINGW32__)

    if ( (m_serialPort = CreateFileA(
        serialPortName.data(),        // pointer to name of the file 
        (GENERIC_READ | GENERIC_WRITE), // access (read-write) mode 
        0,                              // share mode 
        nullptr,                        // pointer to security descriptor
        OPEN_EXISTING,                  // how to create 
        0,                              // file attributes 
        nullptr                         // handle to file with attributes to copy
        )) == INVALID_HANDLE_VALUE )
    { 
        trace( "%s(%d):ERROR(1): %s() [ CreateFile fail (%u) ]\n", __FILE__, __LINE__, __func__, GetLastError() );
        return IV_ERROR_FAIL;
    }
#else
    errno = 0;

    if ( (rc = ::open( serialPortName.data(), O_RDWR | O_NDELAY )) < 0 )
    {
        trace( "%s(%d):ERROR(1): %s() [ open fail ]\n", __FILE__, __LINE__, __func__ );
        return IV_ERROR_FAIL;
    }

    m_serialPort = rc;

    if ( (rc = fcntl( m_serialPort, F_SETFL, O_RDWR )) < 0 )
    {
        trace( "%s(%d):ERROR(1): %s() [ fcntl fail ]\n", __FILE__, __LINE__, __func__ );
        return IV_ERROR_FAIL;
    }

    m_flock.l_type    = F_WRLCK;  // lock stream for read and write
    m_flock.l_whence  = 0;        // lock whole file -> 0,0,0
    m_flock.l_start   = 0;
    m_flock.l_len     = 0;

    if ( (rc = fcntl( m_serialPort, F_SETLK, &m_flock )) < 0 )
    {
        trace( "%s(%d):WARINING(1): %s() [ fcntl (file locking) fail ]\n", __FILE__, __LINE__, __func__ );
        //return IV_ERROR_FAIL;
    }
#endif

    if ( (rc = changeSerialState( m_serialPort, baudRate, communicationReadTimeout )) != IV_ERROR_SUCCESS )
    {
        return IV_ERROR_FAIL;
    }

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
* @brief Aendern der aktuellen Parameter zur seriellen Kommunikationsschnittstelle.
*/
int32_t Datafox::Communication::ConnectionBase::changeSerialState( HANDLE connectionSerialPort, uint32_t baudRate, uint16_t communicationReadTimeout )
{
    trace( "%s(%d):INFO(3): %s( %p, %u, %u )\n", __FILE__, __LINE__, __func__, connectionSerialPort, baudRate, communicationReadTimeout);

#if defined(_MSC_VER) || defined(__MINGW32__)
    DCB          dcb;
    COMMTIMEOUTS ctout;
    DWORD        cerr = 0;

    memset( &dcb, 0, sizeof( dcb ) );
    dcb.DCBlength = sizeof( dcb );
    dcb.fBinary   = 1;

    dcb.BaudRate  = baudRate;
    dcb.ByteSize  = 8;
    dcb.Parity    = NOPARITY;
    dcb.StopBits  = ONESTOPBIT;

    if ( !SetCommState( connectionSerialPort, &dcb ) )
    {
        trace( "%s(%d):ERROR(1): %s() [ SetCommState fail (%u) ]\n", __FILE__, __LINE__, __func__, GetLastError() );
        closeSerial();
        return IV_ERROR_FAIL;
    }

    // Handelt es sich um die eigene Schnittstelle?
    if ( connectionSerialPort == m_serialPort )
    {
        // Akutuelle Baudrate anpassen.
        m_value = baudRate;
    }

    ctout.ReadIntervalTimeout         = 0;
    ctout.ReadTotalTimeoutMultiplier  = 0;
    ctout.ReadTotalTimeoutConstant    = communicationReadTimeout;
    ctout.WriteTotalTimeoutConstant   = 0;
    ctout.WriteTotalTimeoutMultiplier = 0;

    if ( !SetCommTimeouts( connectionSerialPort, &ctout ) )
    {
        trace( "%s(%d):ERROR(1): %s() [ SetCommTimeouts fail ]\n", __FILE__, __LINE__, __func__ );
        closeSerial();
        return IV_ERROR_FAIL;
    }

    // Handelt es sich um die eigene Schnittstelle?
    if ( connectionSerialPort == m_serialPort )
    {
        // Aktuellen Timeout anpassen
        m_readTimeout = communicationReadTimeout;
    }

    // unsolved: fuer modems + RS485 notwendig !?!
    if ( !EscapeCommFunction( connectionSerialPort, SETRTS ) )
    {
        trace( "%s(%d):ERROR(1): %s() [ EscapeCommFunction fail ]\n", __FILE__, __LINE__, __func__ );
        closeSerial();
        return IV_ERROR_FAIL;
    }
    if ( !EscapeCommFunction( connectionSerialPort, SETDTR ) )
    {
        trace( "%s(%d):ERROR(1): %s() [ EscapeCommFunction fail ]\n", __FILE__, __LINE__, __func__ );
        closeSerial();
        return IV_ERROR_FAIL;
    }

    if ( !ClearCommError( connectionSerialPort, &cerr, nullptr ) )
    {
        trace( "%s(%d):ERROR(1): %s() [ ClearCommError fail ]\n", __FILE__, __LINE__, __func__ );
        closeSerial();
        return IV_ERROR_FAIL;
    }

#else
    int32_t rc;
    int32_t baud_mode, char_mode, pari_check, stop_mode;
    int32_t termio_cm;

    switch( baudRate )
    {
    case 9600:
        baud_mode = B9600;
        break;

    case 19200:
        baud_mode = B19200;
        break;

    case 38400:
        baud_mode = B38400;
        break;

    case 57600:
        baud_mode = B57600;
        break;
    case 115200:

        baud_mode = B115200;
        break;

    default:
        return IV_ERROR_FAIL;
    }

    char_mode = CS8;
    pari_check = 0;
    stop_mode = 0;

    m_termio.c_iflag = 0;
    m_termio.c_oflag = 0;
    m_termio.c_cflag = 0;
    m_termio.c_lflag = 0;
    m_termio.c_line  = 0;
    memset( (void *)m_termio.c_cc, 0, NCCS );

    m_termio.c_cflag |= HUPCL;        // hangup line when close/exit(DTR of)
    m_termio.c_cflag |= CLOCAL;       // local/direct connection

    m_termio.c_cflag |= baud_mode;     // set baudrate
    m_termio.c_cflag |= char_mode;     // set databits
    m_termio.c_cflag |= pari_check;    // set parity
    m_termio.c_cflag |= stop_mode;     // set stopbits

    m_termio.c_cflag |= CREAD;        // enable input queue

    m_termio.c_cc[VMIN]  = 1;         // read waiting for 1 char

    if ( (rc = ioctl( connectionSerialPort, TCSETA, &m_termio )) < 0 )
    {
        trace( "%s(%d):ERROR(1): %s() [ ioctl fail ]\n", __FILE__, __LINE__, __func__ );
        closeSerial();
        return IV_ERROR_FAIL;
    }

    if ( (rc = ioctl( connectionSerialPort, TCFLSH, TCIOFLUSH )) < 0 )
    {
        trace( "%s(%d):ERROR(1): %s() [ ioctl fail ]\n", __FILE__, __LINE__, __func__ );
        closeSerial();
        return IV_ERROR_FAIL;
    }

    if ( (rc = ioctl( connectionSerialPort, TIOCMGET, &termio_cm )) < 0 )
    {
        trace( "%s(%d):ERROR(1): %s() [ ioctl fail ]\n", __FILE__, __LINE__, __func__ );
        closeSerial();
        return IV_ERROR_FAIL;
    }

    if ( !(termio_cm & TIOCM_RTS) )
    {
        termio_cm |= TIOCM_RTS;
    }

    if ( !(termio_cm & TIOCM_DTR) )
    {
        termio_cm |= TIOCM_DTR;
    }

    if ( (rc = ioctl( connectionSerialPort, TIOCMSET, &termio_cm )) < 0 )
    {
        trace( "%s(%d):ERROR(1): %s() [ ioctl fail ]\n", __FILE__, __LINE__, __func__ );
        closeSerial();
        return IV_ERROR_FAIL;
    }

    // Handelt es sich um die eigene Schnittstelle?
    if ( connectionSerialPort == m_serialPort )
    {
        // Akutuelle Baudrate und Timeout anpassen.
        m_value = baudRate;
        m_readTimeout = communicationReadTimeout;
    }
#endif

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
* @brief Schliessen der seriellen Kommunikationsschnittstelle.
*/
int32_t Datafox::Communication::ConnectionBase::closeSerial()
{
    int32_t res = IV_ERROR_SUCCESS;
    trace( "%s(%d):INFO(3): %s()\n", __FILE__, __LINE__, __func__ );

#if defined(_MSC_VER) || defined(__MINGW32__)
    try
    {
        if ( !CancelIo( m_serialPort ) )
        {
            trace( "%s(%d):ERROR(1): %s() [ CancelIo fail (%u) ]\n", __FILE__, __LINE__, __func__, GetLastError() );
        }
    }
    catch ( ... )
    {
    }

    if ( (res == IV_ERROR_SUCCESS)  && !EscapeCommFunction( m_serialPort, CLRRTS ) )
    {
        trace( "%s(%d):ERROR(1): %s() [ EscapeCommFunction fail (%u) ]\n", __FILE__, __LINE__, __func__, GetLastError() );
        res = IV_ERROR_FAIL;
    }
    if ( (res == IV_ERROR_SUCCESS)  && !EscapeCommFunction( m_serialPort, CLRDTR ) )
    {
        trace( "%s(%d):ERROR(1): %s() [ EscapeCommFunction fail (%u) ]\n", __FILE__, __LINE__, __func__, GetLastError() );
        res = IV_ERROR_FAIL;
    }

    try
    {
        if ( !CloseHandle( m_serialPort ) )
        {
            trace( "%s(%d):ERROR(1): %s() [ CloseHandle fail (%u) ]\n", __FILE__, __LINE__, __func__, GetLastError() );
        }
    }
    catch ( ... )
    {
    }
#else
    int32_t rc;
    int32_t termio_cm;

    if ( (res == IV_ERROR_SUCCESS) && (rc = ioctl( m_serialPort, TIOCMGET, &termio_cm )) < 0 )
    {
        trace( "%s(%d):ERROR(1): %s() [ ioctl fail ]\n", __FILE__, __LINE__, __func__ );
        res = IV_ERROR_FAIL;
    }

    if ( (termio_cm & TIOCM_RTS) )
    {
        termio_cm &= ~TIOCM_RTS;
    }

    if ( (termio_cm & TIOCM_DTR) )
    {
        termio_cm &= ~TIOCM_DTR;
    }

    if ( (res == IV_ERROR_SUCCESS) && (rc = ioctl( m_serialPort, TIOCMSET, &termio_cm )) < 0 )
    {
        trace( "%s(%d):ERROR(1): %s() [ ioctl fail ]\n", __FILE__, __LINE__, __func__ );
        res = IV_ERROR_FAIL;
    }

    ::close( m_serialPort );
#endif
    m_serialPort = INVALID_HANDLE_VALUE;

    return res;
}

// ##################################################################
/**
* @brief Lesen aus der seriellen Kommunikationsschnittstelle.
*/
int32_t Datafox::Communication::ConnectionBase::readSerial( uint8_t *buffer, uint32_t readCount, uint32_t &readInCount, uint8_t readTrials )
{
    trace("%s(%d):INFO(3): %s( %p, %u, %u, %u, %u )\n", __FILE__, __LINE__, __func__, buffer, readCount, readInCount, readTrials);

#if defined(_MSC_VER) || defined(__MINGW32__)
    uint32_t res;
#else
    int32_t res{};
#endif
    uint8_t trials = 0;
    readInCount = 0;
#if !defined(_MSC_VER) && !defined(__MINGW32__)
    struct timeval seltime;
    fd_set readfds;
    errno = 0;
#endif

    // Daten emfpangen
    for ( ; (readCount != readInCount) && (trials < readTrials); ++trials )
    {
        res = 0;
#if defined(_MSC_VER) || defined(__MINGW32__)
        if ( !ReadFile( m_serialPort, buffer + readInCount, DWORD(readCount - readInCount), (LPDWORD)&res, nullptr ) )
        {
            if ( GetLastError() == ERROR_FILE_NOT_FOUND )
            {
                trace( "%s(%d):ERROR(1): %s() [ Schnittstelle wird geschlossen, error: %u ]\n", __FILE__, __LINE__, __func__, GetLastError() );
                close();
            }

            trace( "%s(%d):ERROR(1): %s() [ trial: %u, readCount: %u, readInCount: %u ]\n", __FILE__, __LINE__, __func__, trials, readCount, readInCount );
            return -IV_ERROR_RECEIVE;
        }
#else
        FD_ZERO( (fd_set FAR *) &readfds );
        FD_SET( m_serialPort, (fd_set FAR *) &readfds );
        seltime.tv_sec  = m_readTimeout / 1000;
        seltime.tv_usec = (m_readTimeout % 1000) * 1000;

        if ( (res = select( FD_SETSIZE,
            (fd_set FAR *)&readfds, (fd_set FAR *) 0, (fd_set FAR *) 0,
            (struct timeval *)&seltime )) < 0 )
        {
            trace( "%s(%d):ERROR(1): %s() [ trial: %u, readCount: %u, readInCount: %u ]\n", __FILE__, __LINE__, __func__, trials, readCount, readInCount );
            return -IV_ERROR_RECEIVE;
        }

        // Liegen Daten an?
        if ( res )
        { 
            if ( (res = read( m_serialPort, buffer + readInCount, readCount - readInCount )) < 0 )
            {
                trace( "%s(%d):ERROR(1): %s() [ trial: %u, readCount: %u, readInCount: %u ]\n", __FILE__, __LINE__, __func__, trials, readCount, readInCount );
                return -IV_ERROR_RECEIVE;
            }
        }
#endif
        if ( res == 0 )
        {
#if defined(_MSC_VER) || defined(__MINGW32__)
            trace( "%s(%d):WARNING(1): %s() [ Leseversuch ohne Daten. (WinError: %u)]\n", __FILE__, __LINE__, __func__, GetLastError() );
#else
            trace( "%s(%d):WARNING(1): %s() [ Leseversuch ohne Daten. ]\n", __FILE__, __LINE__, __func__ );
#endif
            break;
        }
        else if ( static_cast<uint32_t>(res) < (readCount - readInCount) )
        {
            trace( "%s(%d):WARNING(1): %s() [ Nachlesen erforderlich. trial: %u, waren %u, gefordert %u, bereits %u ]\n", __FILE__, __LINE__, __func__, trials, res, readCount - readInCount, readInCount );
        }

        readInCount += res;
        // Gesamtanzahl gelesener Bytes fuer Statistik halten
        m_sizeRead += res;
    }

    // Wurde gewuenschte Anzahl nicht empfangen?
    if ( readCount != readInCount )
    {
        trace( "%s(%d):ERROR(1): %s() [ trial: %u, readCount: %u, readInCount: %u ]\n", __FILE__, __LINE__, __func__, trials, readCount, readInCount );
        return -IV_ERROR_RECEIVE;
    }

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
* @brief Schreiben in die serielle Kommunikationsschnittstelle.
*/
int32_t Datafox::Communication::ConnectionBase::writeSerial( const uint8_t *buffer, uint32_t writeCount, uint32_t &writeOutCount, uint8_t writeTrials )
{
    trace( "%s(%d):INFO(3): %s( %p, %u, %u, %u )\n", __FILE__, __LINE__, __func__, buffer, writeCount, writeOutCount, writeTrials );

#if !defined(_MSC_VER) && !defined(__MINGW32__)
    errno = 0;
#endif

#if defined(_MSC_VER) || defined(__MINGW32__)
    uint32_t res{ IV_ERROR_SUCCESS };
#else
    int32_t res{ IV_ERROR_SUCCESS };
#endif
    uint8_t trials = 0;
    writeOutCount = 0;

    for ( ; (writeCount != writeOutCount) && (trials < writeTrials); ++trials )
    {
#if defined(_MSC_VER) || defined(__MINGW32__)
        if ( !WriteFile( m_serialPort, buffer + writeOutCount, DWORD(writeCount - writeOutCount), (LPDWORD)&res, nullptr ) )
        {
            if ( GetLastError() == ERROR_FILE_NOT_FOUND )
            {
                trace( "%s(%d):ERROR(1): %s() [ Schnittstelle wird geschlossen, error: %u ]\n", __FILE__, __LINE__, __func__, GetLastError() );
                close();
            }
#else

        if ( (res = write( m_serialPort, (unsigned char*)buffer + writeOutCount, writeCount - writeOutCount )) < 0 )
        {
#endif
#if defined(_MSC_VER) || defined(__MINGW32__)
            trace( "%s(%d):ERROR(1): %s() [ trial: %u, writeCount: %u, writeOutCount: %u, error: %u ]\n", __FILE__, __LINE__, __func__, trials, writeCount, writeOutCount, GetLastError() );
            if ( GetLastError() == ERROR_ALREADY_EXISTS )
            {            
                // 2014-04-03
                // Workaround um bei USB-COM-Ports ein blockieren zu verhindern.
                std::this_thread::sleep_for( std::chrono::seconds( 3 ) );
            }
#else
            trace( "%s(%d):ERROR(1): %s() [ trial: %u, writeCount: %u, writeOutCount: %u ]\n", __FILE__, __LINE__, __func__, trials, writeCount, writeOutCount );
#endif
            return -IV_ERROR_TRANSMIT;
        }

        // Wurden alle Zeichen versendet?
        if ( res == 0 )
        {
            trace( "%s(%d):WARNING(1): %s() [ Es wurden keine Daten gesendet. ]\n", __FILE__, __LINE__, __func__ );
            break;

        }
        else if ( static_cast<uint32_t>(res) < (writeCount - writeOutCount) )
        {		
            trace( "%s(%d):WARNING(1): %s() [ Nachsenden erforderlich. trial: %u, waren %u, gefordert %u, bereits %u]\n", __FILE__, __LINE__, __func__, trials, res, writeCount - writeOutCount, writeOutCount );
        }

        writeOutCount += res;
        // Gesamtanzahl geschriebener Bytes fuer Statistik halten
        m_sizeWritten += res;
    }

    // Wurden nicht alle Zeichen versendet?
    if ( writeCount != writeOutCount )
    {
        trace( "%s(%d):ERROR(1): %s() [ trial: %u, writeCount: %u, writeOutCount: %u ]\n", __FILE__, __LINE__, __func__, trials, writeCount, writeOutCount );
        return -IV_ERROR_TRANSMIT;
    }

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
* @brief Oeffnen einer TCP/IP Kommunikationsschnittstelle.
*/
int32_t Datafox::Communication::ConnectionBase::openSocket( const std::string &hostName, uint16_t portNumber, uint16_t communicationReadTimeout )
{
    trace( "%s(%d):INFO(3): %s( %s, %u, %u )\n", __FILE__, __LINE__, __func__, hostName.c_str(), portNumber, communicationReadTimeout );

    SOCKADDR_IN sin;
    memset( &sin, 0, sizeof(sin) );
    sin.sin_family = AF_INET;
    sin.sin_port = htons( portNumber );
    sin.sin_addr.s_addr = inet_addr( hostName.data() );

    // Kanal bereits belegt?
    if ( type() != ConnectionType::undefined )
    {
        return IV_ERROR_FAIL;
    }

    if ( sin.sin_addr.s_addr == INADDR_NONE )
    { 
        // Adresse ueber DNS aufloesen
        HOSTENT *he;
        he = gethostbyname( hostName.data() );
        if ( !he )
        {
            trace( "%s(%d):ERROR(1): %s() [ gethostbyname fail (%u) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );
            return IV_ERROR_FAIL;
        }

        sin.sin_addr.s_addr = *((uint32_t*)he->h_addr);
        //sin.sin_addr.s_addr = inet_addr( he->h_addr );
    }

    if ( (m_socket = ::socket( AF_INET, SOCK_STREAM, 0 )) == INVALID_SOCKET )
    {
        trace( "%s(%d):ERROR(1): %s() [ socket fail (%u) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );
        return IV_ERROR_FAIL;
    }

    if ( connect( m_socket, (SOCKADDR*)&sin, sizeof(sin) ) == SOCKET_ERROR )
    {
        trace( "%s(%d):ERROR(1): %s() [ connect fail (%u) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );
        closeSocket();
        return IV_ERROR_FAIL;
    }

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
* @brief Schliessen der TCP/IP Kommunikationsschnittstelle.
*/
int32_t Datafox::Communication::ConnectionBase::closeSocket()
{
    trace( "%s(%d):INFO(3): %s()\n", __FILE__, __LINE__, __func__ );

#if defined(_MSC_VER) || defined(__MINGW32__)
    if ( closesocket( m_socket ) == SOCKET_ERROR )
    {
        trace( "%s(%d):ERROR(1): %s() [ closesocket fail (%u) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );
        return IV_ERROR_FAIL;
    }
#else
    if ( shutdown( m_socket, 2 ) == SOCKET_ERROR )
    {
        trace( "%s(%d):ERROR(1): %s() [ shutdown fail (%u) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );
    }
    if ( ::close( m_socket ) == SOCKET_ERROR )
    {
        trace( "%s(%d):ERROR(1): %s() [ close fail (%u) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );
        return IV_ERROR_FAIL;
    }
#endif

    m_socket = INVALID_SOCKET;

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
* @brief Lesen aus der TCP/IP Kommunikationsschnittstelle.
*/
int32_t Datafox::Communication::ConnectionBase::readSocket( uint8_t *buffer, uint32_t readCount, uint32_t &readInCount, uint8_t readTrials )
{
    trace( "%s(%d):INFO(3): %s( %p, %u, %u, %u )\n", __FILE__, __LINE__, __func__, buffer, readCount, readInCount, readTrials );

#if defined(_MSC_VER)
#if !defined(MYCONST)
#define MYCONST const
#endif
#else
#define MYCONST
#endif
    struct timeval seltime;
    fd_set readfds;

    int32_t res;
    uint8_t trials = 0;
    uint8_t lTrial = (uint8_t)readCount;

    readInCount = 0;

    // Daten emfpangen
    for ( ; (readCount != readInCount) && (trials < readTrials); ++trials )
    {
        FD_ZERO( (fd_set FAR *) &readfds );
        FD_SET( m_socket, (fd_set FAR *) &readfds );
        seltime.tv_sec  = m_readTimeout / 1000;
        seltime.tv_usec = (m_readTimeout % 1000) * 1000;

        if ( (res = select( FD_SETSIZE,
            (fd_set FAR *)&readfds, (fd_set FAR *) 0, (fd_set FAR *) 0,
            (MYCONST struct timeval FAR *)&seltime )) < 0 )
        {
            trace( "%s(%d):ERROR(1): %s() [ id: %d, trial: %u, readCount: %u, readInCount: %u, error: %u ]\n", __FILE__, __LINE__, __func__, number(), trials, readCount, readInCount, WSAGetLastError() );
            return -IV_ERROR_RECEIVE;
        }

        // Liegen Daten an?
        if ( res )
        {
            if ( (res = recv( m_socket, (char *)buffer + readInCount, int32_t(readCount - readInCount), 0 )) == SOCKET_ERROR )
            {
#if defined(_MSC_VER) || defined(__MINGW32__)
                if ( WSAECONNRESET == WSAGetLastError() || WSAECONNABORTED == WSAGetLastError() )
                {
                    trace( "%s(%d):ERROR(1): %s() [ id: %d, recv error: %u ]\n", __FILE__, __LINE__, __func__, number(), WSAGetLastError() );
                    return -IV_ERROR_CHANNEL_OPEN_REMOTE;
                }
#endif
                trace( "%s(%d):ERROR(1): %s() [ id: %d, trial: %u, readCount: %u, readInCount: %u, error: %u ]\n", __FILE__, __LINE__, __func__, number(), trials, readCount, readInCount, WSAGetLastError() );
                return -IV_ERROR_RECEIVE;
            }
            else if ( res == 0 )
            {
                // Wenn Daten gemeldet werden,
                // die Anzahl aber 0 ist, so wurde die Verbindung beendet.

                // Verbindung wurde von Gegenseite geschlossen.
                trace( "%s(%d):ERROR(1): %s() [ id: %d, trial: %u, readCount: %u, readInCount: %u ]\n", __FILE__, __LINE__, __func__, number(), trials, readCount, readInCount );
                return -IV_ERROR_CHANNEL_OPEN_REMOTE;
            }
            else if ( res < int32_t(readCount - readInCount) )
            {
                trace( "%s(%d):WARNING(1): %s() [ Nachlesen erforderlich. id: %d, trial: %u, waren %u, gefordert %u, bereits %u ]\n", __FILE__, __LINE__, __func__, number(), trials, res, readCount - readInCount, readInCount );
                // Fuer Kommunikation ueber Com-Server mit Mehrfachdocking viele Teilpakete
                if ( lTrial )
                    trials--;
                else
                    lTrial = uint8_t(readCount - readInCount);

                lTrial--;
            }
        }
        else
        {
            trace( "%s(%d):WARNING(1): %s() [ Es wurden keine Daten gelesen. id: %d ]\n", __FILE__, __LINE__, __func__, number() );
            break;
        }

        readInCount += static_cast<uint32_t>( res );
        // Gesamtanzahl gelesener Bytes fuer Statistik halten
        m_sizeRead += static_cast<uint32_t>( res );
    }

    // Wurde gewuenschte Anzahl nicht empfangen?
    if ( readCount != readInCount )
    {
        trace( "%s(%d):ERROR(1): %s() [ id: %d, trial: %u, readCount: %u, readInCount: %u ]\n", __FILE__, __LINE__, __func__, number(), trials, readCount, readInCount );
        return -IV_ERROR_RECEIVE;
    }

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
* @brief Schreiben in die TCP/IP Kommunikationsschnittstelle.
*/
int32_t Datafox::Communication::ConnectionBase::writeSocket( const uint8_t *buffer, uint32_t writeCount, uint32_t &writeOutCount, uint8_t writeTrials )
{
    trace( "%s(%d):INFO(3): %s( %p, %u, %u, %u )\n", __FILE__, __LINE__, __func__, buffer, writeCount, writeOutCount, writeTrials );

    int32_t res;
    uint8_t trials = 0;

    writeOutCount = 0;

    for ( ; (writeCount != writeOutCount) && (trials < writeTrials); ++trials )
    {
        // Daten senden
        if ( (res = send( m_socket, reinterpret_cast<const char *>( buffer ) + writeOutCount, static_cast<int32_t>( writeCount - writeOutCount ), 0 )) == SOCKET_ERROR )
        {
#if defined(_MSC_VER) || defined(__MINGW32__)
            if ( WSAECONNRESET == WSAGetLastError() || WSAECONNABORTED == WSAGetLastError() )
            {
                trace( "%s(%d):ERROR(1): %s) [ id: %d, send error: %u ]\n", __FILE__, __LINE__, __func__, number(), WSAGetLastError() );
                return -IV_ERROR_CHANNEL_OPEN_REMOTE;
            }
#endif
            trace( "%s(%d):ERROR(1): %s() [ id: %d, trial: %u, writeCount %u, writeOutCount: %u, Error: %u ]\n", __FILE__, __LINE__, __func__, number(), trials, writeCount, writeOutCount, WSAGetLastError() );
            return -IV_ERROR_TRANSMIT;
        }

        // Wurden alle Daten versendet?
        if ( res == 0 )
        {
            trace( "%s(%d):WARNING(1): %s() [ Es wurden keine Daten gesendet. id: %d ]\n", __FILE__, __LINE__, __func__, number() );
            break;
        }
        else if ( res < static_cast<int32_t>( writeCount - writeOutCount ) )
        {
            trace( "%s(%d):WARNING(1): %s() [ Nachsenden erforderlich. id: %d, trial: %u, waren %u, gefordert %u, bereits %u ]\n", __FILE__, __LINE__, __func__, number(), trials, res, writeCount - writeOutCount, writeOutCount );
        }

        writeOutCount += static_cast<uint32_t>( res );
        // Gesamtanzahl geschriebener Bytes fuer Statistik halten
        m_sizeWritten += static_cast<uint32_t>( res );
    }

    // Wurden nicht alle Daten im vorgegebenen Rahmen versendet?
    if ( writeCount != writeOutCount )
    {
        trace( "%s(%d):ERROR(1): %s() [ id: %d, trial: %u, writeCount %u, writeOutCount: %u ]\n", __FILE__, __LINE__, __func__, number(), trials, writeCount, writeOutCount );
        return -IV_ERROR_TRANSMIT;
    }

    // Ausfuehrung ist erfolgt.
    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
* @brief Verwerfen der Daten an der TCP/IP Kommunikationsschnittstelle.
*/
int32_t Datafox::Communication::ConnectionBase::flushSocket()
{
    trace( "%s(%d):INFO(3): %s()\n", __FILE__, __LINE__, __func__ );

#if defined(_MSC_VER)
#define MYCONST const
#else
#define MYCONST
#endif
    struct timeval seltime;
    fd_set readfds;

    int32_t res;
    char buffer[1024];

    // Daten emfpangen
    for ( ;; )
    {
        FD_ZERO( (fd_set FAR *) &readfds );
        FD_SET( m_socket, (fd_set FAR *) &readfds );
        seltime.tv_sec  = 0;
        seltime.tv_usec = 500000;

        if ( (res = select( FD_SETSIZE,
            (fd_set FAR *)&readfds, (fd_set FAR *) 0, (fd_set FAR *) 0,
            (MYCONST struct timeval FAR *)&seltime )) < 0 )
        {
            trace( "%s(%d):ERROR(1): %s() [ id: %d, error: %d ]\n", __FILE__, __LINE__, __func__, number(), WSAGetLastError() );
            return -IV_ERROR_RECEIVE;
        }

        // Liegen Daten an?
        if ( res )
        {
            if ( (res = recv( m_socket, buffer, sizeof(buffer), 0 )) == SOCKET_ERROR )
            {
#if defined(_MSC_VER) || defined(__MINGW32__)
                if ( WSAECONNRESET == WSAGetLastError() || WSAECONNABORTED == WSAGetLastError() )
                {
                    trace( "%s(%d):ERROR(1): %s() [ id: %d, recv error: %d ]\n", __FILE__, __LINE__, __func__, number(), WSAGetLastError() );
                    return -IV_ERROR_CHANNEL_OPEN_REMOTE;
                }
#endif
                trace( "%s(%d):ERROR(1): %s() [ id: %d, error: %d ]\n", __FILE__, __LINE__, __func__, number(), WSAGetLastError() );
                return -IV_ERROR_RECEIVE;
            }
            else if ( res == 0 )
            {
                // Wenn Daten gemeldet werden,
                // die Anzahl aber 0 ist, so wurde die Verbindung beendet.

                // Verbindung wurde von Gegenseite geschlossen.
                trace( "%s(%d):ERROR(1): %s() [ id: %d ]\n", __FILE__, __LINE__, __func__, number() );
                return -IV_ERROR_CHANNEL_OPEN_REMOTE;
            }
        }
        else
        {
            break;
        }
    }

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
* @brief Lesen aus der Kommunikationsschnittstelle.
*/
int32_t Datafox::Communication::ConnectionBase::readData( uint8_t *buffer, uint32_t readCount, uint32_t &readInCount, uint8_t readTrials )
{
    trace( "%s(%d):INFO(3): %s( %p, %u, %u, %u )\n", __FILE__, __LINE__, __func__, buffer, readCount, readInCount, readTrials );

    int32_t result = IV_ERROR_SUCCESS;
    switch ( type() )
    {
    case ConnectionType::serial:
        result = readSerial( buffer, readCount, readInCount, readTrials );
        break;

    case ConnectionType::socket:
        result = readSocket( buffer, readCount, readInCount, readTrials );
        break;

    default:
        trace( "%s(%d):ERROR(1): %s() [ unknown connection type: %d ]\n", __FILE__, __LINE__, __func__, type() );
        result = -IV_ERROR_RECEIVE;
    }

    return result;
}

// ##################################################################
/**
* @brief Schreiben in die Kommunikationsschnittstelle.
*/
int32_t Datafox::Communication::ConnectionBase::writeData( const uint8_t *buffer, uint32_t writeCount, uint32_t &writeOutCount, std::uint8_t writeTrials )
{
    trace( "%s(%d):INFO(3): %s( %p, %u, %u, %u )\n", __FILE__, __LINE__, __func__, buffer, writeCount, writeOutCount, writeTrials );

    int32_t result = IV_ERROR_SUCCESS;
    switch ( type() )
    {
    case ConnectionType::serial:
        result = writeSerial( buffer, writeCount, writeOutCount, writeTrials );
        break;

    case ConnectionType::socket:
        result = writeSocket( buffer, writeCount, writeOutCount, writeTrials );
        break;

    default:
        trace( "%s(%d):ERROR(1): %s() [ unknown connection type: %d ]\n", __FILE__, __LINE__, __func__, type() );
        result = -IV_ERROR_TRANSMIT;
    }

    return result;
}

