/****************************************************************************
**
** Copyright (C) Datafox GmbH.
** All rights reserved.
** Contact: Datafox GmbH (develop@datafox.de)
**
** This file contains the class for processing XML configuration files
** (read/write).
**
****************************************************************************/
#include "file/file_xml.h"
#include "file/file_aes.h"
#include "common/utils.h"
#include <boost/format.hpp>
#include <boost/algorithm/string.hpp>
#include <algorithm>
#include <iostream>
#include <fstream>

std::string FileXML::TAB = "    ";// "\t";
std::string FileXML::LT = "&lt;";
std::string FileXML::GT = "&gt;";
std::string FileXML::sTAG = "";//"<small>";
std::string FileXML::eTAG = "";// "</small>";

std::map<FileXML::ContentType, std::string> FileXML::nodeNames = {
    {ContentType::MainNode, "documentStudio"},
    {ContentType::FileHeader, "header"},
    {ContentType::FlashHeader, "flashHeader"},
    {ContentType::ImageScope, "imageScope"},
    {ContentType::ExtendedScope, "extendedScope"},
    {ContentType::Config, "config"},
    {ContentType::GlobalVar, "globalVar"},
    {ContentType::GlobalVarValue, "valueGlobalVar"},
    {ContentType::ConfigRoot, "configRoot"},
    {ContentType::Operating, "operating"},
    {ContentType::Menu, "menu"},
    {ContentType::MenuItem, "menuItem"},
    {ContentType::ReadRecord, "readRecord"},
    {ContentType::ReadField, "readField"},
    {ContentType::ReadField_TextEx, "textEx"},
    {ContentType::ReadField_ConfirmationEx, "confirmationEx"},
    {ContentType::ReadField_NumFieldEx, "numFieldEx"},
    {ContentType::ReadField_ListSelectionEx, "listSelectionEx"},
    {ContentType::ReadField_DataEx, "dataEx"},
    {ContentType::ReadField_GlobVarToHIDEx, "globVarToHIDEx"},
    {ContentType::ReadField_SwitchModuleEnergieEx, "switchModuleEnergieEx"},
    {ContentType::ReadField_SystemMessageEx, "systemMessageEx"},
    {ContentType::ReadField_TakePhotoEx, "takePhotoEx"},
    {ContentType::ReadField_FeverControl, "feverControlEx"},
    {ContentType::ReadField_MultiInputSource, "multiInputSourceEx"},
    {ContentType::Transponder, "transponder"},
    {ContentType::TransponderGlobal, "transponderGlobal"},
    {ContentType::TransponderAccessControl, "transponderAccessControl"},
    {ContentType::Compare, "compare"},
    {ContentType::Records, "records"},
    {ContentType::RecordDescription, "recordDescription"},
    {ContentType::RecordField, "recordField"},
    {ContentType::Lists, "lists"},
    {ContentType::ExtendedFlash, "extendedFlash"},
    {ContentType::MachineDefinitions, "machineDefinitions"},
    {ContentType::Machine, "machine"},
    {ContentType::VibrationSensor, "vibrationSensor"},
    {ContentType::SignalProcessing, "signalProcessing"},
    {ContentType::DigitalInputs, "digitalInputs"},
    {ContentType::DigitalInput, "digitalInput"},
    {ContentType::DigitalInputEx, "digitalInputExt"},
    {ContentType::AnalogInputs, "analogInputs"},
    {ContentType::AnalogInput, "analogInput"},
    {ContentType::Timers, "timers"},
    {ContentType::Timer, "timer"},
    {ContentType::EventChains, "eventChains"},
    {ContentType::IOModule, "ioModule"},
    {ContentType::AccessControl, "accessControl"},
    {ContentType::Access, "access"},
    {ContentType::PowerSetupNode, "powerSetupNode"},
    {ContentType::PowerSetup, "powerSetup"},
    {ContentType::PowerSetup_CPU, "cpuModule"},
    {ContentType::PowerSetup_Timer, "timerModule"},
    {ContentType::PowerSetup_Inputs, "inputsModule"},
    {ContentType::PowerSetup_DeviceGlobal, "deviceGlobalModule"},
    {ContentType::PowerSetup_U5V, "U5VModule"},
    {ContentType::PowerSetup_U12V, "U12VModule"},
    {ContentType::PowerSetup_Communication, "communicationModule"},
    {ContentType::PowerSetup_DisplayLight, "displayLightModule"},
    {ContentType::PowerSetup_DisplayText, "displayTextModule"},
    {ContentType::PowerSetup_Fingerprint, "fingerprintModule"},
    {ContentType::PowerSetup_Transponder, "transponderModule"},
    {ContentType::PowerSetup_GPS, "gpsModule"},
    {ContentType::PowerSetup_USBHost, "usbHostModule"},
    {ContentType::PowerSetup_MMC, "mmcModule"},
    {ContentType::PowerEvent, "powerEvent"},
    {ContentType::PowerMemo, "memo"},
    {ContentType::TransponderConfiguration, "transponderConfiguration"},
    {ContentType::TransponderLegicAdvantHeader, "transponderLegicAdvantHeader"},
    {ContentType::TransponderLegicAdvant, "transponderLegicAdvant"},
    {ContentType::TransponderMifareDesfireHeader, "transponderMifareDesfireHeader"},
    {ContentType::TransponderMifareDesfire, "transponderMifareDesfire"},
    {ContentType::TransponderMifarePlusHeader, "transponderMifarePlusHeader"},
    {ContentType::TransponderMifarePlus, "transponderMifarePlus"},
    {ContentType::TransponderMifareUltralightHeader, "transponderMifareUltralightHeader"},
    {ContentType::TransponderMifareUltralight, "transponderMifareUltralight"},
    {ContentType::TransponderiCodeHeader, "transponderiCodeHeader"},
    {ContentType::TransponderiCode, "transponderiCode"},
    {ContentType::TransponderJavaCardHeader, "transponderJavaCardHeader"},
    {ContentType::TransponderJavaCard, "transponderJavaCard"},
    {ContentType::TransponderMyDHeader, "transponderMyDHeader"},
    {ContentType::TransponderMyD, "transponderMyD"},
    {ContentType::TransponderISOTagsHeader, "transponderISOTagsHeader"},
    {ContentType::TransponderISOTags, "transponderISOTags"},
    {ContentType::TransponderWiegandClockDataHeader, "transponderWiegandClockDataHeader"},
    {ContentType::TransponderWiegandClockData, "transponderWiegandClockData"},
    {ContentType::TransponderETSITagsHeader, "transponderETSITagsHeader"},
    {ContentType::TransponderETSITags, "transponderETSITags"},
    {ContentType::TransponderQ5Header, "transponderQ5Header"},
    {ContentType::TransponderQ5, "transponderQ5"},
    {ContentType::TransponderHitag1FeigHeader, "transponderHitag1FeigHeader"},
    {ContentType::TransponderHitag1Feig, "transponderHitag1Feig"},
    {ContentType::TransponderPreconfiguredReaderHeader, "transponderPreconfiguredReaderHeader"},
    {ContentType::TransponderPreconfiguredReader, "transponderPreconfiguredReader"},
    {ContentType::Extended, "extended"},
    {ContentType::Fingerprint, "fingerprint"},
    {ContentType::TimeboyAdaption, "timeboyAdaption"},
    {ContentType::GPRS, "GPRS"},
    {ContentType::EnergieOption, "energieOption"},
    {ContentType::GPS, "GPS"},
    {ContentType::ProtectionPassword, "protectionPassword"},
    {ContentType::AutoKeys, "autoKeys"},
    {ContentType::AutoKey, "autoKey"},
    {ContentType::SystemMessages, "systemMessages"},
    {ContentType::SystemMessage, "systemMessage"},
};

FileXML::FileXML()
{
}

void FileXML::setVersions( const std::string &studioVersion, const std::string &libraryVersion )
{
    m_studioVersion.parse( studioVersion );
    m_libraryVersion.parse( libraryVersion );
}

/** Methode zur Initialisierung der XML-Datei.

Zuvor erfolgte Initialisierungen werden ohne Meldung überschrieben. Es wird die grundlegende XML-Struktur
aufgebaut und der Wurzelknoten gesetzt.

@param studioVersion Version von Datafox Studio
@param libraryVersion Version der DLL

@return @ref ErrorCodes
*/
FileXML::ErrorCodes FileXML::init()
{
    auto result = reset();

    if ( result == ErrorCodes::Success )
    {
        // Loading the doxygen tag file to determine the links in the documentation.
        m_pDoxygenTagXml = std::make_shared<tinyxml2::XMLDocument>( new tinyxml2::XMLDocument() );
        std::string tagFilePath = Utils::applicationDirectoryPath() + "/data/res/doxygen_tag_file.xml";
        if ( m_pDoxygenTagXml->LoadFile( tagFilePath.c_str() ) != tinyxml2::XML_SUCCESS )
        {
            return ErrorCodes::LoadingHelpFileFailed;
        }
        m_pDoxygenRoot = m_pDoxygenTagXml->FirstChildElement( "tagfile" );
    }
    return result;
}

FileXML::ErrorCodes FileXML::reset()
{
    // ACHTUNG: Der neu erstellte Knoten wird per Zeiger übergeben und ist fortan im Besitz von tinyXml.
    m_xmlDoc = std::make_shared<tinyxml2::XMLDocument>( new tinyxml2::XMLDocument() );
    if ( !m_xmlDoc )
    {
        return ErrorCodes::DocumentCreationFailed;
    }

    tinyxml2::XMLDeclaration *pDecl = m_xmlDoc->NewDeclaration( "xml version=\"1.0\" encoding=\"UTF-8\"" );
    m_xmlDoc->LinkEndChild( pDecl );

    m_parent = m_xmlDoc->NewElement( "documentStudio" );
    m_xmlDoc->LinkEndChild( m_parent );

    return ErrorCodes::Success;
}

FileXML::ErrorCodes FileXML::save( const std::string &filePath )
{
    // Zieldatei öffnen, wenn nötig vorher erstellen
    FILE *pFile{};
    FILE *pHelpFile{};
    if ( (pFile = fopen( filePath.c_str(), "w+" )) == NULL )
    {
        return ErrorCodes::FileCreationFailed;
    }

    ErrorCodes result = ErrorCodes::Success;
    if ( m_xmlDoc->SaveFile( pFile ) != tinyxml2::XML_SUCCESS )
    {
        result = ErrorCodes::FileSaveFailed;
    }
    fclose( pFile );

    // Depending on the configuration in the file DFCom.ini an additional help file (*.html) is created.
    if ( m_createHelpFile )
    {
        std::string helpPath = Utils::File::dir( filePath ) + "/" + Utils::File::name( filePath ) + ".html";
        writeHelpFile( helpPath );
    }

    return result;
}

void FileXML::processItem( std::ofstream &helpFile, tinyxml2::XMLElement *element, std::string shifting, const tinyxml2::XMLNode *lastValidCompound )
{
    std::string docRef;
    if ( element != nullptr )
    {
        const char *tag = changeSpecialTag( element->Name() );
        if ( tag != nullptr )
        {
            auto docCompound = findStructCompound( tagToStruct( tag ) );
            if ( docCompound != nullptr )
            {
                lastValidCompound = docCompound;
                auto filename = docCompound->FirstChildElement( "filename" );
                if ( filename != nullptr )
                {
                    docRef = std::string( filename->GetText() );
                }
            }

            auto child = element->FirstChildElement();
            bool hasChilds = (child != nullptr);
            if ( hasChilds )
            {
                helpFile << shifting << sTAG << LT;
                if ( !docRef.empty() )
                {
                    helpFile << "<a href=\"#\"" << " onclick=\"return open_new_window('" << docRef << "');\">" << element->Name() << "</a>";
                }
                else
                {
                    helpFile << element->Name();
                }

                // Copy the attributes
                auto attribute = element->FirstAttribute();
                while ( attribute != nullptr )
                {
                    if ( (attribute->Name() != nullptr) && (attribute->Value() != nullptr) )
                    {
                        helpFile << " " << attribute->Name() << "=" << attribute->Value();
                    }
                    attribute = attribute->Next();
                }

                helpFile << GT << eTAG << std::endl;

                do
                {
                    processItem( helpFile, child, shifting + TAB, lastValidCompound );
                    child = child->NextSiblingElement();
                }
                while ( child != nullptr );

                helpFile << shifting << sTAG << LT << "/";
                if ( !docRef.empty() )
                {
                    helpFile << "<a href=\"#\"" << " onclick=\"return open_new_window('" << docRef << "');\">" << element->Name() << "</a>";
                }
                else
                {
                    helpFile << element->Name();
                }
                helpFile << GT << eTAG << std::endl;
            }
            else
            {
                auto text = element->GetText();

                docRef.clear();
                //auto compound = (docCompound == nullptr) ? lastValidCompound : docCompound;
                if ( lastValidCompound != nullptr )
                {
                    auto member = lastValidCompound->FirstChildElement( "member" );
                    while ( member != nullptr )
                    {
                        auto name = member->FirstChildElement( "name" );
                        if ( name == nullptr )
                        {
                            break;
                        }

                        // Found definition for current element.
                        if ( std::string( name->GetText() ).compare( element->Name() ) == 0 )
                        {
                            auto anchorfile = member->FirstChildElement( "anchorfile" );
                            auto anchor = member->FirstChildElement( "anchor" );
                            if ( (anchorfile == nullptr) || (anchor == nullptr) )
                            {
                                break;
                            }

                            const char *filenameText = anchorfile->GetText();
                            const char *anchorText = anchor->GetText();
                            if ( (filenameText != nullptr) && (anchorText != nullptr) )
                            {
                                docRef = std::string( filenameText ) + "#" + std::string( anchorText );
                            }
                        }
                        member = member->NextSiblingElement( "member" );
                    }
                }


                helpFile << shifting << sTAG << LT;
                if ( !docRef.empty() )
                {
                    helpFile << "<a href=\"#\"" << " onclick=\"return open_new_window('" << docRef << "');\">" << element->Name() << "</a>";
                }
                else
                {
                    helpFile << element->Name();
                }

                if ( text != nullptr )
                {
                    helpFile << GT << eTAG << "<b>" << text << "</b>" << sTAG << LT << "/";

                    if ( !docRef.empty() )
                    {
                        helpFile << "<a href=\"#\"" << " onclick=\"return open_new_window('" << docRef << "');\">" << element->Name() << "</a>";
                    }
                    else
                    {
                        helpFile << element->Name();
                    }
                    helpFile << GT << eTAG << std::endl;
                }
                else
                {
                    helpFile << "/" << GT << eTAG << std::endl;
                }
            }
        }
    }
}

const char *FileXML::changeSpecialTag( const char *name )
{
    static std::string powerModule = "powerModule";
    static std::string powerMemo = "powerMemo";
    static std::string transponder = "transponder";

    const char *result = name;
    if ( name != nullptr )
    {
        std::string tmp( name );
        if ( (tmp.compare( "cpuModule" ) == 0) ||
             (tmp.compare( "timerModule" ) == 0) ||
             (tmp.compare( "inputsModule" ) == 0) ||
             (tmp.compare( "deviceGlobalModule" ) == 0) ||
             (tmp.compare( "U5VModule" ) == 0) ||
             (tmp.compare( "U12VModule" ) == 0) ||
             (tmp.compare( "communicationModule" ) == 0) ||
             (tmp.compare( "displayLightModule" ) == 0) ||
             (tmp.compare( "displayTextModule" ) == 0) ||
             (tmp.compare( "fingerprintModule" ) == 0) ||
             (tmp.compare( "transponderModule" ) == 0) ||
             (tmp.compare( "gpsModule" ) == 0) ||
             (tmp.compare( "usbHostModule" ) == 0) ||
             (tmp.compare( "mmcModule" ) == 0) )
        {
            result = powerModule.c_str();
        }
        else if ( tmp.compare( "memo" ) == 0 )
        {
            result = powerMemo.c_str();
        }
        else if (( tmp.compare( "transponderGlobal" ) == 0 ) || (tmp.compare( "transponderAccessControl" ) == 0))
        {
            result = transponder.c_str();
        }
    }
    return result;
}

void FileXML::writeHelpFile( const std::string &path )
{
    std::ofstream helpFile( path.c_str() );

    // header
    helpFile << "<!DOCTYPE html>" << std::endl;
    helpFile << "<html lang=\"en\">" << std::endl;
    helpFile << "<head>" << std::endl;
    helpFile << TAB << "<meta charset=\"utf-8\">" << std::endl;
    helpFile << TAB << "<title>Dokumentation - " << Utils::to_utf8( Utils::File::name( path ) ) << "</title>" << std::endl;
    helpFile << TAB << "<style>" << std::endl;
    helpFile << TAB << TAB << "a{ color:#0000FF; }" << std::endl;
    helpFile << TAB << TAB << "a:visited{ color:#0000FF; }" << std::endl;
    helpFile << TAB << TAB << "a:hover{ color:#0000FF; }" << std::endl;
    helpFile << TAB << TAB << "a:active{ color:#0000FF; }" << std::endl;
    helpFile << TAB << "</style>" << std::endl;
    helpFile << TAB << "<script type=\"text/javascript\">" << std::endl;
    helpFile << TAB << TAB << "function open_new_window(link)" << std::endl;
    helpFile << TAB << TAB << "{" << std::endl;
    helpFile << TAB << TAB << TAB << "window.open(\"" << m_docPath << "/\" + link, 'doc', '');" << std::endl;
    helpFile << TAB << TAB << TAB << "return false;" << std::endl;
    helpFile << TAB << TAB << "}" << std::endl;
    helpFile << TAB << "</script>" << std::endl;
    helpFile << "<head>" << std::endl;

    // body
    helpFile << "<body><pre>" << std::endl;

    std::string shifting;
    auto root = m_xmlDoc->FirstChildElement( "documentStudio" );
    processItem( helpFile, root, shifting );

    helpFile << "</pre></body>" << std::endl;
    helpFile.close();
}

void FileXML::configureAdditionalHelpFile( bool createHelpFile, const std::string &path )
{
    m_createHelpFile = createHelpFile;
    m_docPath = path;
    Utils::File::replaceDirSeparator( m_docPath );
}

tinyxml2::XMLElement *FileXML::nodeByType( ContentType type )
{
    const std::string &name = nodeNameByType( type );
    // cancel processing if name is empty
    if ( name.empty() )
    {
        return nullptr;
    }

    // Finding the node NAME and return it. (creating if not exist).
    auto node = m_parent->LastChildElement( name.c_str() );
    // add if not exist
    if ( node == nullptr )
    {
        node = m_xmlDoc->NewElement( name.c_str() );
        m_parent->LinkEndChild( node );
    }
    return node;
}

const std::string &FileXML::nodeNameByType( ContentType type ) const
{
    auto it = nodeNames.find( type );
    if ( it != nodeNames.end() )
    {
        return it->second;
    }

    static std::string dummy;
    return dummy;
}

uint32_t FileXML::numberOfChildItems( ContentType type )
{
    const std::string &name = nodeNameByType( type );
    uint32_t count{};

    for ( auto child = m_parent->FirstChildElement( name.c_str() ); child; child = child->NextSiblingElement( name.c_str() ) )
    {
        ++count;
    }
    return count;
}


////////////////////////////////////////
// Methods to write and read values into the current (m_xmlDoc) xml document.
//

template<>
void FileXML::writeValue<uint32_t>( tinyxml2::XMLElement *node, const std::string &name, const uint32_t &value )
{
    if ( node != nullptr )
    {
        auto pItem = m_xmlDoc->NewElement( name.c_str() );
        node->LinkEndChild( pItem );
        pItem->LinkEndChild( m_xmlDoc->NewText( std::to_string( value ).c_str() ) );
    }
}
template<>
uint32_t &FileXML::readValue<uint32_t>( tinyxml2::XMLElement *node, tinyxml2::XMLElement *&lastItem, const std::string &name, uint32_t &value, bool &errorPresent )
{
    errorPresent |= (node == nullptr);
    if ( !errorPresent )
    {
        auto item = (lastItem == nullptr) ? node->FirstChildElement( name.c_str() ) : lastItem->NextSiblingElement( name.c_str() );
        if ( item != nullptr )
        {
            value = static_cast<uint32_t>(item->UnsignedText());
        }
        else
        {
            errorPresent = true;
        }
        lastItem = item;
    }
    return value;
}


template<>
void FileXML::writeValue<uint16_t>( tinyxml2::XMLElement *node, const std::string &name, const uint16_t &value )
{
    writeValue( node, name, static_cast<uint32_t>(value) );
}
template<>
uint16_t &FileXML::readValue<uint16_t>( tinyxml2::XMLElement *node, tinyxml2::XMLElement *&lastItem, const std::string &name, uint16_t &value, bool &errorPresent )
{
    uint32_t tmp{};
    readValue( node, lastItem, name, tmp, errorPresent );

    if ( !errorPresent )
    {
        value = static_cast<uint16_t>(tmp);
    }
    return value;
}


template<>
void FileXML::writeValue<uint8_t>( tinyxml2::XMLElement *node, const std::string &name, const uint8_t &value )
{
    writeValue( node, name, static_cast<uint32_t>(value) );
}
template<>
uint8_t &FileXML::readValue<uint8_t>( tinyxml2::XMLElement *node, tinyxml2::XMLElement *&lastItem, const std::string &name, uint8_t &value, bool &errorPresent )
{
    uint32_t tmp{};
    readValue( node, lastItem, name, tmp, errorPresent );

    if ( !errorPresent )
    {
        value = static_cast<uint8_t>(tmp);
    }
    return value;
}

template<>
void FileXML::writeValue<int32_t>( tinyxml2::XMLElement *node, const std::string &name, const int32_t &value )
{
    if ( node != nullptr )
    {
        auto pItem = m_xmlDoc->NewElement( name.c_str() );
        node->LinkEndChild( pItem );
        pItem->LinkEndChild( m_xmlDoc->NewText( std::to_string( value ).c_str() ) );
    }
}
template<>
int32_t &FileXML::readValue<int32_t>( tinyxml2::XMLElement *node, tinyxml2::XMLElement *&lastItem, const std::string &name, int32_t &value, bool &errorPresent )
{
    errorPresent |= (node == nullptr);
    if ( !errorPresent )
    {
        auto item = (lastItem == nullptr) ? node->FirstChildElement( name.c_str() ) : lastItem->NextSiblingElement( name.c_str() );
        if ( item != nullptr )
        {
            value = static_cast<int32_t>(item->IntText());
        }
        else
        {
            errorPresent = true;
        }
        lastItem = item;
    }
    return value;
}


template<>
void FileXML::writeValue<float>( tinyxml2::XMLElement *node, const std::string &name, const float &value )
{
    if ( node != nullptr )
    {
        auto pItem = m_xmlDoc->NewElement( name.c_str() );
        node->LinkEndChild( pItem );
        pItem->LinkEndChild( m_xmlDoc->NewText( std::to_string( value ).c_str() ) );
    }
}
template<>
float &FileXML::readValue<float>( tinyxml2::XMLElement *node, tinyxml2::XMLElement *&lastItem, const std::string &name, float &value, bool &errorPresent )
{
    errorPresent |= (node == nullptr);
    if ( !errorPresent )
    {
        auto item = (lastItem == nullptr) ? node->FirstChildElement( name.c_str() ) : lastItem->NextSiblingElement( name.c_str() );
        if ( item != nullptr )
        {
            value = static_cast<float>(item->DoubleText());
        }
        else
        {
            errorPresent = true;
        }
        lastItem = item;
    }
    return value;
}


template<>
void FileXML::writeValue<std::string>( tinyxml2::XMLElement *node, const std::string &name, const std::string &value )
{
    if ( node != nullptr )
    {
        auto pItem = m_xmlDoc->NewElement( name.c_str() );
        node->LinkEndChild( pItem );
        if ( !value.empty() )
        {
            pItem->LinkEndChild( m_xmlDoc->NewText( Utils::to_utf8( value ).c_str() ) );
        }
    }
}
template<>
bool FileXML::updateValue<std::string>( tinyxml2::XMLElement *node, const std::string &name, const std::string &value )
{
    if ( node != nullptr )
    {
        auto pItem = node->FirstChildElement( name.c_str() );
        if ( pItem != nullptr )
        {
            if ( !value.empty() )
            {
                pItem->SetValue( Utils::to_utf8( value ).c_str() );
                return true;
            }
        }
    }
    return false;
}
template<>
std::string &FileXML::readValue<std::string>( tinyxml2::XMLElement *node, tinyxml2::XMLElement *&lastItem, const std::string &name, std::string &value, bool &errorPresent )
{
    errorPresent |= (node == nullptr);
    if ( !errorPresent )
    {
        auto item = (lastItem == nullptr) ? node->FirstChildElement( name.c_str() ) : lastItem->NextSiblingElement( name.c_str() );
        if ( item != nullptr )
        {
            const char *text = item->GetText();
            if ( text != nullptr )
            {
                value = Utils::from_utf8( text );
            }
        }
        else
        {
            errorPresent = true;
        }
        lastItem = item;
    }
    return value;
}


template<>
void FileXML::writeValue<bool>( tinyxml2::XMLElement *node, const std::string &name, const bool &value )
{
    writeValue( node, name, value ? "true" : "false" );
}
template<>
bool &FileXML::readValue<bool>( tinyxml2::XMLElement *node, tinyxml2::XMLElement *&lastItem, const std::string &name, bool &value, bool &errorPresent )
{
    std::string tmp;
    readValue( node, lastItem, name, tmp, errorPresent );

    if ( !errorPresent )
    {
        boost::trim( tmp );
        boost::to_lower( tmp );
        if ( tmp == "true" /* || tmp == "1"*/ )
        {
            value = true;
        }
        else if ( tmp == "false" /*|| tmp == "0"*/ )
        {
            value = false;
        }
        else
        {
            // Fehlerhafter Feldwert
            errorPresent = true;
        }
    }
    return value;
}


void FileXML::writeValue( tinyxml2::XMLElement *node, const std::string &name, const char *value )
{
    if ( value != nullptr )
    {
        writeValue( node, name, std::string( value ) );
    }
}


void FileXML::writeValue( tinyxml2::XMLElement *node, const std::string &name, const uint8_t *value, int16_t size )
{
    writeValue( node, name, Utils::toHexString( value, size ) );
}

void FileXML::readValue( tinyxml2::XMLElement *node, tinyxml2::XMLElement *&lastItem, const std::string &name, uint8_t *value, uint16_t size, bool &errorPresent, bool nullTermination )
{
    errorPresent |= ((node == nullptr) || (value == nullptr));
    if ( !errorPresent )
    {
        auto item = (lastItem == nullptr) ? node->FirstChildElement( name.c_str() ) : lastItem->NextSiblingElement( name.c_str() );
        errorPresent = (item == nullptr);
        if ( !errorPresent && (size > 0) )
        {
            // clear data
            memset( value, 0x00, size );

            const char *text = item->GetText();
            if ( text != nullptr )
            {
                std::vector<uint8_t> tmp = Utils::fromHexString( text, false );

                // Reduction of the maximum array size used due to null termination.
                uint16_t maxSize = nullTermination ? size - 1 : size;
                if ( tmp.size() <= maxSize )
                {
                    for ( size_t i = 0; i < tmp.size(); ++i )
                    {
                        value[i] = tmp[i];
                    }
                }
                else
                {
                    errorPresent = true;
                }
            }
        }
        lastItem = item;
    }
}

void FileXML::readValue( tinyxml2::XMLElement *node, const std::string &name, uint8_t *value, uint16_t size, bool &errorPresent, bool nullTermination )
{
    tinyxml2::XMLElement *lastItem{};
    readValue( node, lastItem, name, reinterpret_cast<uint8_t *>(value), size, errorPresent, nullTermination );
}

void FileXML::readValue( tinyxml2::XMLElement *node, tinyxml2::XMLElement *&lastItem, const std::string &name, char *value, uint16_t size, bool &errorPresent, bool nullTermination )
{
    errorPresent |= ((node == nullptr) || (value == nullptr));
    if ( !errorPresent )
    {
        auto item = (lastItem == nullptr) ? node->FirstChildElement( name.c_str() ) : lastItem->NextSiblingElement( name.c_str() );
        if ( (item != nullptr) && (size > 0) )
        {
            // clear data
            memset( value, 0x00, size );

            const char *tmp = item->GetText();
            if ( tmp != nullptr )
            {
                std::string text = Utils::from_utf8( tmp );

                // Reduction of the maximum array size used due to null termination.
                uint16_t maxSize = nullTermination ? size - 1 : size;
                if ( text.size() <= maxSize )
                {
                    for ( size_t i = 0; i < text.size(); ++i )
                    {
                        value[i] = text[i];
                    }
                }
                else
                {
                    errorPresent = true;
                }
            }
        }
        else
        {
            errorPresent = true;
        }
        lastItem = item;
    }
}

void FileXML::readValue( tinyxml2::XMLElement *node, const std::string &name, char *value, uint16_t size, bool &errorPresent, bool nullTermination )
{
    tinyxml2::XMLElement *lastItem{};
    readValue( node, lastItem, name, value, size, errorPresent, nullTermination );
}


////////////////////////////////////////
// Methods to write and read complete structures into the current (m_xmlDoc) xml document.
//

template<>
void FileXML::writeContent<StructFileStudioHeader>( ContentType type, const StructFileStudioHeader &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "flags", data.flags );
        writeValue( subNode, "studioVersion", Utils::toHexString( data.studioVersion, true ) );
        writeValue( subNode, "minStudioVersion", Utils::toHexString( data.minStudioVersion, true ) );
        writeValue( subNode, "shouldStudioVersion", Utils::toHexString( data.shouldStudioVersion, true ) );
        writeValue( subNode, "minLibraryVersion", Utils::toHexString( data.minLibraryVersion, true ) );
        writeValue( subNode, "shouldLibraryVersion", Utils::toHexString( data.shouldLibraryVersion, true ) );
        writeValue( subNode, "minFirmware1Version", Utils::toHexString( data.minFirmware1Version, true ) );
        writeValue( subNode, "shouldFirmware1Version", Utils::toHexString( data.shouldFirmware1Version, true ) );
        writeValue( subNode, "minFirmware2Version", Utils::toHexString( data.minFirmware2Version, true ) );
        writeValue( subNode, "shouldFirmware2Version", Utils::toHexString( data.shouldFirmware2Version, true ) );
        writeValue( subNode, "minFirmware3Version", Utils::toHexString( data.minFirmware3Version, true ) );
        writeValue( subNode, "shouldFirmware3Version", Utils::toHexString( data.shouldFirmware3Version, true ) );
        writeValue( subNode, "size", data.size );
    }
}
template<>
bool FileXML::readContent<StructFileStudioHeader>( ContentType type, StructFileStudioHeader &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            std::string dummy;
            readValue( subNode, "flags", data.flags, errorPresent );
            data.studioVersion = Utils::fromHexString_uint( readValue( subNode, "studioVersion", dummy, errorPresent ), uint32_t() );
            data.minStudioVersion = Utils::fromHexString_uint( readValue( subNode, "minStudioVersion", dummy, errorPresent ), uint32_t() );
            m_minStudioVersion = SoftwarebuildVersion( data.minStudioVersion );
            data.shouldStudioVersion = Utils::fromHexString_uint( readValue( subNode, "shouldStudioVersion", dummy, errorPresent ), uint32_t() );
            data.minLibraryVersion = Utils::fromHexString_uint( readValue( subNode, "minLibraryVersion", dummy, errorPresent ), uint32_t() );
            data.shouldLibraryVersion = Utils::fromHexString_uint( readValue( subNode, "shouldLibraryVersion", dummy, errorPresent ), uint32_t() );
            data.minFirmware1Version = Utils::fromHexString_uint( readValue( subNode, "minFirmware1Version", dummy, errorPresent ), uint32_t() );
            data.shouldFirmware1Version = Utils::fromHexString_uint( readValue( subNode, "shouldFirmware1Version", dummy, errorPresent ), uint32_t() );
            data.minFirmware2Version = Utils::fromHexString_uint( readValue( subNode, "minFirmware2Version", dummy, errorPresent ), uint32_t() );
            data.shouldFirmware2Version = Utils::fromHexString_uint( readValue( subNode, "shouldFirmware2Version", dummy, errorPresent ), uint32_t() );
            data.minFirmware3Version = Utils::fromHexString_uint( readValue( subNode, "minFirmware3Version", dummy, errorPresent ), uint32_t() );
            data.shouldFirmware3Version = Utils::fromHexString_uint( readValue( subNode, "shouldFirmware3Version", dummy, errorPresent ), uint32_t() );
            readValue( subNode, "minFirmware3Version", data.size, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructAutoKey>( ContentType type, const StructAutoKey &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "function", static_cast<uint8_t>(data.function) );
        writeValue( subNode, "hour", data.hour );
        writeValue( subNode, "minute", data.minute );
        writeValue( subNode, "weekdays", static_cast<uint8_t>(data.weekdays) );
    }
}
template<>
bool FileXML::readContent<StructAutoKey>( ContentType type, StructAutoKey &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            uint8_t tmp{};
            data.function = static_cast<EnumAutoKeyFunction>(readValue( subNode, "function", tmp, errorPresent ));
            readValue( subNode, "hour", data.hour, errorPresent );
            readValue( subNode, "minute", data.minute, errorPresent );
            data.weekdays = static_cast<EnumAutoKeyWeekday>(readValue( subNode, "weekdays", tmp, errorPresent ));
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructAutoKeys>( ContentType type, const StructAutoKeys &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "version", data.version );

        // Update parent node for sub-structure.
        auto oldParent = setParentNode( subNode );

        for ( size_t i = 0; i < sizeof( data.key ) / sizeof( data.key[0] ); ++i )
        {
            writeContent( ContentType::AutoKey, data.key[i] );
        }

        // Reset parent node because the end of the structure is reached.
        setParentNode( oldParent );
    }
}
template<>
bool FileXML::readContent<StructAutoKeys>( ContentType type, StructAutoKeys &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "version", data.version, errorPresent );

            if ( !errorPresent )
            {
                // Update parent node for sub-structure.
                auto oldParent = setParentNode( subNode );

                tinyxml2::XMLElement *subItem{};
                for ( size_t i = 0; i < sizeof( data.key ) / sizeof( data.key[0] ) && !errorPresent; ++i )
                {
                    errorPresent |= !readContent( ContentType::AutoKey, data.key[i], subItem );
                    if ( subItem == nullptr )
                    {
                        // No further element was found.
                        errorPresent = true;
                        break;
                    }
                }

                // Reset parent node because the end of the structure is reached.
                setParentNode( oldParent );
            }
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructFlashHeader>( ContentType type, const StructFlashHeader &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "version", Utils::toHexString( data.version, true ) );
        writeValue( subNode, "config", Utils::toHexString( data.config, true ) );
        writeValue( subNode, "headline", (const char *)(data.headline) );
        writeValue( subNode, "listStruct", data.listStruct );

        auto subNode2 = m_xmlDoc->NewElement( "lists" );
        subNode->LinkEndChild( subNode2 );

        for ( size_t i = 0; i < sizeof( data.list ) / sizeof( data.list[0] ); ++i )
        {
            writeValue( subNode2, "list", Utils::toHexString( data.list[i], true ) );
        }

        writeValue( subNode, "dataStruct", Utils::toHexString( data.dataStruct, true ) );
        writeValue( subNode, "flashSizeSUS", Utils::toHexString( data.flashSizeSUS, true ) );
        writeValue( subNode, "inputChainSizeMDE", Utils::toHexString( data.inputChainSizeMDE, true ) );
        writeValue( subNode, "inputChainSizeEntrance", Utils::toHexString( data.inputChainSizeEntrance, true ) );

        subNode2 = m_xmlDoc->NewElement( "listKeyFields" );
        subNode->LinkEndChild( subNode2 );

        for ( size_t i = 0; i < sizeof( data.listKeyField ) / sizeof( data.listKeyField[0] ); ++i )
        {
            writeValue( subNode2, "listKeyField", data.listKeyField[i] );
        }

        // Update parent node for sub-structure.
        auto oldParent = setParentNode( subNode );

        writeContent( ContentType::AutoKeys, data.autoKeys );

        // Reset parent node because the end of the structure is reached.
        setParentNode( oldParent );

        writeValue( subNode, "studioVersion", Utils::toHexString( data.studioVersion, true ) );
        writeValue( subNode, "setupType", static_cast<uint8_t>(data.setupType) );
        writeValue( subNode, "listAccess", Utils::toHexString( data.listAccess, true ) );
        writeValue( subNode, "inputChainAccess", Utils::toHexString( data.inputChainAccess, true ) );
        writeValue( subNode, "fileVersion", Utils::toHexString( data.fileVersion, true ) );
        writeValue( subNode, "headerSize", Utils::toHexString( data.headerSize, true ) );
        writeValue( subNode, "extendedSize", Utils::toHexString( data.extendedSize, true ) );
        writeValue( subNode, "libraryVersion", Utils::toHexString( data.libraryVersion, true ) );
    }
}
template<>
bool FileXML::readContent<StructFlashHeader>( ContentType type, StructFlashHeader &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            std::string dummy;
            data.version = Utils::fromHexString_uint( readValue( subNode, "version", dummy, errorPresent ), uint16_t() );
            data.config = Utils::fromHexString_uint( readValue( subNode, "config", dummy, errorPresent ), uint32_t() );
            readValue( subNode, "headline", data.headline, sizeof( data.headline ) / sizeof( data.headline[0] ), errorPresent, true );
            readValue( subNode, "listStruct", data.listStruct, errorPresent );

            // !!! Anpassung an readValue mit Schleife nötig !!!
            if ( !errorPresent )
            {
                auto subNode2 = subNode->FirstChildElement( "lists" );
                errorPresent = (subNode2 == nullptr);
                if ( !errorPresent )
                {
                    auto item = subNode2->FirstChildElement( "list" );
                    for ( size_t i = 0; i < sizeof( data.list ) / sizeof( data.list[0] ); ++i )
                    {
                        if ( i > 0 )
                        {
                            item = item->NextSiblingElement( "list" );
                        }

                        if ( item != nullptr )
                        {
                            const char *text = item->GetText();
                            if ( text != nullptr )
                            {
                                data.list[i] = Utils::fromHexString_uint( Utils::from_utf8( text ), uint32_t() );
                            }
                        }
                        else
                        {
                            errorPresent = true;
                        }
                    }
                }
            }

            data.dataStruct = Utils::fromHexString_uint( readValue( subNode, "dataStruct", dummy, errorPresent ), uint32_t() );
            data.flashSizeSUS = Utils::fromHexString_uint( readValue( subNode, "flashSizeSUS", dummy, errorPresent ), uint32_t() );
            data.inputChainSizeMDE = Utils::fromHexString_uint( readValue( subNode, "inputChainSizeMDE", dummy, errorPresent ), uint32_t() );
            data.inputChainSizeEntrance = Utils::fromHexString_uint( readValue( subNode, "inputChainSizeEntrance", dummy, errorPresent ), uint32_t() );

            if ( !errorPresent )
            {
                auto subNode2 = subNode->FirstChildElement( "listKeyFields" );
                errorPresent = (subNode2 == nullptr);
                if ( !errorPresent )
                {
                    auto item = subNode2->FirstChildElement( "listKeyField" );
                    for ( size_t i = 0; i < sizeof( data.listKeyField ) / sizeof( data.listKeyField[0] ) && (item != nullptr); ++i )
                    {
                        if ( i > 0 )
                        {
                            item = item->NextSiblingElement( "listKeyField" );
                        }

                        if ( item != nullptr )
                        {
                            data.listKeyField[i] = static_cast<uint8_t>(item->UnsignedText());
                            errorPresent = false;
                        }
                        else
                        {
                            errorPresent = true;
                        }
                    }
                }
            }

            if ( !errorPresent )
            {
                // Update parent node for sub-structure.
                auto oldParent = setParentNode( subNode );

                errorPresent |= !readContent( FileXML::ContentType::AutoKeys, data.autoKeys );

                // Reset parent node because the end of the structure is reached.
                setParentNode( oldParent );
            }

            data.studioVersion = Utils::fromHexString_uint( readValue( subNode, "studioVersion", dummy, errorPresent ), uint32_t() );
            data.setupType = static_cast<SetupType>(readValue( subNode, "setupType", (uint8_t &)(data.setupType), errorPresent ));
            data.listAccess = Utils::fromHexString_uint( readValue( subNode, "listAccess", dummy, errorPresent ), uint32_t() );
            data.inputChainAccess = Utils::fromHexString_uint( readValue( subNode, "inputChainAccess", dummy, errorPresent ), uint32_t() );
            data.fileVersion = Utils::fromHexString_uint( readValue( subNode, "inputChainAccess", dummy, errorPresent ), uint16_t() );
            data.headerSize = Utils::fromHexString_uint( readValue( subNode, "headerSize", dummy, errorPresent ), uint32_t() );
            data.extendedSize = Utils::fromHexString_uint( readValue( subNode, "extendedSize", dummy, errorPresent ), uint32_t() );
            data.libraryVersion = Utils::fromHexString_uint( readValue( subNode, "libraryVersion", dummy, errorPresent ), uint32_t() );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructGlobalVar>( ContentType type, const StructGlobalVar &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        constexpr size_t numberOfItems = sizeof( data.names ) / sizeof( data.names[0] );
        for ( size_t i = 0; i < numberOfItems; ++i )
        {
            writeValue( subNode, "names", data.names[i] );
        }
    }
}
template<>
bool FileXML::readContent<StructGlobalVar>( ContentType type, StructGlobalVar &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            tinyxml2::XMLElement *subItem{};
            constexpr int numberOfItems = sizeof( data.names ) / sizeof( data.names[0] );
            constexpr int stringSize = sizeof( data.names[0] ) / sizeof( data.names[0][0] );
            for ( size_t i = 0; i < numberOfItems && !errorPresent; ++i )
            {
                readValue( subNode, subItem, "names", data.names[i], stringSize, errorPresent, true );
                if ( subItem == nullptr )
                {
                    // No further element was found.
                    errorPresent = true;
                    break;
                }
            }
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructConfigRoot>( ContentType type, const StructConfigRoot &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "autoSwitchOff", data.autoSwitchOff );
        writeValue( subNode, "numberOfEntries", data.numberOfEntries );
        writeValue( subNode, "behaviourAfterSwitchOn", data.behaviourAfterSwitchOn );
    }
}
template<>
bool FileXML::readContent<StructConfigRoot>( ContentType type, StructConfigRoot &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "autoSwitchOff", data.autoSwitchOff, errorPresent );
            readValue( subNode, "numberOfEntries", data.numberOfEntries, errorPresent );
            readValue( subNode, "behaviourAfterSwitchOn", data.behaviourAfterSwitchOn, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructOperating>( ContentType type, const StructOperating &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "type", data.type );
    }
}
template<>
bool FileXML::readContent<StructOperating>( ContentType type, StructOperating &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "type", data.type, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructMenu>( ContentType type, const StructMenu &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "flags", data.flags );
        writeValue( subNode, "numberOfEntries", data.numberOfEntries );
        writeValue( subNode, "heading1", data.heading1 );
        writeValue( subNode, "heading2", data.heading2 );
        writeValue( subNode, "header1Text", data.header1Text );
        writeValue( subNode, "header2Text", data.header2Text );
        writeValue( subNode, "globVar", data.globVar );
        writeValue( subNode, "behaviourAtEsc", data.behaviourAtEsc );
    }
}
template<>
bool FileXML::readContent<StructMenu>( ContentType type, StructMenu &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "flags", data.flags, errorPresent );
            readValue( subNode, "numberOfEntries", data.numberOfEntries, errorPresent );
            readValue( subNode, "heading1", data.heading1, sizeof( data.heading1 ) / sizeof( data.heading1[0] ), errorPresent, true );
            readValue( subNode, "heading2", data.heading2, sizeof( data.heading2 ) / sizeof( data.heading2[0] ), errorPresent, true );
            readValue( subNode, "header1Text", data.header1Text, errorPresent );
            readValue( subNode, "header2Text", data.header2Text, errorPresent );
            readValue( subNode, "globVar", data.globVar, errorPresent );
            readValue( subNode, "behaviourAtEsc", data.behaviourAtEsc, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructMenuItem>( ContentType type, const StructMenuItem &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "name", data.name );
        writeValue( subNode, "numberOfEntries", data.numberOfEntries );
        writeValue( subNode, "password", data.password );
        writeValue( subNode, "flags", data.flags );
        writeValue( subNode, "machine", data.machine );
    }
}
template<>
bool FileXML::readContent<StructMenuItem>( ContentType type, StructMenuItem &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "name", data.name, sizeof( data.name ) / sizeof( data.name[0] ), errorPresent, true );
            readValue( subNode, "numberOfEntries", data.numberOfEntries, errorPresent );
            readValue( subNode, "password", data.password, errorPresent );
            readValue( subNode, "flags", data.flags, errorPresent );
            readValue( subNode, "machine", data.machine, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructReadRecord>( ContentType type, const StructReadRecord &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "number", data.number );
        writeValue( subNode, "fields", data.fields );
        writeValue( subNode, "behaviourAtCR", data.behaviourAtCR );
        writeValue( subNode, "name", data.name );
        writeValue( subNode, "type", data.type );
        writeValue( subNode, "flags", data.flags );
        writeValue( subNode, "header1Text", data.header1Text );
        writeValue( subNode, "behaviourAtESC", data.behaviourAtESC );
        writeValue( subNode, "vehicleRelease", data.vehicleRelease );
        writeValue( subNode, "relayNum", data.relayNum );
        writeValue( subNode, "relayTime", data.relayTime );
    }
}
template<>
bool FileXML::readContent<StructReadRecord>( ContentType type, StructReadRecord &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "number", data.number, errorPresent );
            readValue( subNode, "fields", data.fields, errorPresent );
            readValue( subNode, "behaviourAtCR", data.behaviourAtCR, errorPresent );
            readValue( subNode, "name", data.name, sizeof( data.name ) / sizeof( data.name[0] ), errorPresent, true );
            readValue( subNode, "type", data.type, errorPresent );
            readValue( subNode, "flags", data.flags, errorPresent );
            readValue( subNode, "header1Text", data.header1Text, errorPresent );
            readValue( subNode, "behaviourAtESC", data.behaviourAtESC, errorPresent );
            readValue( subNode, "vehicleRelease", data.vehicleRelease, errorPresent );
            readValue( subNode, "relayNum", data.relayNum, errorPresent );
            readValue( subNode, "relayTime", data.relayTime, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructTransponder>( ContentType type, const StructTransponder &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "tpSegment1", data.tpSegment1 );
        writeValue( subNode, "tpMode1", data.tpMode1 );
        writeValue( subNode, "tpSegment2", data.tpSegment2 );
        writeValue( subNode, "tpMode2", data.tpMode2 );
        writeValue( subNode, "tpSegment3", data.tpSegment3 );
        writeValue( subNode, "tpMode3", data.tpMode3 );
    }
}
template<>
bool FileXML::readContent<StructTransponder>( ContentType type, StructTransponder &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "tpSegment1", data.tpSegment1, errorPresent );
            readValue( subNode, "tpMode1", data.tpMode1, errorPresent );
            readValue( subNode, "tpSegment2", data.tpSegment2, errorPresent );
            readValue( subNode, "tpMode2", data.tpMode2, errorPresent );
            readValue( subNode, "tpSegment3", data.tpSegment3, errorPresent );
            readValue( subNode, "tpMode3", data.tpMode3, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructCompare>( ContentType type, const StructCompare &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "type", data.type );
        writeValue( subNode, "formatGlobalVar", data.formatGlobalVar );
        writeValue( subNode, "formatString", data.formatString );
        writeValue( subNode, "negate", data.negate );
        writeValue( subNode, "jumpDestination", data.jumpDestination );
    }
}
template<>
bool FileXML::readContent<StructCompare>( ContentType type, StructCompare &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "type", data.type, errorPresent );
            readValue( subNode, "formatGlobalVar", data.formatGlobalVar, errorPresent );
            readValue( subNode, "formatString", data.formatString, sizeof( data.formatString ) / sizeof( data.formatString[0] ), errorPresent, true );
            readValue( subNode, "negate", data.negate, errorPresent );
            readValue( subNode, "jumpDestination", data.jumpDestination, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructNumFieldEx>( ContentType type, const StructNumFieldEx &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "flags", data.flags );
        writeValue( subNode, "param1", data.param1 );
        writeValue( subNode, "param2", data.param2 );
        writeValue( subNode, "param3", data.param3 );
        writeValue( subNode, "relayNumber", data.relayNumber );
        writeValue( subNode, "relayAction", data.relayAction );
        writeValue( subNode, "relayTime", data.relayTime );
        writeValue( subNode, "contentOfConstant", data.contentOfConstant );
        // Update parent node for sub-structure.
        auto oldParent = setParentNode( subNode );

        writeContent( ContentType::Transponder, data.tag );
        writeContent( ContentType::Compare, data.compare );

        // Reset parent node because the end of the structure is reached.
        setParentNode( oldParent );
    }
}
template<>
bool FileXML::readContent<StructNumFieldEx>( ContentType type, StructNumFieldEx &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "flags", data.flags, errorPresent );
            readValue( subNode, "param1", data.param1, errorPresent );
            readValue( subNode, "param2", data.param2, errorPresent );
            readValue( subNode, "param3", data.param3, errorPresent );
            uint8_t tmp{};
            data.relayNumber = readValue( subNode, "relayNumber", tmp, errorPresent ) & 0x0F;
            data.relayAction = readValue( subNode, "relayAction", tmp, errorPresent ) & 0x0F;
            readValue( subNode, "relayTime", data.relayTime, errorPresent );
            readValue( subNode, "contentOfConstant", data.contentOfConstant, sizeof( data.contentOfConstant ) / sizeof( data.contentOfConstant[0] ), errorPresent, true );

            // Update parent node for sub-structure.
            auto oldParent = setParentNode( subNode );

            errorPresent |= !readContent( ContentType::Transponder, data.tag );
            errorPresent |= !readContent( ContentType::Compare, data.compare );

            // Reset parent node because the end of the structure is reached.
            setParentNode( oldParent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructListSelectionEx>( ContentType type, const StructListSelectionEx &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "selection", data.selection );
        writeValue( subNode, "selectionRecordField", data.selectionRecordField );
        writeValue( subNode, "selectionListField", data.selectionListField );
        writeValue( subNode, "listNumber", data.listNumber );
        writeValue( subNode, "selectionField", data.selectionField );
        writeValue( subNode, "dataField", data.dataField );
        writeValue( subNode, "selectionBarcode", data.selectionBarcode );
        writeValue( subNode, "flags", data.flags );
        writeValue( subNode, "noValueCheck", data.noValueCheck );
        writeValue( subNode, "jumpSelectionNoEntry", data.jumpSelectionNoEntry );
        writeValue( subNode, "inputModuleTimeout", data.inputModuleTimeout );
        writeValue( subNode, "transponderConfig", data.transponderConfig );
        writeValue( subNode, "doubleColumn", data.doubleColumn );
        writeValue( subNode, "selectionRecordField2", data.selectionRecordField2 );
        writeValue( subNode, "selectionListField2", data.selectionListField2 );
        writeValue( subNode, "selectionNoEntry", data.selectionNoEntry );
        writeValue( subNode, "usedFontSize", data.usedFontSize );

        // Update parent node for sub-structure.
        auto oldParent = setParentNode( subNode );

        writeContent( ContentType::Compare, data.compare );

        // Reset parent node because the end of the structure is reached.
        setParentNode( oldParent );
    }
}
template<>
bool FileXML::readContent<StructListSelectionEx>( ContentType type, StructListSelectionEx &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "selection", data.selection, errorPresent );
            readValue( subNode, "selectionRecordField", data.selectionRecordField, errorPresent );
            readValue( subNode, "selectionListField", data.selectionListField, errorPresent );
            readValue( subNode, "listNumber", data.listNumber, errorPresent );
            readValue( subNode, "selectionField", data.selectionField, errorPresent );
            readValue( subNode, "dataField", data.dataField, errorPresent );
            readValue( subNode, "selectionBarcode", data.selectionBarcode, errorPresent );
            readValue( subNode, "flags", data.flags, errorPresent );
            readValue( subNode, "noValueCheck", data.noValueCheck, errorPresent );
            readValue( subNode, "jumpSelectionNoEntry", data.jumpSelectionNoEntry, errorPresent );
            readValue( subNode, "inputModuleTimeout", data.inputModuleTimeout, errorPresent );
            readValue( subNode, "transponderConfig", data.transponderConfig, errorPresent );
            readValue( subNode, "doubleColumn", data.doubleColumn, errorPresent );
            readValue( subNode, "selectionRecordField2", data.selectionRecordField2, errorPresent );
            readValue( subNode, "selectionListField2", data.selectionListField2, errorPresent );
            readValue( subNode, "selectionNoEntry", data.selectionNoEntry, errorPresent );
            readValue( subNode, "usedFontSize", data.usedFontSize, errorPresent );

            // Update parent node for sub-structure.
            auto oldParent = setParentNode( subNode );

            errorPresent |= !readContent( ContentType::Compare, data.compare );

            // Reset parent node because the end of the structure is reached.
            setParentNode( oldParent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructConfirmationEx>( ContentType type, const StructConfirmationEx &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "text1", data.text1 );
        writeValue( subNode, "text2", data.text2 );
        writeValue( subNode, "text3", data.text3 );
    }
}
template<>
bool FileXML::readContent<StructConfirmationEx>( ContentType type, StructConfirmationEx &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "text1", data.text1, sizeof( data.text1 ) / sizeof( data.text1[0] ), errorPresent, true );
            readValue( subNode, "text2", data.text2, sizeof( data.text2 ) / sizeof( data.text2[0] ), errorPresent, true );
            readValue( subNode, "text3", data.text3, sizeof( data.text3 ) / sizeof( data.text3[0] ), errorPresent, true );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructTextEx>( ContentType type, const StructTextEx &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "text", data.text );
    }
}
template<>
bool FileXML::readContent<StructTextEx>( ContentType type, StructTextEx &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "text", data.text, sizeof( data.text ) / sizeof( data.text[0] ), errorPresent, true );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructDataEx>( ContentType type, const StructDataEx &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "type", data.type );
        writeValue( subNode, "flags", data.flags );
        writeValue( subNode, "time", data.time );
        writeValue( subNode, "redOK", data.redOK );
        writeValue( subNode, "greenOK", data.greenOK );
        writeValue( subNode, "blueOK", data.blueOK );
        writeValue( subNode, "whiteOK", data.whiteOK );
        writeValue( subNode, "redESC", data.redESC );
        writeValue( subNode, "greenESC", data.greenESC );
        writeValue( subNode, "blueESC", data.blueESC );
        writeValue( subNode, "whiteESC", data.whiteESC );
    }
}
template<>
bool FileXML::readContent<StructDataEx>( ContentType type, StructDataEx &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "type", data.type, errorPresent );
            readValue( subNode, "flags", data.flags, errorPresent );
            readValue( subNode, "time", data.time, errorPresent );
            readValue( subNode, "redOK", data.redOK, errorPresent );
            readValue( subNode, "greenOK", data.greenOK, errorPresent );
            readValue( subNode, "blueOK", data.blueOK, errorPresent );
            readValue( subNode, "whiteOK", data.whiteOK, errorPresent );
            readValue( subNode, "redESC", data.redESC, errorPresent );
            readValue( subNode, "greenESC", data.greenESC, errorPresent );
            readValue( subNode, "blueESC", data.blueESC, errorPresent );
            readValue( subNode, "whiteESC", data.whiteESC, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructGlobVarToHIDEx>( ContentType type, const StructGlobVarToHIDEx &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "nullByte", data.nullByte );
        writeValue( subNode, "ending", data.ending );
        writeValue( subNode, "interfacePort", data.interfacePort );
    }
}
template<>
bool FileXML::readContent<StructGlobVarToHIDEx>( ContentType type, StructGlobVarToHIDEx &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "nullByte", data.nullByte, errorPresent );
            readValue( subNode, "ending", data.ending, errorPresent );
            readValue( subNode, "interfacePort", data.interfacePort, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructSwitchModuleEnergieEx>( ContentType type, const StructSwitchModuleEnergieEx &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "module", data.module );
        writeValue( subNode, "item", data.item );
        writeValue( subNode, "time", data.time );
        writeValue( subNode, "value", data.value );
        writeValue( subNode, "interval", data.interval );
    }
}
template<>
bool FileXML::readContent<StructSwitchModuleEnergieEx>( ContentType type, StructSwitchModuleEnergieEx &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "module", data.module, errorPresent );
            readValue( subNode, "item", data.item, errorPresent );
            readValue( subNode, "time", data.time, errorPresent );
            readValue( subNode, "value", data.value, errorPresent );
            readValue( subNode, "interval", data.interval, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructSystemMessageEx>( ContentType type, const StructSystemMessageEx &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "value", data.value );
    }
}
template<>
bool FileXML::readContent<StructSystemMessageEx>( ContentType type, StructSystemMessageEx &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "value", data.value, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructTakePhotoEx>( ContentType type, const StructTakePhotoEx &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "flags", data.flags );
        writeValue( subNode, "showTime", data.showTime );
        writeValue( subNode, "showPeep", data.showPeep );
        writeValue( subNode, "text", data.text );
    }
}
template<>
bool FileXML::readContent<StructTakePhotoEx>( ContentType type, StructTakePhotoEx &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "flags", data.flags, errorPresent );
            readValue( subNode, "showTime", data.showTime, errorPresent );
            readValue( subNode, "showPeep", data.showPeep, errorPresent );
            readValue( subNode, "text", data.text, sizeof( data.text ) / sizeof( data.text[0] ), errorPresent, true );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructFeverControlEx>( ContentType type, const StructFeverControlEx &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );
        
        writeValue( subNode, "flagUnitFahrenheit", static_cast<bool>( data.flagUnitFahrenheit ) );
        writeValue( subNode, "flagShowTemperaturValue", static_cast<bool>( data.flagShowTemperaturValue ) );
        writeValue( subNode, "flagValidMeasureRepeatable", static_cast<bool>( data.flagValidMeasureRepeatable ) );
        writeValue( subNode, "flagInvalidMeasureRepeatable", static_cast<bool>( data.flagInvalidMeasureRepeatable ) );
        writeValue( subNode, "measureMin", data.measureMin );
        writeValue( subNode, "measureMax", data.measureMax );
        writeValue( subNode, "measureFever", data.measureFever );
        writeValue( subNode, "indexGlobVar", data.indexGlobVar );
        writeValue( subNode, "indexRecordField", data.indexRecordField );
        
    }
}
template<>
bool FileXML::readContent<StructFeverControlEx>( ContentType type, StructFeverControlEx &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            bool bitfield;
            readValue( subNode, "flagUnitFahrenheit", bitfield, errorPresent ); data.flagUnitFahrenheit = bitfield;
            readValue( subNode, "flagShowTemperaturValue", bitfield, errorPresent ); data.flagShowTemperaturValue = bitfield;
            readValue( subNode, "flagValidMeasureRepeatable", bitfield, errorPresent ); data.flagValidMeasureRepeatable = bitfield;
            readValue( subNode, "flagInvalidMeasureRepeatable", bitfield, errorPresent ); data.flagInvalidMeasureRepeatable = bitfield;
            readValue( subNode, "measureMin", data.measureMin, errorPresent );
            readValue( subNode, "measureMax", data.measureMax, errorPresent );
            readValue( subNode, "measureFever", data.measureFever, errorPresent );
            readValue( subNode, "indexGlobVar", data.indexGlobVar, errorPresent );
            readValue( subNode, "indexRecordField", data.indexRecordField, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructMultiInputSourceEx>( ContentType type, const StructMultiInputSourceEx &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "childCount", data.childCount );
        writeValue( subNode, "keypadInput", static_cast<bool>( data.keypadInput ) );
        writeValue( subNode, "barcodeInput", static_cast<bool>( data.barcodeInput ) );
        writeValue( subNode, "transponderInput", static_cast<bool>( data.transponderInput ) );
        writeValue( subNode, "fingerprintIdentificationInput", static_cast<bool>( data.fingerprintIdentificationInput ) );
        writeValue( subNode, "fingerprintVerificationInput", static_cast<bool>( data.fingerprintVerificationInput ) );
        writeValue( subNode, "indexGlobaleVariable", data.indexGlobaleVariable );
        writeValue( subNode, "indexRecordField", data.indexRecordField );

        // Update parent node for sub-structure.
        auto oldParent = setParentNode( subNode );

        writeContent( ContentType::Compare, data.compare );

        // Reset parent node because the end of the structure is reached.
        setParentNode( oldParent );
    }
}
template<>
bool FileXML::readContent<StructMultiInputSourceEx>( ContentType type, StructMultiInputSourceEx &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            bool bitfield;
            readValue( subNode, "childCount", data.childCount, errorPresent );
            readValue( subNode, "keypadInput", bitfield, errorPresent ); data.keypadInput = bitfield;
            readValue( subNode, "barcodeInput", bitfield, errorPresent ); data.barcodeInput = bitfield;
            readValue( subNode, "transponderInput", bitfield, errorPresent ); data.transponderInput = bitfield;
            readValue( subNode, "fingerprintIdentificationInput", bitfield, errorPresent ); data.fingerprintIdentificationInput = bitfield;
            readValue( subNode, "fingerprintVerificationInput", bitfield, errorPresent ); data.fingerprintVerificationInput = bitfield;
            readValue( subNode, "indexGlobaleVariable", data.indexGlobaleVariable, errorPresent );
            readValue( subNode, "indexRecordField", data.indexRecordField, errorPresent );

            // Update parent node for sub-structure.
            auto oldParent = setParentNode( subNode );

            errorPresent |= !readContent( ContentType::Compare, data.compare );

            // Reset parent node because the end of the structure is reached.
            setParentNode( oldParent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructReadField>( ContentType type, const StructReadField &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "type", data.type );
        writeValue( subNode, "recordFieldNumber", data.recordFieldNumber );
        writeValue( subNode, "behaviourAtEsc", data.behaviourAtEsc );
        writeValue( subNode, "name", data.name );
        writeValue( subNode, "header2Text", data.header2Text );
        writeValue( subNode, "globalVariable", data.globalVariable );
        writeValue( subNode, "systemVariable", data.systemVariable );
        writeValue( subNode, "barcodeRangeFrom", data.barcodeRangeFrom );
        writeValue( subNode, "barcodeRangeUpTo", data.barcodeRangeUpTo );
        writeValue( subNode, "inputTypeCount", data.inputTypeCount );
        writeValue( subNode, "variable1", data.variable1 );
        writeValue( subNode, "variable2", data.variable2 );

        // Update parent node for sub-structure.
        auto oldParent = setParentNode( subNode );

        switch ( data.type )
        {
            case TBFE_NORM:
            case TBFE_GET_TIME:
            case TBFE_KONSTANTE:
            case TBFE_COPY_GV_TO_FIELD:
            case TBFE_TIMEBOY_COPY_GV_TO_FIELD:
            case TBFE_SERIENNUMMER:
            case TBFE_ANALOG_WERT:
            case TBFE_ANALOG_STATUS:
            case TBFE_DIGITAL_WERT:
            case TBFE_ZAEHLER_WERT:
            case TBFE_FUNC:
            case TBFE_SPS_WERT:
            case TBFE_ENTRANCE_ZM_NUM:
            case TBFE_ENTRANCE_READER_NUM:
            case TBFE_ENTRANCE_PNR:
            case TBFE_ENTRANCE_STATUS:
            case TBFE_DEBUG:
            case TBFE_GPRS_ALIVE_COUNTER:
            case TBFE_FIRMWARE_VERSION:
            case TBFE_ONLINE_STATUS:
            case TBFE_SUMMER_WINTER_STATUS:
            case TBFE_GPS_RECORD_FIX:
            case TBFE_GPS_RECORD_VARIABLE:
            case TBFE_ENTRANCE2_IDENTIFICATION:
            case TBFE_COPY_FIELD_TO_GV:
            case TBFE_TIMEBOY_COPY_FIELD_TO_GV:
            case TBFE_FINGERPRINT_SCANN:
            case TBFE_FINGERPRINT_CARD:
            case TBFE_FINGERPRINT_TEACH:
            case TBFE_FINGERPRINT_IDENTIFICATION:
            case TBFE_FINGERPRINT_VERIFICATION:
            case TBFE_FINGERPRINT_DELETE:
            case TBFE_OPERATION:
            case TBFE_SWITCH_RELAY:
            case TBFE_SWITCH_LED:
            case TBFE_TIMER:
            case TBFE_TRANSPONDER_WRITE:
            case TBFE_GPS_DISTANCE:
            case TBFE_MOBILE_PARAMETER:
            case TBFE_SWITCH_BUZZER:
            case TBFE_KONSTANTE_TYPE_DEVICE_NAME:
            case TBFE_MOBILMASTERIV_MENUENTRY:
            case TBFE_SWITCH_COMMUNICATION:
            case TBFE_FINGERPRINT_ADMNISTARTION:
            case TBFE_TIMER_STATE:
            case TBFE_VIBRATION_VALUE:
            {
                writeContent( ContentType::ReadField_NumFieldEx, data.ex.numField );
                break;
            }

            case TBFE_LIST_SELECT_GET:
            case TBFE_LIST_SELECT_SET:
            case TBFE_DATA_ON_CARD:
            {
                writeContent( ContentType::ReadField_ListSelectionEx, data.ex.listSelection );
                break;
            }

            case TBFE_BESTAETIGUNG:
            {
                writeContent( ContentType::ReadField_ConfirmationEx, data.ex.confirmation );
                break;
            }

            case TBFE_SEND_SMS:
            case TBFE_SCRIPT:
            case TBFE_SHOW_MESSAGE:
            case TBFE_READ_VARIABLE:
            case TBFE_WRITE_VARIABLE:
            case TBFE_SIGNATURE:
            case TBFE_LONG_BARCODE:
            {
                writeContent( ContentType::ReadField_TextEx, data.ex.text );
                break;
            }

            case TBFE_SWITCH_BACKLIGHT:
            {
                writeContent( ContentType::ReadField_DataEx, data.ex.data );
                break;
            }

            case TBFE_GV_TO_INTERFACE:
            {
                writeContent( ContentType::ReadField_GlobVarToHIDEx, data.ex.globVarToHID );
                break;
            }

            case TBFE_SWITCH_MODULE_ENERGIE:
            {
                writeContent( ContentType::ReadField_SwitchModuleEnergieEx, data.ex.switchModuleEnergie );
                break;
            }

            case TBFE_SYSTEM_MESSAGE:
            {
                writeContent( ContentType::ReadField_SystemMessageEx, data.ex.systemMessage );
                break;
            }

            case TBFE_TAKE_PHOTO:
            {
                writeContent( ContentType::ReadField_TakePhotoEx, data.ex.takePhoto );
                break;
            }

            case TBFE_FEVER_CONTROL:
            {
                writeContent( ContentType::ReadField_FeverControl, data.ex.feverControl );
                break;
            }

            case TBFE_MULTI_INPUT_SOURCE:
            {
                writeContent( ContentType::ReadField_MultiInputSource, data.ex.multiInputSource );
                break;
            }
        }

        // Reset parent node because the end of the structure is reached.
        setParentNode( oldParent );
    }
}
template<>
bool FileXML::readContent<StructReadField>( ContentType type, StructReadField &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "type", data.type, errorPresent );
            readValue( subNode, "recordFieldNumber", data.recordFieldNumber, errorPresent );
            readValue( subNode, "behaviourAtEsc", data.behaviourAtEsc, errorPresent );
            readValue( subNode, "name", data.name, sizeof( data.name ) / sizeof( data.name[0] ), errorPresent, true );
            readValue( subNode, "header2Text", data.header2Text, errorPresent );
            readValue( subNode, "globalVariable", data.globalVariable, errorPresent );
            readValue( subNode, "systemVariable", data.systemVariable, errorPresent );
            readValue( subNode, "barcodeRangeFrom", data.barcodeRangeFrom, errorPresent );
            readValue( subNode, "barcodeRangeUpTo", data.barcodeRangeUpTo, errorPresent );
            readValue( subNode, "inputTypeCount", data.inputTypeCount, errorPresent );
            readValue( subNode, "variable1", data.variable1, errorPresent );
            readValue( subNode, "variable2", data.variable2, errorPresent );

            // Update parent node for sub-structure.
            auto oldParent = setParentNode( subNode );

            switch ( data.type )
            {
                case TBFE_NORM:
                case TBFE_GET_TIME:
                case TBFE_KONSTANTE:
                case TBFE_COPY_GV_TO_FIELD:
                case TBFE_TIMEBOY_COPY_GV_TO_FIELD:
                case TBFE_SERIENNUMMER:
                case TBFE_ANALOG_WERT:
                case TBFE_ANALOG_STATUS:
                case TBFE_DIGITAL_WERT:
                case TBFE_ZAEHLER_WERT:
                case TBFE_FUNC:
                case TBFE_SPS_WERT:
                case TBFE_ENTRANCE_ZM_NUM:
                case TBFE_ENTRANCE_READER_NUM:
                case TBFE_ENTRANCE_PNR:
                case TBFE_ENTRANCE_STATUS:
                case TBFE_DEBUG:
                case TBFE_GPRS_ALIVE_COUNTER:
                case TBFE_FIRMWARE_VERSION:
                case TBFE_ONLINE_STATUS:
                case TBFE_SUMMER_WINTER_STATUS:
                case TBFE_GPS_RECORD_FIX:
                case TBFE_GPS_RECORD_VARIABLE:
                case TBFE_ENTRANCE2_IDENTIFICATION:
                case TBFE_COPY_FIELD_TO_GV:
                case TBFE_TIMEBOY_COPY_FIELD_TO_GV:
                case TBFE_FINGERPRINT_SCANN:
                case TBFE_FINGERPRINT_CARD:
                case TBFE_FINGERPRINT_TEACH:
                case TBFE_FINGERPRINT_IDENTIFICATION:
                case TBFE_FINGERPRINT_VERIFICATION:
                case TBFE_FINGERPRINT_DELETE:
                case TBFE_OPERATION:
                case TBFE_SWITCH_RELAY:
                case TBFE_SWITCH_LED:
                case TBFE_TIMER:
                case TBFE_TRANSPONDER_WRITE:
                case TBFE_GPS_DISTANCE:
                case TBFE_MOBILE_PARAMETER:
                case TBFE_SWITCH_BUZZER:
                case TBFE_KONSTANTE_TYPE_DEVICE_NAME:
                case TBFE_MOBILMASTERIV_MENUENTRY:
                case TBFE_SWITCH_COMMUNICATION:
                case TBFE_FINGERPRINT_ADMNISTARTION:
                case TBFE_TIMER_STATE:
                case TBFE_VIBRATION_VALUE:
                {
                    errorPresent |= !readContent( ContentType::ReadField_NumFieldEx, data.ex.numField );
                    break;
                }

                case TBFE_LIST_SELECT_GET:
                case TBFE_LIST_SELECT_SET:
                case TBFE_DATA_ON_CARD:
                {
                    errorPresent |= !readContent( ContentType::ReadField_ListSelectionEx, data.ex.listSelection );
                    break;
                }

                case TBFE_BESTAETIGUNG:
                {
                    errorPresent |= !readContent( ContentType::ReadField_ConfirmationEx, data.ex.confirmation );
                    break;
                }

                case TBFE_SEND_SMS:
                case TBFE_SCRIPT:
                case TBFE_SHOW_MESSAGE:
                case TBFE_READ_VARIABLE:
                case TBFE_WRITE_VARIABLE:
                case TBFE_SIGNATURE:
                case TBFE_LONG_BARCODE:
                {
                    errorPresent |= !readContent( ContentType::ReadField_TextEx, data.ex.text );
                    break;
                }

                case TBFE_SWITCH_BACKLIGHT:
                {
                    errorPresent |= !readContent( ContentType::ReadField_DataEx, data.ex.data );
                    break;
                }

                case TBFE_GV_TO_INTERFACE:
                {
                    errorPresent |= !readContent( ContentType::ReadField_GlobVarToHIDEx, data.ex.globVarToHID );
                    break;
                }

                case TBFE_SWITCH_MODULE_ENERGIE:
                {
                    errorPresent |= !readContent( ContentType::ReadField_SwitchModuleEnergieEx, data.ex.switchModuleEnergie );
                    break;
                }

                case TBFE_SYSTEM_MESSAGE:
                {
                    errorPresent |= !readContent( ContentType::ReadField_SystemMessageEx, data.ex.systemMessage );
                    break;
                }

                case TBFE_TAKE_PHOTO:
                {
                    errorPresent |= !readContent( ContentType::ReadField_TakePhotoEx, data.ex.takePhoto );
                    break;
                }

                case TBFE_FEVER_CONTROL:
                {
                    errorPresent |= !readContent( ContentType::ReadField_FeverControl, data.ex.feverControl );
                    break;
                }

                case TBFE_MULTI_INPUT_SOURCE:
                {
                    errorPresent |= !readContent( ContentType::ReadField_MultiInputSource, data.ex.multiInputSource );
                    break;
                }
            }

            // Reset parent node because the end of the structure is reached.
            setParentNode( oldParent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructRecordDescription>( ContentType type, const StructRecordDescription &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "fields", data.fields );
        writeValue( subNode, "name", data.name );
    }
}
template<>
bool FileXML::readContent<StructRecordDescription>( ContentType type, StructRecordDescription &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "fields", data.fields, errorPresent );
            readValue( subNode, "name", data.name, sizeof( data.name ) / sizeof( data.name[0] ), errorPresent, true );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructRecordField>( ContentType type, const StructRecordField &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "type", data.type );
        writeValue( subNode, "count", data.count );
        writeValue( subNode, "name", data.name );
    }
}
template<>
bool FileXML::readContent<StructRecordField>( ContentType type, StructRecordField &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "type", data.type, errorPresent );
            readValue( subNode, "count", data.count, errorPresent );
            readValue( subNode, "name", data.name, sizeof( data.name ) / sizeof( data.name[0] ), errorPresent, true );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructMachineDefinition>( ContentType type, const StructMachineDefinition &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "flags", data.flags );
        writeValue( subNode, "name", data.name );
        writeValue( subNode, "inputs", data.inputs );
    }
}
template<>
bool FileXML::readContent<StructMachineDefinition>( ContentType type, StructMachineDefinition &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "flags", data.flags, errorPresent );
            readValue( subNode, "name", data.name, sizeof( data.name ) / sizeof( data.name[0] ), errorPresent, true );
            readValue( subNode, "inputs", data.inputs, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructVibrationSensor>( ContentType type, const StructVibrationSensor &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "useAs", data.useAs );
        writeValue( subNode, "level", data.level );
        writeValue( subNode, "timeout", data.timeout );

        auto subNode2 = m_xmlDoc->NewElement( "invokeInputChain" );
        subNode->LinkEndChild( subNode2 );

        for ( size_t i = 0; i < sizeof( data.invokeInputChain ) / sizeof( data.invokeInputChain[0] ); ++i )
        {
            writeValue( subNode2, "inputChain", data.invokeInputChain[i] );
        }
    }
}
template<>
bool FileXML::readContent<StructVibrationSensor>( ContentType type, StructVibrationSensor &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "useAs", data.useAs, errorPresent );
            readValue( subNode, "level", data.level, errorPresent );
            readValue( subNode, "timeout", data.timeout, errorPresent );

            tinyxml2::XMLElement *subItem{};
            auto subNode2 = subNode->FirstChildElement( "invokeInputChain" );

            constexpr int numberOfItems = sizeof( data.invokeInputChain ) / sizeof( data.invokeInputChain[0] );
            for ( size_t i = 0; i < numberOfItems && !errorPresent; ++i )
            {
                readValue( subNode2, subItem, "inputChain", data.invokeInputChain[i], errorPresent );
            }
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructDigitalInput>( ContentType type, const StructDigitalInput &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "useAs", data.useAs );
        writeValue( subNode, "startLowHigh", data.startLowHigh );
        writeValue( subNode, "startHighLow", data.startHighLow );
        writeValue( subNode, "countDivisor", data.countDivisor );
        writeValue( subNode, "cycle", data.cycle );
        writeValue( subNode, "timeout", data.timeout );
        writeValue( subNode, "countDivisor32Bit", data.countDivisor32Bit );
        writeValue( subNode, "debounceHigh", data.debounceHigh );
        writeValue( subNode, "debounceLow", data.debounceLow );
        writeValue( subNode, "invokeInputChainCountDivisor", data.invokeInputChainCountDivisor );
        writeValue( subNode, "invokeInputChainCycle", data.invokeInputChainCycle );
    }
}
template<>
bool FileXML::readContent<StructDigitalInput>( ContentType type, StructDigitalInput &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "useAs", data.useAs, errorPresent );
            readValue( subNode, "startLowHigh", data.startLowHigh, errorPresent );
            readValue( subNode, "startHighLow", data.startHighLow, errorPresent );
            readValue( subNode, "countDivisor", data.countDivisor, errorPresent );
            readValue( subNode, "cycle", data.cycle, errorPresent );
            readValue( subNode, "timeout", data.timeout, errorPresent );
            readValue( subNode, "countDivisor32Bit", data.countDivisor32Bit, errorPresent );
            readValue( subNode, "debounceHigh", data.debounceHigh, errorPresent );
            readValue( subNode, "debounceLow", data.debounceLow, errorPresent );
            readValue( subNode, "invokeInputChainCountDivisor", data.invokeInputChainCountDivisor, errorPresent );
            readValue( subNode, "invokeInputChainCycle", data.invokeInputChainCycle, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructDigitalInputExt>( ContentType type, const StructDigitalInputExt &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "cavity", data.cavity );
        writeValue( subNode, "strokes", data.strokes );
    }
}
template<>
bool FileXML::readContent<StructDigitalInputExt>( ContentType type, StructDigitalInputExt &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "cavity", data.cavity, errorPresent );
            readValue( subNode, "strokes", data.strokes, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructAnalogInput>( ContentType type, const StructAnalogInput &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "useAs", data.useAs );
        writeValue( subNode, "show", data.show );
        writeValue( subNode, "sampleRate", data.sampleRate );
        writeValue( subNode, "name", data.name );
        writeValue( subNode, "unit", data.unit );
        writeValue( subNode, "minLimit", data.minLimit );
        writeValue( subNode, "maxLimit", data.maxLimit );
        writeValue( subNode, "minInput", data.minInput );
        writeValue( subNode, "maxInput", data.maxInput );
        writeValue( subNode, "limitUpperAlarm", data.limitUpperAlarm );
        writeValue( subNode, "limitUpperWarning", data.limitUpperWarning );
        writeValue( subNode, "limitLowerWarning", data.limitLowerWarning );
        writeValue( subNode, "limitLowerAlarm", data.limitLowerAlarm );
        writeValue( subNode, "limitTolerance", data.limitTolerance );

        auto subNode2 = m_xmlDoc->NewElement( "invokeInputChain" );
        subNode->LinkEndChild( subNode2 );

        for ( size_t i = 0; i < sizeof( data.invokeInputChain ) / sizeof( data.invokeInputChain[0] ); ++i )
        {
            writeValue( subNode2, "inputChain", data.invokeInputChain[i] );
        }
    }
}
template<>
bool FileXML::readContent<StructAnalogInput>( ContentType type, StructAnalogInput &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "useAs", data.useAs, errorPresent );
            readValue( subNode, "show", data.show, errorPresent );
            readValue( subNode, "sampleRate", data.sampleRate, errorPresent );
            readValue( subNode, "name", data.name, sizeof( data.name ) / sizeof( data.name[0] ), errorPresent, true );
            readValue( subNode, "unit", data.unit, sizeof( data.unit ) / sizeof( data.unit[0] ), errorPresent, true );
            readValue( subNode, "minLimit", data.minLimit, errorPresent );
            readValue( subNode, "maxLimit", data.maxLimit, errorPresent );
            readValue( subNode, "minInput", data.minInput, errorPresent );
            readValue( subNode, "maxInput", data.maxInput, errorPresent );
            readValue( subNode, "limitUpperAlarm", data.limitUpperAlarm, errorPresent );
            readValue( subNode, "limitUpperWarning", data.limitUpperWarning, errorPresent );
            readValue( subNode, "limitLowerWarning", data.limitLowerWarning, errorPresent );
            readValue( subNode, "limitLowerAlarm", data.limitLowerAlarm, errorPresent );
            readValue( subNode, "limitTolerance", data.limitTolerance, errorPresent );

            tinyxml2::XMLElement *subItem{};
            auto subNode2 = subNode->FirstChildElement( "invokeInputChain" );

            constexpr int numberOfItems = sizeof( data.invokeInputChain ) / sizeof( data.invokeInputChain[0] );
            for ( size_t i = 0; i < numberOfItems && !errorPresent; ++i )
            {
                readValue( subNode2, subItem, "inputChain", data.invokeInputChain[i], errorPresent );
            }
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructTimer>( ContentType type, const StructTimer &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "useAs", data.useAs );
        writeValue( subNode, "cycle", data.cycle );
        writeValue( subNode, "timeout", data.timeout );
        writeValue( subNode, "invokeInputChain", data.invokeInputChain );
    }
}
template<>
bool FileXML::readContent<StructTimer>( ContentType type, StructTimer &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "useAs", data.useAs, errorPresent );
            readValue( subNode, "cycle", data.cycle, errorPresent );
            readValue( subNode, "timeout", data.timeout, errorPresent );
            readValue( subNode, "invokeInputChain", data.invokeInputChain, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructIOModule>( ContentType type, const StructIOModule &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        auto subNode2 = m_xmlDoc->NewElement( "voltageInputs" );
        subNode->LinkEndChild( subNode2 );

        writeValue( subNode2, "voltageInputsActive", data.voltageInputsActive );
        writeValue( subNode2, "voltageInputsShowOnScreen", data.voltageInputsShowOnScreen );

        constexpr size_t countVoltage = sizeof( data.voltageInputsMinValue ) / sizeof( data.voltageInputsMinValue[0] );
        for ( size_t i = 0; i < countVoltage; ++i )
        {
            auto subNode3 = m_xmlDoc->NewElement( "input" );
            subNode2->LinkEndChild( subNode3 );

            writeValue( subNode3, "voltageInputsName", data.voltageInputsName[i] );
            writeValue( subNode3, "voltageInputsMinValue", data.voltageInputsMinValue[i] );
            writeValue( subNode3, "voltageInputsMaxValue", data.voltageInputsMaxValue[i] );
            writeValue( subNode3, "voltageInputsUnit", data.voltageInputsUnit[i] );
        }

        subNode2 = m_xmlDoc->NewElement( "currentInputs" );
        subNode->LinkEndChild( subNode2 );

        writeValue( subNode2, "currentInputsActive", data.currentInputsActive );
        writeValue( subNode2, "currentInputsShowOnScreen", data.currentInputsShowOnScreen );

        constexpr size_t countCurrent = sizeof( data.currentInputsMinValue ) / sizeof( data.currentInputsMinValue[0] );
        for ( size_t i = 0; i < countVoltage; ++i )
        {
            auto subNode3 = m_xmlDoc->NewElement( "input" );
            subNode2->LinkEndChild( subNode3 );

            writeValue( subNode3, "currentInputsName", data.currentInputsName[i] );
            writeValue( subNode3, "currentInputsMinValue", data.currentInputsMinValue[i] );
            writeValue( subNode3, "currentInputsMaxValue", data.currentInputsMaxValue[i] );
            writeValue( subNode3, "currentInputsUnit", data.currentInputsUnit[i] );
        }
    }
}
template<>
bool FileXML::readContent<StructIOModule>( ContentType type, StructIOModule &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);

        // voltage inputs
        if ( !errorPresent )
        {
            auto subNode2 = subNode->FirstChildElement( "voltageInputs" );
            errorPresent = (subNode2 == nullptr);

            if ( !errorPresent )
            {
                readValue( subNode2, "voltageInputsActive", data.voltageInputsActive, errorPresent );
                readValue( subNode2, "voltageInputsShowOnScreen", data.voltageInputsShowOnScreen, errorPresent );

                constexpr size_t size = sizeof( data.voltageInputsMinValue ) / sizeof( data.voltageInputsMinValue[0] );

                tinyxml2::XMLElement *lastSubItem{};
                for ( size_t i = 0; i < size && !errorPresent; ++i )
                {
                    lastSubItem = (lastSubItem == nullptr) ? subNode2->FirstChildElement( "input" ) : lastSubItem->NextSiblingElement( "input" );
                    errorPresent = (lastSubItem == nullptr);
                    if ( !errorPresent )
                    {
                        readValue( lastSubItem, "voltageInputsName", data.voltageInputsName[i], sizeof( data.voltageInputsName[i] ) / sizeof( data.voltageInputsName[i][0] ), errorPresent, true );
                        readValue( lastSubItem, "voltageInputsMinValue", data.voltageInputsMinValue[i], errorPresent );
                        readValue( lastSubItem, "voltageInputsMaxValue", data.voltageInputsMaxValue[i], errorPresent );
                        readValue( lastSubItem, "voltageInputsUnit", data.voltageInputsUnit[i], sizeof( data.voltageInputsUnit[i] ) / sizeof( data.voltageInputsUnit[i][0] ), errorPresent, true );
                    }
                }
            }
        }

        // current inputs
        if ( !errorPresent )
        {
            auto subNode2 = subNode->FirstChildElement( "currentInputs" );
            errorPresent = (subNode2 == nullptr);

            if ( !errorPresent )
            {
                readValue( subNode2, "currentInputsActive", data.currentInputsActive, errorPresent );
                readValue( subNode2, "currentInputsShowOnScreen", data.currentInputsShowOnScreen, errorPresent );

                constexpr size_t size = sizeof( data.currentInputsMinValue ) / sizeof( data.currentInputsMinValue[0] );

                tinyxml2::XMLElement *lastSubItem{};
                for ( size_t i = 0; i < size && !errorPresent; ++i )
                {
                    lastSubItem = (lastSubItem == nullptr) ? subNode2->FirstChildElement( "input" ) : lastSubItem->NextSiblingElement( "input" );
                    errorPresent = (lastSubItem == nullptr);
                    if ( !errorPresent )
                    {
                        readValue( lastSubItem, "currentInputsName", data.currentInputsName[i], sizeof( data.currentInputsName[i] ) / sizeof( data.currentInputsName[i][0] ), errorPresent, true );
                        readValue( lastSubItem, "currentInputsMinValue", data.currentInputsMinValue[i], errorPresent );
                        readValue( lastSubItem, "currentInputsMaxValue", data.currentInputsMaxValue[i], errorPresent );
                        readValue( lastSubItem, "currentInputsUnit", data.currentInputsUnit[i], sizeof( data.currentInputsUnit[i] ) / sizeof( data.currentInputsUnit[i][0] ), errorPresent, true );
                    }
                }
            }
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<PowerSetupNode>( ContentType type, const PowerSetupNode &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "version", data.version );
        writeValue( subNode, "countProfile", data.countProfile );
        writeValue( subNode, "countEvents", data.countEvents );
        writeValue( subNode, "active", data.active );
    }
}
template<>
bool FileXML::updateContent<PowerSetupNode>( ContentType type, const PowerSetupNode &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            errorPresent |= !updateValue( subNode, "version", data.version );
            errorPresent |= !updateValue( subNode, "countProfile", data.countProfile );
            errorPresent |= !updateValue( subNode, "countEvents", data.countEvents );
            errorPresent |= !updateValue( subNode, "active", data.active );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}
template<>
bool FileXML::readContent<PowerSetupNode>( ContentType type, PowerSetupNode &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "version", data.version, errorPresent );
            readValue( subNode, "countProfile", data.countProfile, errorPresent );
            readValue( subNode, "countEvents", data.countEvents, errorPresent );
            readValue( subNode, "active", data.active, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<PowerModule>( ContentType type, const PowerModule &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "item", data.item );
        writeValue( subNode, "time", data.time );
        writeValue( subNode, "value", data.value );
        writeValue( subNode, "interval", data.interval );
    }
}
template<>
bool FileXML::readContent<PowerModule>( ContentType type, PowerModule &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "item", data.item, errorPresent );
            readValue( subNode, "time", data.time, errorPresent );
            readValue( subNode, "value", data.value, errorPresent );
            readValue( subNode, "interval", data.interval, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<PowerSetup>( ContentType type, const PowerSetup &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "name", (const char *)(data.name) );

        // Update parent node for sub-structure.
        auto oldParent = setParentNode( subNode );

        writeContent( ContentType::PowerSetup_CPU, data.cpu );
        writeContent( ContentType::PowerSetup_Timer, data.timer );
        writeContent( ContentType::PowerSetup_Inputs, data.inputs );
        writeContent( ContentType::PowerSetup_DeviceGlobal, data.deviceGlobal );
        writeContent( ContentType::PowerSetup_U5V, data.U5V );
        writeContent( ContentType::PowerSetup_U12V, data.U12V );
        writeContent( ContentType::PowerSetup_Communication, data.communication );
        writeContent( ContentType::PowerSetup_DisplayLight, data.displayLight );
        writeContent( ContentType::PowerSetup_DisplayText, data.displayText );
        writeContent( ContentType::PowerSetup_Fingerprint, data.fingerprint );
        writeContent( ContentType::PowerSetup_Transponder, data.transponder );
        writeContent( ContentType::PowerSetup_GPS, data.gps );
        writeContent( ContentType::PowerSetup_USBHost, data.usbHost );
        writeContent( ContentType::PowerSetup_MMC, data.mmc );

        // Reset parent node because the end of the structure is reached.
        setParentNode( oldParent );
    }
}
template<>
bool FileXML::readContent<PowerSetup>( ContentType type, PowerSetup &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "name", (char*)data.name, sizeof( data.name ) / sizeof( data.name[0] ), errorPresent, true );

            // Update parent node for sub-structure.
            auto oldParent = setParentNode( subNode );

            errorPresent |= !readContent( ContentType::PowerSetup_CPU, data.cpu );
            errorPresent |= !readContent( ContentType::PowerSetup_Timer, data.timer );
            errorPresent |= !readContent( ContentType::PowerSetup_Inputs, data.inputs );
            errorPresent |= !readContent( ContentType::PowerSetup_DeviceGlobal, data.deviceGlobal );
            errorPresent |= !readContent( ContentType::PowerSetup_U5V, data.U5V );
            errorPresent |= !readContent( ContentType::PowerSetup_U12V, data.U12V );
            errorPresent |= !readContent( ContentType::PowerSetup_Communication, data.communication );
            errorPresent |= !readContent( ContentType::PowerSetup_DisplayLight, data.displayLight );
            errorPresent |= !readContent( ContentType::PowerSetup_DisplayText, data.displayText );
            errorPresent |= !readContent( ContentType::PowerSetup_Fingerprint, data.fingerprint );
            errorPresent |= !readContent( ContentType::PowerSetup_Transponder, data.transponder );
            errorPresent |= !readContent( ContentType::PowerSetup_GPS, data.gps );
            errorPresent |= !readContent( ContentType::PowerSetup_USBHost, data.usbHost );
            errorPresent |= !readContent( ContentType::PowerSetup_MMC, data.mmc );

            // Reset parent node because the end of the structure is reached.
            setParentNode( oldParent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<PowerEvent>( ContentType type, const PowerEvent &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "type", data.type );
        writeValue( subNode, "profile", data.profile );
        writeValue( subNode, "time", data.time );
        writeValue( subNode, "nextprofile", data.nextprofile );
        writeValue( subNode, "value", data.value );
        writeValue( subNode, "channel", data.channel );
    }
}
template<>
bool FileXML::readContent<PowerEvent>( ContentType type, PowerEvent &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "type", data.type, errorPresent );
            readValue( subNode, "profile", data.profile, errorPresent );
            readValue( subNode, "time", data.time, errorPresent );
            readValue( subNode, "nextprofile", data.nextprofile, errorPresent );
            readValue( subNode, "value", data.value, errorPresent );
            readValue( subNode, "channel", data.channel, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<PowerMemo>( ContentType type, const PowerMemo &data )
{
    assert( m_parent != nullptr );
    if ( m_parent != nullptr )
    {
        writeValue( m_parent, "memo", data.memo );
    }
}
template<>
bool FileXML::readContent<PowerMemo>( ContentType type, PowerMemo &data, tinyxml2::XMLElement *&lastItem )
{
    bool errorPresent = (m_parent == nullptr);
    assert( !errorPresent );
    if ( !errorPresent )
    {
        readValue( m_parent, "memo", data.memo, sizeof( data.memo ) / sizeof( data.memo[0] ), errorPresent, true );
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructConfig>( ContentType type, const StructConfig &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "backlightLCD", data.backlightLCD );
        writeValue( subNode, "waitingTimeForReply", data.waitingTimeForReply );
        writeValue( subNode, "onlineOperation", data.onlineOperation );
        writeValue( subNode, "passwordSystemMenu", data.passwordSystemMenu );
        writeValue( subNode, "operatingMode", data.operatingMode );
        writeValue( subNode, "deactivateOnOffButton", data.deactivateOnOffButton );
        writeValue( subNode, "showInfoPage", data.showInfoPage );
        writeValue( subNode, "presentationModeUnique", data.presentationModeUnique );
        writeValue( subNode, "transponderModeUnique", data.transponderModeUnique );
        writeValue( subNode, "autoIncrementFlags", data.autoIncrementFlags );
        writeValue( subNode, "autoIncrementValue", data.autoIncrementValue );
        writeValue( subNode, "transponderCutFlags", data.transponderCutFlags );
        writeValue( subNode, "transponderCutFrom", data.transponderCutFrom );
        writeValue( subNode, "transponderCutLength", data.transponderCutLength );
        writeValue( subNode, "ioModuleActive", data.ioModuleActive );
        writeValue( subNode, "transponderSegment", data.transponderSegment );
        writeValue( subNode, "transponderType", data.transponderType );
        writeValue( subNode, "keySound", data.keySound );
        writeValue( subNode, "signalLastFunctionKey", data.signalLastFunctionKey );
        writeValue( subNode, "initialValue3", data.initialValue3 );
        writeValue( subNode, "enableEntrance", data.enableEntrance );
        writeValue( subNode, "useExternalReader", data.useExternalReader );
        writeValue( subNode, "doorSignalTime", data.doorSignalTime );
        writeValue( subNode, "doorMonitoring", data.doorMonitoring );
        writeValue( subNode, "doorAlarmTime", data.doorAlarmTime );
        writeValue( subNode, "entranceMainMenu", data.entranceMainMenu );
        writeValue( subNode, "enableCut", data.enableCut );
        writeValue( subNode, "cutFrom", data.cutFrom );
        writeValue( subNode, "cutUpTo", data.cutUpTo );
        writeValue( subNode, "initialValue1", data.initialValue1 );
        writeValue( subNode, "entranceFlag", data.entranceFlag );
        writeValue( subNode, "showBigWatchAlways", data.showBigWatchAlways );
        writeValue( subNode, "timeoutPZE", data.timeoutPZE );
        writeValue( subNode, "lampFlagsMDE", data.lampFlagsMDE );
        writeValue( subNode, "peepFlags", data.peepFlags );
        writeValue( subNode, "transponderUserPassword", data.transponderUserPassword );
        writeValue( subNode, "initialValue2", data.initialValue2 );
        writeValue( subNode, "enableEntrance2", data.enableEntrance2 );

        // Update parent node for sub-structure.
        auto oldParent = setParentNode( subNode );

        writeContent( ContentType::TransponderGlobal, data.transponderGlobal );
        writeContent( ContentType::TransponderAccessControl, data.transponderAccessControl );

        // Reset parent node because the end of the structure is reached.
        setParentNode( oldParent );
    }
}
template<>
bool FileXML::readContent<StructConfig>( ContentType type, StructConfig &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "backlightLCD", data.backlightLCD, errorPresent );
            readValue( subNode, "waitingTimeForReply", data.waitingTimeForReply, errorPresent );
            readValue( subNode, "onlineOperation", data.onlineOperation, errorPresent );
            readValue( subNode, "passwordSystemMenu", data.passwordSystemMenu, errorPresent );
            readValue( subNode, "operatingMode", data.operatingMode, errorPresent );
            readValue( subNode, "deactivateOnOffButton", data.deactivateOnOffButton, errorPresent );
            readValue( subNode, "showInfoPage", data.showInfoPage, errorPresent );
            readValue( subNode, "presentationModeUnique", data.presentationModeUnique, errorPresent );
            readValue( subNode, "transponderModeUnique", data.transponderModeUnique, errorPresent );
            readValue( subNode, "autoIncrementFlags", data.autoIncrementFlags, errorPresent );
            readValue( subNode, "autoIncrementValue", data.autoIncrementValue, errorPresent );
            readValue( subNode, "transponderCutFlags", data.transponderCutFlags, errorPresent );
            readValue( subNode, "transponderCutFrom", data.transponderCutFrom, errorPresent );
            readValue( subNode, "transponderCutLength", data.transponderCutLength, errorPresent );
            readValue( subNode, "ioModuleActive", data.ioModuleActive, errorPresent );
            readValue( subNode, "transponderSegment", data.transponderSegment, errorPresent );
            readValue( subNode, "transponderType", data.transponderType, errorPresent );
            readValue( subNode, "keySound", data.keySound, errorPresent );
            readValue( subNode, "signalLastFunctionKey", data.signalLastFunctionKey, errorPresent );
            readValue( subNode, "initialValue3", data.initialValue3, errorPresent );
            readValue( subNode, "enableEntrance", data.enableEntrance, errorPresent );
            readValue( subNode, "useExternalReader", data.useExternalReader, errorPresent );
            readValue( subNode, "doorSignalTime", data.doorSignalTime, errorPresent );
            readValue( subNode, "doorMonitoring", data.doorMonitoring, errorPresent );
            readValue( subNode, "doorAlarmTime", data.doorAlarmTime, errorPresent );
            readValue( subNode, "entranceMainMenu", data.entranceMainMenu, errorPresent );
            readValue( subNode, "enableCut", data.enableCut, errorPresent );
            readValue( subNode, "cutFrom", data.cutFrom, errorPresent );
            readValue( subNode, "cutUpTo", data.cutUpTo, errorPresent );
            readValue( subNode, "initialValue1", data.initialValue1, errorPresent );
            readValue( subNode, "entranceFlag", data.entranceFlag, errorPresent );
            readValue( subNode, "showBigWatchAlways", data.showBigWatchAlways, errorPresent );
            readValue( subNode, "timeoutPZE", data.timeoutPZE, errorPresent );
            readValue( subNode, "lampFlagsMDE", data.lampFlagsMDE, errorPresent );
            readValue( subNode, "peepFlags", data.peepFlags, errorPresent );
            readValue( subNode, "transponderUserPassword", data.transponderUserPassword, errorPresent );
            readValue( subNode, "initialValue2", data.initialValue2, errorPresent );
            readValue( subNode, "enableEntrance2", data.enableEntrance2, errorPresent );

            if ( !errorPresent )
            {
                // Update parent node for sub-structure.
                auto oldParent = setParentNode( subNode );

                errorPresent |= !readContent( ContentType::TransponderGlobal, data.transponderGlobal );
                errorPresent |= !readContent( ContentType::TransponderAccessControl, data.transponderAccessControl );

                // Reset parent node because the end of the structure is reached.
                setParentNode( oldParent );
            }
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructValueGlobalVar>( ContentType type, const StructValueGlobalVar &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "flags", data.flags );

        constexpr size_t numberOfItems = sizeof( data.value ) / sizeof( data.value[0] );
        for ( size_t i = 0; i < numberOfItems; ++i )
        {
            writeValue( subNode, "value", data.value[i] );
        }
    }
}
template<>
bool FileXML::readContent<StructValueGlobalVar>( ContentType type, StructValueGlobalVar &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "flags", data.flags, errorPresent );

            if ( !errorPresent )
            {
                tinyxml2::XMLElement *subItem{};
                constexpr int itemCount = sizeof( data.value ) / sizeof( data.value[0] );
                constexpr int stringSize = sizeof( data.value[0] ) / sizeof( data.value[0][0] );
                for ( size_t i = 0; i < itemCount && !errorPresent; ++i )
                {
                    readValue( subNode, subItem, "value", data.value[i], stringSize, errorPresent, true );
                    if ( subItem == nullptr )
                    {
                        // No further element was found.
                        errorPresent = true;
                        break;
                    }
                }
            }
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructTransponderLegicAdvantHeader>( ContentType type, const StructTransponderLegicAdvantHeader &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "structSize", data.structSize );
        writeValue( subNode, "structFlags", data.structFlags );
        writeValue( subNode, "transponderType", data.transponderType );
        writeValue( subNode, "count", data.count );
        writeValue( subNode, "mode", data.mode );
        writeValue( subNode, "flags", data.flags );
    }
}
template<>
bool FileXML::readContent<StructTransponderLegicAdvantHeader>( ContentType type, StructTransponderLegicAdvantHeader &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "structSize", data.structSize, errorPresent );
            readValue( subNode, "structFlags", data.structFlags, errorPresent );
            readValue( subNode, "transponderType", data.transponderType, errorPresent );
            readValue( subNode, "count", data.count, errorPresent );
            readValue( subNode, "mode", data.mode, errorPresent );
            readValue( subNode, "flags", data.flags, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructTransponderLegicAdvant>( ContentType type, const StructTransponderLegicAdvant &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "rwType", data.rwType );
        writeValue( subNode, "rwFlags", data.rwFlags );
        writeValue( subNode, "rwSegment", data.rwSegment );
        writeValue( subNode, "rwOffset", data.rwOffset );
        writeValue( subNode, "rwLength", data.rwLength );
        constexpr size_t arraySize = sizeof( data.kghSearch ) / sizeof( data.kghSearch[0] );
        writeValue( subNode, "kghSearch", data.kghSearch, arraySize < data.kghSearchLength ? arraySize : data.kghSearchLength );
        writeValue( subNode, "kghSearchLength", data.kghSearchLength );
        writeValue( subNode, "crcCheck", data.crcCheck );
        writeValue( subNode, "crcAddress", data.crcAddress );
        writeValue( subNode, "outputFormat", data.outputFormat );
        writeValue( subNode, "outputLength", data.outputLength );
    }
}
template<>
bool FileXML::readContent<StructTransponderLegicAdvant>( ContentType type, StructTransponderLegicAdvant &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "rwType", data.rwType, errorPresent );
            readValue( subNode, "rwFlags", data.rwFlags, errorPresent );
            readValue( subNode, "rwSegment", data.rwSegment, errorPresent );
            readValue( subNode, "rwOffset", data.rwOffset, errorPresent );
            readValue( subNode, "rwLength", data.rwLength, errorPresent );
            readValue( subNode, "kghSearchLength", data.kghSearchLength, errorPresent );
            constexpr size_t arraySize = sizeof( data.kghSearch ) / sizeof( data.kghSearch[0] );
            readValue( subNode, "kghSearch", data.kghSearch, arraySize < data.kghSearchLength ? arraySize : data.kghSearchLength, errorPresent );
            readValue( subNode, "crcCheck", data.crcCheck, errorPresent );
            readValue( subNode, "crcAddress", data.crcAddress, errorPresent );
            readValue( subNode, "outputFormat", data.outputFormat, errorPresent );
            readValue( subNode, "outputLength", data.outputLength, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructTransponderMifareDesfireHeader>( ContentType type, const StructTransponderMifareDesfireHeader &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "structSize", data.structSize );
        writeValue( subNode, "structFlags", data.structFlags );
        writeValue( subNode, "transponderType", data.transponderType );
        writeValue( subNode, "count", data.count );
    }
}
template<>
bool FileXML::readContent<StructTransponderMifareDesfireHeader>( ContentType type, StructTransponderMifareDesfireHeader &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "structSize", data.structSize, errorPresent );
            readValue( subNode, "structFlags", data.structFlags, errorPresent );
            readValue( subNode, "transponderType", data.transponderType, errorPresent );
            readValue( subNode, "count", data.count, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructTransponderMifareDesfire>( ContentType type, const StructTransponderMifareDesfire &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "rwFlags", data.rwFlags );
        writeValue( subNode, "applicationNumber", data.applicationNumber );
        writeValue( subNode, "fileNumber", data.fileNumber );
        writeValue( subNode, "rwOffset", data.rwOffset );
        writeValue( subNode, "rwLength", data.rwLength );
        writeValue( subNode, "fileType", data.fileType );
        writeValue( subNode, "communicationMode", data.communicationMode );
        writeValue( subNode, "keyNumber", data.keyNumber );
        writeValue( subNode, "keyFlags", data.keyFlags );
        writeValue( subNode, "applicationKey", data.applicationKey, sizeof( data.applicationKey ) / sizeof( data.applicationKey[0] ) );
        writeValue( subNode, "cryptType", data.cryptType );
        writeValue( subNode, "outputFormat", data.outputFormat );
        writeValue( subNode, "outputLength", data.outputLength );
    }
}
template<>
bool FileXML::readContent<StructTransponderMifareDesfire>( ContentType type, StructTransponderMifareDesfire &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "rwFlags", data.rwFlags, errorPresent );
            readValue( subNode, "applicationNumber", data.applicationNumber, errorPresent );
            readValue( subNode, "fileNumber", data.fileNumber, errorPresent );
            readValue( subNode, "rwOffset", data.rwOffset, errorPresent );
            readValue( subNode, "rwLength", data.rwLength, errorPresent );
            readValue( subNode, "fileType", data.fileType, errorPresent );
            readValue( subNode, "communicationMode", data.communicationMode, errorPresent );
            readValue( subNode, "keyNumber", data.keyNumber, errorPresent );
            readValue( subNode, "keyFlags", data.keyFlags, errorPresent );
            readValue( subNode, "applicationKey", data.applicationKey, sizeof( data.applicationKey ) / sizeof( data.applicationKey[0] ), errorPresent );
            readValue( subNode, "cryptType", data.cryptType, errorPresent );
            readValue( subNode, "outputFormat", data.outputFormat, errorPresent );
            readValue( subNode, "outputLength", data.outputLength, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructTransponderMifarePlusHeader>( ContentType type, const StructTransponderMifarePlusHeader &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "structSize", data.structSize );
        writeValue( subNode, "structFlags", data.structFlags );
        writeValue( subNode, "transponderType", data.transponderType );
        writeValue( subNode, "count", data.count );
        writeValue( subNode, "securityLevel", data.securityLevel );
        writeValue( subNode, "iso14443A", data.iso14443A );
    }
}
template<>
bool FileXML::readContent<StructTransponderMifarePlusHeader>( ContentType type, StructTransponderMifarePlusHeader &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "structSize", data.structSize, errorPresent );
            readValue( subNode, "structFlags", data.structFlags, errorPresent );
            readValue( subNode, "transponderType", data.transponderType, errorPresent );
            readValue( subNode, "count", data.count, errorPresent );
            readValue( subNode, "securityLevel", data.securityLevel, errorPresent );
            readValue( subNode, "iso14443A", data.iso14443A, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructTransponderMifarePlus>( ContentType type, const StructTransponderMifarePlus &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "rwFlags", data.rwFlags );
        writeValue( subNode, "rwSector", data.rwSector );
        writeValue( subNode, "rwBlock", data.rwBlock );
        writeValue( subNode, "rwOffset", data.rwOffset );
        writeValue( subNode, "rwLength", data.rwLength );
        writeValue( subNode, "passwordUsage", data.passwordUsage );
        writeValue( subNode, "passwordFlags", data.passwordFlags );
        writeValue( subNode, "passwordA_Crypto1", data.passwordA_Crypto1, sizeof( data.passwordA_Crypto1 ) / sizeof( data.passwordA_Crypto1[0] ) );
        writeValue( subNode, "passwordB_Crypto1", data.passwordB_Crypto1, sizeof( data.passwordB_Crypto1 ) / sizeof( data.passwordB_Crypto1[0] ) );
        writeValue( subNode, "passwordA_AES", data.passwordA_AES, sizeof( data.passwordA_AES ) / sizeof( data.passwordA_AES[0] ) );
        writeValue( subNode, "passwordB_AES", data.passwordB_AES, sizeof( data.passwordB_AES ) / sizeof( data.passwordB_AES[0] ) );
        writeValue( subNode, "instID", data.instID, sizeof( data.instID ) / sizeof( data.instID[0] ) );
        writeValue( subNode, "outputFormat", data.outputFormat );
        writeValue( subNode, "outputLength", data.outputLength );
    }
}
template<>
bool FileXML::readContent<StructTransponderMifarePlus>( ContentType type, StructTransponderMifarePlus &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "rwFlags", data.rwFlags, errorPresent );
            readValue( subNode, "rwSector", data.rwSector, errorPresent );
            readValue( subNode, "rwBlock", data.rwBlock, errorPresent );
            readValue( subNode, "rwOffset", data.rwOffset, errorPresent );
            readValue( subNode, "rwLength", data.rwLength, errorPresent );
            readValue( subNode, "passwordUsage", data.passwordUsage, errorPresent );
            readValue( subNode, "passwordFlags", data.passwordFlags, errorPresent );
            readValue( subNode, "passwordA_Crypto1", data.passwordA_Crypto1, sizeof( data.passwordA_Crypto1 ) / sizeof( data.passwordA_Crypto1[0] ), errorPresent );
            readValue( subNode, "passwordB_Crypto1", data.passwordB_Crypto1, sizeof( data.passwordB_Crypto1 ) / sizeof( data.passwordB_Crypto1[0] ), errorPresent );
            readValue( subNode, "passwordA_AES", data.passwordA_AES, sizeof( data.passwordA_AES ) / sizeof( data.passwordA_AES[0] ), errorPresent );
            readValue( subNode, "passwordB_AES", data.passwordB_AES, sizeof( data.passwordB_AES ) / sizeof( data.passwordB_AES[0] ), errorPresent );
            readValue( subNode, "instID", data.instID, sizeof( data.instID ) / sizeof( data.instID[0] ), errorPresent );
            readValue( subNode, "outputFormat", data.outputFormat, errorPresent );
            readValue( subNode, "outputLength", data.outputLength, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructTransponderMifareUltralightHeader>( ContentType type, const StructTransponderMifareUltralightHeader &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "structSize", data.structSize );
        writeValue( subNode, "structFlags", data.structFlags );
        writeValue( subNode, "transponderType", data.transponderType );
        writeValue( subNode, "count", data.count );
    }
}
template<>
bool FileXML::readContent<StructTransponderMifareUltralightHeader>( ContentType type, StructTransponderMifareUltralightHeader &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "structSize", data.structSize, errorPresent );
            readValue( subNode, "structFlags", data.structFlags, errorPresent );
            readValue( subNode, "transponderType", data.transponderType, errorPresent );
            readValue( subNode, "count", data.count, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructTransponderMifareUltralight>( ContentType type, const StructTransponderMifareUltralight &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "rwFlags", data.rwFlags );
        writeValue( subNode, "rwBlock", data.rwBlock );
        writeValue( subNode, "rwOffset", data.rwOffset );
        writeValue( subNode, "rwLength", data.rwLength );
        writeValue( subNode, "keyUsage", data.keyUsage );
        writeValue( subNode, "keyFlags", data.keyFlags );
        writeValue( subNode, "key", data.key, sizeof( data.key ) / sizeof( data.key[0] ) );
        writeValue( subNode, "outputFormat", data.outputFormat );
        writeValue( subNode, "outputLength", data.outputLength );
    }
}
template<>
bool FileXML::readContent<StructTransponderMifareUltralight>( ContentType type, StructTransponderMifareUltralight &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "rwFlags", data.rwFlags, errorPresent );
            readValue( subNode, "rwBlock", data.rwBlock, errorPresent );
            readValue( subNode, "rwOffset", data.rwOffset, errorPresent );
            readValue( subNode, "rwLength", data.rwLength, errorPresent );
            readValue( subNode, "keyUsage", data.keyUsage, errorPresent );
            readValue( subNode, "keyFlags", data.keyFlags, errorPresent );
            readValue( subNode, "key", data.key, sizeof( data.key ) / sizeof( data.key[0] ), errorPresent );
            readValue( subNode, "outputFormat", data.outputFormat, errorPresent );
            readValue( subNode, "outputLength", data.outputLength, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructTransponderiCodeHeader>( ContentType type, const StructTransponderiCodeHeader &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "structSize", data.structSize );
        writeValue( subNode, "structFlags", data.structFlags );
        writeValue( subNode, "transponderType", data.transponderType );
        writeValue( subNode, "count", data.count );
    }
}
template<>
bool FileXML::readContent<StructTransponderiCodeHeader>( ContentType type, StructTransponderiCodeHeader &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "structSize", data.structSize, errorPresent );
            readValue( subNode, "structFlags", data.structFlags, errorPresent );
            readValue( subNode, "transponderType", data.transponderType, errorPresent );
            readValue( subNode, "count", data.count, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructTransponderiCode>( ContentType type, const StructTransponderiCode &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "rwFlags", data.rwFlags );
        writeValue( subNode, "rwBlock", data.rwBlock );
        writeValue( subNode, "rwOffset", data.rwOffset );
        writeValue( subNode, "rwLength", data.rwLength );
        writeValue( subNode, "outputFormat", data.outputFormat );
        writeValue( subNode, "outputLength", data.outputLength );
    }
}
template<>
bool FileXML::readContent<StructTransponderiCode>( ContentType type, StructTransponderiCode &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "rwFlags", data.rwFlags, errorPresent );
            readValue( subNode, "rwBlock", data.rwBlock, errorPresent );
            readValue( subNode, "rwOffset", data.rwOffset, errorPresent );
            readValue( subNode, "rwLength", data.rwLength, errorPresent );
            readValue( subNode, "outputFormat", data.outputFormat, errorPresent );
            readValue( subNode, "outputLength", data.outputLength, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructTransponderJavaCardHeader>( ContentType type, const StructTransponderJavaCardHeader &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "structSize", data.structSize );
        writeValue( subNode, "structFlags", data.structFlags );
        writeValue( subNode, "transponderType", data.transponderType );
        writeValue( subNode, "count", data.count );
    }
}
template<>
bool FileXML::readContent<StructTransponderJavaCardHeader>( ContentType type, StructTransponderJavaCardHeader &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "structSize", data.structSize, errorPresent );
            readValue( subNode, "structFlags", data.structFlags, errorPresent );
            readValue( subNode, "transponderType", data.transponderType, errorPresent );
            readValue( subNode, "count", data.count, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructTransponderJavaCard>( ContentType type, const StructTransponderJavaCard &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "rwFlags", data.rwFlags );
        writeValue( subNode, "rwBlock", data.rwBlock );
        writeValue( subNode, "rwOffset", data.rwOffset );
        writeValue( subNode, "rwLength", data.rwLength );
        writeValue( subNode, "outputFormat", data.outputFormat );
        writeValue( subNode, "outputLength", data.outputLength );
    }
}
template<>
bool FileXML::readContent<StructTransponderJavaCard>( ContentType type, StructTransponderJavaCard &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "rwFlags", data.rwFlags, errorPresent );
            readValue( subNode, "rwBlock", data.rwBlock, errorPresent );
            readValue( subNode, "rwOffset", data.rwOffset, errorPresent );
            readValue( subNode, "rwLength", data.rwLength, errorPresent );
            readValue( subNode, "outputFormat", data.outputFormat, errorPresent );
            readValue( subNode, "outputLength", data.outputLength, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructTransponderMyDHeader>( ContentType type, const StructTransponderMyDHeader &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "structSize", data.structSize );
        writeValue( subNode, "structFlags", data.structFlags );
        writeValue( subNode, "transponderType", data.transponderType );
        writeValue( subNode, "count", data.count );
        writeValue( subNode, "isoFlags", data.isoFlags );
    }
}
template<>
bool FileXML::readContent<StructTransponderMyDHeader>( ContentType type, StructTransponderMyDHeader &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "structSize", data.structSize, errorPresent );
            readValue( subNode, "structFlags", data.structFlags, errorPresent );
            readValue( subNode, "transponderType", data.transponderType, errorPresent );
            readValue( subNode, "count", data.count, errorPresent );
            readValue( subNode, "isoFlags", data.isoFlags, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructTransponderMyD>( ContentType type, const StructTransponderMyD &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "rwFlags", data.rwFlags );
        writeValue( subNode, "rwBlock", data.rwBlock );
        writeValue( subNode, "rwOffset", data.rwOffset );
        writeValue( subNode, "rwLength", data.rwLength );
        writeValue( subNode, "keyUsage", data.keyUsage );
        writeValue( subNode, "keyFlags", data.keyFlags );
        writeValue( subNode, "keyA", data.keyA, sizeof( data.keyA ) / sizeof( data.keyA[0] ) );
        writeValue( subNode, "keyB", data.keyB, sizeof( data.keyB ) / sizeof( data.keyB[0] ) );
        writeValue( subNode, "outputFormat", data.outputFormat );
        writeValue( subNode, "outputLength", data.outputLength );
    }
}
template<>
bool FileXML::readContent<StructTransponderMyD>( ContentType type, StructTransponderMyD &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "rwFlags", data.rwFlags, errorPresent );
            readValue( subNode, "rwBlock", data.rwBlock, errorPresent );
            readValue( subNode, "rwOffset", data.rwOffset, errorPresent );
            readValue( subNode, "rwLength", data.rwLength, errorPresent );
            readValue( subNode, "keyUsage", data.keyUsage, errorPresent );
            readValue( subNode, "keyFlags", data.keyFlags, errorPresent );
            readValue( subNode, "keyA", data.keyA, sizeof( data.keyA ) / sizeof( data.keyA[0] ), errorPresent );
            readValue( subNode, "keyB", data.keyB, sizeof( data.keyB ) / sizeof( data.keyB[0] ), errorPresent );
            readValue( subNode, "outputFormat", data.outputFormat, errorPresent );
            readValue( subNode, "outputLength", data.outputLength, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructTransponderISOTagsHeader>( ContentType type, const StructTransponderISOTagsHeader &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "structSize", data.structSize );
        writeValue( subNode, "structFlags", data.structFlags );
        writeValue( subNode, "transponderType", data.transponderType );
        writeValue( subNode, "count", data.count );
    }
}
template<>
bool FileXML::readContent<StructTransponderISOTagsHeader>( ContentType type, StructTransponderISOTagsHeader &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "structSize", data.structSize, errorPresent );
            readValue( subNode, "structFlags", data.structFlags, errorPresent );
            readValue( subNode, "transponderType", data.transponderType, errorPresent );
            readValue( subNode, "count", data.count, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructTransponderISOTags>( ContentType type, const StructTransponderISOTags &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "isoType", data.isoType );
        writeValue( subNode, "iso14443A", data.iso14443A );
        writeValue( subNode, "iso14443A_2", data.iso14443A_2 );
        writeValue( subNode, "iso15693", data.iso15693 );
        writeValue( subNode, "outputFormat", data.outputFormat );
        writeValue( subNode, "outputLength", data.outputLength );
    }
}
template<>
bool FileXML::readContent<StructTransponderISOTags>( ContentType type, StructTransponderISOTags &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "isoType", data.isoType, errorPresent );
            readValue( subNode, "iso14443A", data.iso14443A, errorPresent );
            readValue( subNode, "iso14443A_2", data.iso14443A_2, errorPresent );
            readValue( subNode, "iso15693", data.iso15693, errorPresent );
            readValue( subNode, "outputFormat", data.outputFormat, errorPresent );
            readValue( subNode, "outputLength", data.outputLength, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructTransponderWiegandClockDataHeader>( ContentType type, const StructTransponderWiegandClockDataHeader &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "structSize", data.structSize );
        writeValue( subNode, "structFlags", data.structFlags );
        writeValue( subNode, "transponderType", data.transponderType );
        writeValue( subNode, "count", data.count );
    }
}
template<>
bool FileXML::readContent<StructTransponderWiegandClockDataHeader>( ContentType type, StructTransponderWiegandClockDataHeader &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "structSize", data.structSize, errorPresent );
            readValue( subNode, "structFlags", data.structFlags, errorPresent );
            readValue( subNode, "transponderType", data.transponderType, errorPresent );
            readValue( subNode, "count", data.count, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructTransponderWiegandClockData>( ContentType type, const StructTransponderWiegandClockData &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "flags", data.flags );
        writeValue( subNode, "clsWireActiveState", data.clsWireActiveState );
        writeValue( subNode, "bitFormat", data.bitFormat );
        writeValue( subNode, "numberOfBits", data.numberOfBits );
        writeValue( subNode, "parityFlags01", data.parityFlags01 );
        writeValue( subNode, "parityBitPos01", data.parityBitPos01 );
        writeValue( subNode, "parityStart01", data.parityStart01 );
        writeValue( subNode, "parityEnd01", data.parityEnd01 );
        writeValue( subNode, "parityFlags02", data.parityFlags02 );
        writeValue( subNode, "parityBitPos02", data.parityBitPos02 );
        writeValue( subNode, "parityStart02", data.parityStart02 );
        writeValue( subNode, "parityEnd02", data.parityEnd02 );
        writeValue( subNode, "facilityCode", data.facilityCode, sizeof( data.facilityCode ) / sizeof( data.facilityCode[0] ) );
        writeValue( subNode, "facilityCodeFormat", data.facilityCodeFormat );
        writeValue( subNode, "facilityCodeBitPos", data.facilityCodeBitPos );
        writeValue( subNode, "facilityCodeNumberOfBits", data.facilityCodeNumberOfBits );
        writeValue( subNode, "cardDataBitPos", data.cardDataBitPos );
        writeValue( subNode, "cardDataNumberOfBits", data.cardDataNumberOfBits );
        writeValue( subNode, "outputData", data.outputData );
        writeValue( subNode, "outputFormat", data.outputFormat );
        writeValue( subNode, "outputLength", data.outputLength );
    }
}
template<>
bool FileXML::readContent<StructTransponderWiegandClockData>( ContentType type, StructTransponderWiegandClockData &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "flags", data.flags, errorPresent );
            readValue( subNode, "clsWireActiveState", data.clsWireActiveState, errorPresent );
            readValue( subNode, "bitFormat", data.bitFormat, errorPresent );
            readValue( subNode, "numberOfBits", data.numberOfBits, errorPresent );
            readValue( subNode, "parityFlags01", data.parityFlags01, errorPresent );
            readValue( subNode, "parityBitPos01", data.parityBitPos01, errorPresent );
            readValue( subNode, "parityStart01", data.parityStart01, errorPresent );
            readValue( subNode, "parityEnd01", data.parityEnd01, errorPresent );
            readValue( subNode, "parityFlags02", data.parityFlags02, errorPresent );
            readValue( subNode, "parityBitPos02", data.parityBitPos02, errorPresent );
            readValue( subNode, "parityStart02", data.parityStart02, errorPresent );
            readValue( subNode, "parityEnd02", data.parityEnd02, errorPresent );
            readValue( subNode, "facilityCode", data.facilityCode, sizeof( data.facilityCode ) / sizeof( data.facilityCode[0] ), errorPresent );
            readValue( subNode, "facilityCodeFormat", data.facilityCodeFormat, errorPresent );
            readValue( subNode, "facilityCodeBitPos", data.facilityCodeBitPos, errorPresent );
            readValue( subNode, "facilityCodeNumberOfBits", data.facilityCodeNumberOfBits, errorPresent );
            readValue( subNode, "cardDataBitPos", data.cardDataBitPos, errorPresent );
            readValue( subNode, "cardDataNumberOfBits", data.cardDataNumberOfBits, errorPresent );
            readValue( subNode, "outputData", data.outputData, errorPresent );
            readValue( subNode, "outputFormat", data.outputFormat, errorPresent );
            readValue( subNode, "outputLength", data.outputLength, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructTransponderETSITagsHeader>( ContentType type, const StructTransponderETSITagsHeader &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "structSize", data.structSize );
        writeValue( subNode, "structFlags", data.structFlags );
        writeValue( subNode, "transponderType", data.transponderType );
        writeValue( subNode, "count", data.count );
    }
}
template<>
bool FileXML::readContent<StructTransponderETSITagsHeader>( ContentType type, StructTransponderETSITagsHeader &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "structSize", data.structSize, errorPresent );
            readValue( subNode, "structFlags", data.structFlags, errorPresent );
            readValue( subNode, "transponderType", data.transponderType, errorPresent );
            readValue( subNode, "count", data.count, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructTransponderETSITags>( ContentType type, const StructTransponderETSITags &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "strength", data.strength );
        writeValue( subNode, "rwFlags", data.rwFlags );
        writeValue( subNode, "rwBank", data.rwBank );
        writeValue( subNode, "rwOffset", data.rwOffset );
        writeValue( subNode, "rwLength", data.rwLength );
        writeValue( subNode, "password", data.password );
        writeValue( subNode, "outputFormat", data.outputFormat );
        writeValue( subNode, "outputLength", data.outputLength );
    }
}
template<>
bool FileXML::readContent<StructTransponderETSITags>( ContentType type, StructTransponderETSITags &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "strength", data.strength, errorPresent );
            readValue( subNode, "rwFlags", data.rwFlags, errorPresent );
            readValue( subNode, "rwBank", data.rwBank, errorPresent );
            readValue( subNode, "rwOffset", data.rwOffset, errorPresent );
            readValue( subNode, "rwLength", data.rwLength, errorPresent );
            readValue( subNode, "password", data.password, errorPresent );
            readValue( subNode, "outputFormat", data.outputFormat, errorPresent );
            readValue( subNode, "outputLength", data.outputLength, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructTransponderQ5Header>( ContentType type, const StructTransponderQ5Header &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "structSize", data.structSize );
        writeValue( subNode, "structFlags", data.structFlags );
        writeValue( subNode, "transponderType", data.transponderType );
        writeValue( subNode, "count", data.count );
        writeValue( subNode, "isoFlags", data.isoFlags );
    }
}
template<>
bool FileXML::readContent<StructTransponderQ5Header>( ContentType type, StructTransponderQ5Header &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "structSize", data.structSize, errorPresent );
            readValue( subNode, "structFlags", data.structFlags, errorPresent );
            readValue( subNode, "transponderType", data.transponderType, errorPresent );
            readValue( subNode, "count", data.count, errorPresent );
            readValue( subNode, "isoFlags", data.isoFlags, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructTransponderQ5>( ContentType type, const StructTransponderQ5 &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "rwFlags", data.rwFlags );
        writeValue( subNode, "rwBlock", data.rwBlock );
        writeValue( subNode, "rwOffset", data.rwOffset );
        writeValue( subNode, "rwLength", data.rwLength );
        writeValue( subNode, "keyUsage", data.keyUsage );
        writeValue( subNode, "keyFlags", data.keyFlags );
        writeValue( subNode, "keyA", data.keyA, sizeof( data.keyA ) / sizeof( data.keyA[0] ) );
        writeValue( subNode, "keyB", data.keyB, sizeof( data.keyB ) / sizeof( data.keyB[0] ) );
        writeValue( subNode, "outputFormat", data.outputFormat );
        writeValue( subNode, "outputLength", data.outputLength );
    }
}
template<>
bool FileXML::readContent<StructTransponderQ5>( ContentType type, StructTransponderQ5 &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "rwFlags", data.rwFlags, errorPresent );
            readValue( subNode, "rwBlock", data.rwBlock, errorPresent );
            readValue( subNode, "rwOffset", data.rwOffset, errorPresent );
            readValue( subNode, "rwLength", data.rwLength, errorPresent );
            readValue( subNode, "keyUsage", data.keyUsage, errorPresent );
            readValue( subNode, "keyFlags", data.keyFlags, errorPresent );
            readValue( subNode, "keyA", data.keyA, sizeof( data.keyA ) / sizeof( data.keyA[0] ), errorPresent );
            readValue( subNode, "keyB", data.keyB, sizeof( data.keyB ) / sizeof( data.keyB[0] ), errorPresent );
            readValue( subNode, "outputFormat", data.outputFormat, errorPresent );
            readValue( subNode, "outputLength", data.outputLength, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructTransponderHitag1FeigHeader>( ContentType type, const StructTransponderHitag1FeigHeader &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "structSize", data.structSize );
        writeValue( subNode, "structFlags", data.structFlags );
        writeValue( subNode, "transponderType", data.transponderType );
        writeValue( subNode, "count", data.count );
    }
}
template<>
bool FileXML::readContent<StructTransponderHitag1FeigHeader>( ContentType type, StructTransponderHitag1FeigHeader &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "structSize", data.structSize, errorPresent );
            readValue( subNode, "structFlags", data.structFlags, errorPresent );
            readValue( subNode, "transponderType", data.transponderType, errorPresent );
            readValue( subNode, "count", data.count, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructTransponderHitag1Feig>( ContentType type, const StructTransponderHitag1Feig &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "flags", data.flags );
        writeValue( subNode, "facilityCodeStart", data.facilityCodeStart );
        writeValue( subNode, "facilityCodeCount", data.facilityCodeCount );
        writeValue( subNode, "facilityCode", data.facilityCode, sizeof( data.facilityCode ) / sizeof( data.facilityCode[0] ) );
        writeValue( subNode, "cardNumberStart", data.cardNumberStart );
        writeValue( subNode, "cardNumberCount", data.cardNumberCount );
        writeValue( subNode, "outputData", data.outputData );
        writeValue( subNode, "outputFormat", data.outputFormat );
        writeValue( subNode, "outputLength", data.outputLength );
    }
}
template<>
bool FileXML::readContent<StructTransponderHitag1Feig>( ContentType type, StructTransponderHitag1Feig &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "flags", data.flags, errorPresent );
            readValue( subNode, "facilityCodeStart", data.facilityCodeStart, errorPresent );
            readValue( subNode, "facilityCodeCount", data.facilityCodeCount, errorPresent );
            readValue( subNode, "facilityCode", data.facilityCode, sizeof( data.facilityCode ) / sizeof( data.facilityCode[0] ), errorPresent );
            readValue( subNode, "cardNumberStart", data.cardNumberStart, errorPresent );
            readValue( subNode, "cardNumberCount", data.cardNumberCount, errorPresent );
            readValue( subNode, "outputData", data.outputData, errorPresent );
            readValue( subNode, "outputFormat", data.outputFormat, errorPresent );
            readValue( subNode, "outputLength", data.outputLength, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructTransponderPreconfiguredReaderHeader>( ContentType type, const StructTransponderPreconfiguredReaderHeader &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "structSize", data.structSize );
        writeValue( subNode, "structFlags", data.structFlags );
        writeValue( subNode, "transponderType", data.transponderType );
        writeValue( subNode, "count", data.count );
    }
}
template<>
bool FileXML::readContent<StructTransponderPreconfiguredReaderHeader>( ContentType type, StructTransponderPreconfiguredReaderHeader &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "structSize", data.structSize, errorPresent );
            readValue( subNode, "structFlags", data.structFlags, errorPresent );
            readValue( subNode, "transponderType", data.transponderType, errorPresent );
            readValue( subNode, "count", data.count, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructTransponderPreconfiguredReader>( ContentType type, const StructTransponderPreconfiguredReader &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "flags", data.flags );
    }
}
template<>
bool FileXML::readContent<StructTransponderPreconfiguredReader>( ContentType type, StructTransponderPreconfiguredReader &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "flags", data.flags, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructFingerprint>( ContentType type, const StructFingerprint &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "method", data.method );
        writeValue( subNode, "segment", data.segment );
        writeValue( subNode, "securityLevel", data.securityLevel );
        writeValue( subNode, "securityFRR", data.securityFRR );
        writeValue( subNode, "securityFAR", data.securityFAR );
        writeValue( subNode, "templateFormat", data.templateFormat );
        writeValue( subNode, "quality", data.quality );
        writeValue( subNode, "minutiaeCount", data.minutiaeCount );
        writeValue( subNode, "passwordGroup", data.passwordGroup );
        writeValue( subNode, "type", data.type );
        writeValue( subNode, "segment2", data.segment2 );
    }
}
template<>
bool FileXML::readContent<StructFingerprint>( ContentType type, StructFingerprint &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "method", data.method, errorPresent );
            readValue( subNode, "segment", data.segment, errorPresent );
            readValue( subNode, "securityLevel", data.securityLevel, errorPresent );
            readValue( subNode, "securityFRR", data.securityFRR, errorPresent );
            readValue( subNode, "securityFAR", data.securityFAR, errorPresent );
            readValue( subNode, "templateFormat", data.templateFormat, errorPresent );
            readValue( subNode, "quality", data.quality, errorPresent );
            readValue( subNode, "minutiaeCount", data.minutiaeCount, errorPresent );
            readValue( subNode, "passwordGroup", data.passwordGroup, errorPresent );
            readValue( subNode, "type", data.type, errorPresent );
            readValue( subNode, "segment2", data.segment2, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructTimeboyAdaption>( ContentType type, const StructTimeboyAdaption &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "method", data.method );
        writeValue( subNode, "count", data.count );
        writeValue( subNode, "records", data.records );
        writeValue( subNode, "invokeInputChainPush", data.invokeInputChainPush );
        writeValue( subNode, "invokeInputChainPull", data.invokeInputChainPull );
    }
}
template<>
bool FileXML::readContent<StructTimeboyAdaption>( ContentType type, StructTimeboyAdaption &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "method", data.method, errorPresent );
            readValue( subNode, "count", data.count, errorPresent );
            readValue( subNode, "records", data.records, errorPresent );
            readValue( subNode, "invokeInputChainPush", data.invokeInputChainPush, errorPresent );
            readValue( subNode, "invokeInputChainPull", data.invokeInputChainPull, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructGPRS>( ContentType type, const StructGPRS &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "method", data.method );
        writeValue( subNode, "invokeInputChain", data.invokeInputChain );
    }
}
template<>
bool FileXML::readContent<StructGPRS>( ContentType type, StructGPRS &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "method", data.method, errorPresent );
            readValue( subNode, "invokeInputChain", data.invokeInputChain, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructEnergieOption>( ContentType type, const StructEnergieOption &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "mode", data.mode );
        writeValue( subNode, "digitalIn", data.digitalIn );
        writeValue( subNode, "timeout", data.timeout );
        writeValue( subNode, "invokeRelease", data.invokeRelease );
        writeValue( subNode, "invokeGPS", data.invokeGPS );
        writeValue( subNode, "digitalOut", data.digitalOut );
        writeValue( subNode, "flags", data.flags );
        writeValue( subNode, "powerLow", data.powerLow );
    }
}
template<>
bool FileXML::readContent<StructEnergieOption>( ContentType type, StructEnergieOption &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "mode", data.mode, errorPresent );
            readValue( subNode, "digitalIn", data.digitalIn, errorPresent );
            readValue( subNode, "timeout", data.timeout, errorPresent );
            readValue( subNode, "invokeRelease", data.invokeRelease, errorPresent );
            readValue( subNode, "invokeGPS", data.invokeGPS, errorPresent );
            readValue( subNode, "digitalOut", data.digitalOut, errorPresent );
            readValue( subNode, "flags", data.flags, errorPresent );
            readValue( subNode, "powerLow", data.powerLow, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructGPS>( ContentType type, const StructGPS &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "startSpeed", data.startSpeed );
        writeValue( subNode, "startMove", data.startMove );
        writeValue( subNode, "invokeInputChainStart", data.invokeInputChainStart );
        writeValue( subNode, "stopTimeout", data.stopTimeout );
        writeValue( subNode, "stopSpeed", data.stopSpeed );
        writeValue( subNode, "invokeInputChainStop", data.invokeInputChainStop );
        writeValue( subNode, "cycle", data.cycle );
        writeValue( subNode, "invokeInputChainCycle", data.invokeInputChainCycle );
        writeValue( subNode, "course", data.course );
        writeValue( subNode, "invokeInputChainCourse", data.invokeInputChainCourse );
        writeValue( subNode, "move", data.move );
        writeValue( subNode, "invokeInputChainMove", data.invokeInputChainMove );
    }
}
template<>
bool FileXML::readContent<StructGPS>( ContentType type, StructGPS &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "startSpeed", data.startSpeed, errorPresent );
            readValue( subNode, "startMove", data.startMove, errorPresent );
            readValue( subNode, "invokeInputChainStart", data.invokeInputChainStart, errorPresent );
            readValue( subNode, "stopTimeout", data.stopTimeout, errorPresent );
            readValue( subNode, "stopSpeed", data.stopSpeed, errorPresent );
            readValue( subNode, "invokeInputChainStop", data.invokeInputChainStop, errorPresent );
            readValue( subNode, "cycle", data.cycle, errorPresent );
            readValue( subNode, "invokeInputChainCycle", data.invokeInputChainCycle, errorPresent );
            readValue( subNode, "course", data.course, errorPresent );
            readValue( subNode, "invokeInputChainCourse", data.invokeInputChainCourse, errorPresent );
            readValue( subNode, "move", data.move, errorPresent );
            readValue( subNode, "invokeInputChainMove", data.invokeInputChainMove, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructKeyBox>( ContentType type, const StructKeyBox &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "random", data.random );
        writeValue( subNode, "flags", data.flags );
        writeValue( subNode, "key", (uint8_t *)data.key, sizeof( data.key ) / sizeof( data.key[0] ) );
        writeValue( subNode, "crc", data.crc );
    }
}
template<>
bool FileXML::readContent<StructKeyBox>( ContentType type, StructKeyBox &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "random", data.random, errorPresent );
            readValue( subNode, "flags", data.flags, errorPresent );
            readValue( subNode, "key", (uint8_t *)data.key, sizeof( data.key ) / sizeof( data.key[0] ), errorPresent );
            readValue( subNode, "crc", data.crc, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructExtended>( ContentType type, const StructExtended &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "doubleInputTime", data.doubleInputTime );
        writeValue( subNode, "confirmationTime", data.confirmationTime );
        writeValue( subNode, "daylightActive", data.daylightActive );
        writeValue( subNode, "daylight1Month", data.daylight1Month );
        writeValue( subNode, "daylight1Week", data.daylight1Week );
        writeValue( subNode, "daylight1Day", data.daylight1Day );
        writeValue( subNode, "daylight1Hour", data.daylight1Hour );
        writeValue( subNode, "daylight2Month", data.daylight2Month );
        writeValue( subNode, "daylight2Week", data.daylight2Week );
        writeValue( subNode, "daylight2Day", data.daylight2Day );
        writeValue( subNode, "daylight2Hour", data.daylight2Hour );
        writeValue( subNode, "transponderPassword32", data.transponderPassword32 );
        writeValue( subNode, "transponderPasswordA0", data.transponderPasswordA0 );
        writeValue( subNode, "phgAESKey", data.phgAESKey, sizeof( data.phgAESKey ) / sizeof( data.phgAESKey[0] ) );
        writeValue( subNode, "transponderPasswordFlag", data.transponderPasswordFlag );
        writeValue( subNode, "transponderPasswordB0", data.transponderPasswordB0 );
        writeValue( subNode, "transponderPasswordA1", data.transponderPasswordA1 );
        writeValue( subNode, "transponderPasswordB1", data.transponderPasswordB1 );
        writeValue( subNode, "transponderPasswordA2", data.transponderPasswordA2 );
        writeValue( subNode, "transponderPasswordB2", data.transponderPasswordB2 );
        writeValue( subNode, "transponderSectorTrailer", Utils::toHexString( data.transponderSectorTrailer, true ) );
        writeValue( subNode, "transponderInitialValue", data.transponderInitialValue, sizeof( data.transponderInitialValue ) / sizeof( data.transponderInitialValue[0] ) );
        writeValue( subNode, "memoryUsage", data.memoryUsage );
        writeValue( subNode, "entranceOnline", data.entranceOnline );
        writeValue( subNode, "entranceOnlineTimeout", data.entranceOnlineTimeout );
        writeValue( subNode, "fileName", data.fileName );
        writeValue( subNode, "fileTimestamp", data.fileTimestamp );
        writeValue( subNode, "passwordSetup", data.passwordSetup, sizeof( data.passwordSetup ) / sizeof( data.passwordSetup[0] ) );
        writeValue( subNode, "userPasswordEnable", data.userPasswordEnable );
        writeValue( subNode, "escJumpToMainmenu", data.escJumpToMainmenu );
        writeValue( subNode, "ledPerFieldFunction", data.ledPerFieldFunction );
        writeValue( subNode, "dataWriteOKTime", data.dataWriteOKTime );
        writeValue( subNode, "dataWriteOKFontType", data.dataWriteOKFontType );
        writeValue( subNode, "powerActiv", data.powerActiv );
        writeValue( subNode, "powerClockTimerCount", data.powerClockTimerCount );
        writeValue( subNode, "spsQuery", data.spsQuery );
        writeValue( subNode, "dataOnCard", data.dataOnCard );
        writeValue( subNode, "acExtension", data.acExtension );
        writeValue( subNode, "acTagTimeToClose", data.acTagTimeToClose );
        writeValue( subNode, "acAlarmTimeout", data.acAlarmTimeout );
        writeValue( subNode, "batteryLoadOptimization", data.batteryLoadOptimization );
        writeValue( subNode, "showSmallWatch", data.showSmallWatch );
        writeValue( subNode, "dontDimBacklightForCharge", data.dontDimBacklightForCharge );
        writeValue( subNode, "flagsEnergie", data.flagsEnergie );
        writeValue( subNode, "displayOffAfter", data.displayOffAfter );
        writeValue( subNode, "flagsShowSetupNote", data.flagsShowSetupNote );
        writeValue( subNode, "phgCryptUseFlags", data.phgCryptUseFlags );
        writeValue( subNode, "readerPollingFlags", data.readerPollingFlags );

        auto subNode2 = m_xmlDoc->NewElement( "recordPriority" );
        subNode->LinkEndChild( subNode2 );
        for ( size_t i = 0; i < sizeof( data.recordPriorityColumn ) / sizeof( data.recordPriorityColumn[0] ); ++i )
        {
            writeValue( subNode2, "recordPriorityColumn", data.recordPriorityColumn[i] );
        }

        writeValue( subNode, "cameraUse", data.cameraUse );
        writeValue( subNode, "cameraGrab", data.cameraGrab );
        writeValue( subNode, "memoryUsageExtData", data.memoryUsageExtData );
        writeValue( subNode, "dateTimeFormatSettings", data.dateTimeFormatSettings );
        writeValue( subNode, "useSignature", data.useSignature );
        writeValue( subNode, "timerTimeBaseDivisor", data.timerTimeBaseDivisor );

        // Update parent node for sub-structure.
        auto oldParent = setParentNode( subNode );

        writeContent( ContentType::Fingerprint, data.fingerprint );
        writeContent( ContentType::TimeboyAdaption, data.adaption );
        writeContent( ContentType::GPRS, data.gprs );
        writeContent( ContentType::EnergieOption, data.energieOption );

        subNode2 = m_xmlDoc->NewElement( "timers" );
        subNode->LinkEndChild( subNode2 );

        // Update parent node for sub-structure.
        auto oldParent2 = setParentNode( subNode2 );
        for ( size_t i = 0; i < sizeof( data.timer ) / sizeof( data.timer[0] ); ++i )
        {
            writeContent( ContentType::Timer, data.timer[i] );
        }
        // Reset parent node because the end of the structure is reached.
        setParentNode( oldParent2 );

        // Backward compatibility of the storage of the digital inputs is not relevant for processing in XML format.
        //subNode2 = m_xmlDoc->NewElement( "digitalInputs" );
        //subNode->LinkEndChild( subNode2 );

        //// Update parent node for sub-structure.
        //oldParent2 = setParentNode( subNode2 );
        //for ( size_t i = 0; i < sizeof( data.digitalExtInput ) / sizeof( data.digitalExtInput[0] ); ++i )
        //{
        //    writeContent( ContentType::DigitalInputEx, data.digitalExtInput[i] );
        //}
        //// Reset parent node because the end of the structure is reached.
        //setParentNode( oldParent2 );

        writeContent( ContentType::GPS, data.gps );
        writeContent( ContentType::ProtectionPassword, data.protectionPassword );

        // Reset parent node because the end of the structure is reached.
        setParentNode( oldParent );
    }
}
template<>
bool FileXML::readContent<StructExtended>( ContentType type, StructExtended &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            std::string dummy;
            readValue( subNode, "doubleInputTime", data.doubleInputTime, errorPresent );
            readValue( subNode, "confirmationTime", data.confirmationTime, errorPresent );
            readValue( subNode, "daylightActive", data.daylightActive, errorPresent );
            readValue( subNode, "daylight1Month", data.daylight1Month, errorPresent );
            readValue( subNode, "daylight1Week", data.daylight1Week, errorPresent );
            readValue( subNode, "daylight1Day", data.daylight1Day, errorPresent );
            readValue( subNode, "daylight1Hour", data.daylight1Hour, errorPresent );
            readValue( subNode, "daylight2Month", data.daylight2Month, errorPresent );
            readValue( subNode, "daylight2Week", data.daylight2Week, errorPresent );
            readValue( subNode, "daylight2Day", data.daylight2Day, errorPresent );
            readValue( subNode, "daylight2Hour", data.daylight2Hour, errorPresent );
            readValue( subNode, "transponderPassword32", data.transponderPassword32, sizeof( data.transponderPassword32 ) / sizeof( data.transponderPassword32[0] ), errorPresent, true );
            readValue( subNode, "transponderPasswordA0", data.transponderPasswordA0, sizeof( data.transponderPasswordA0 ) / sizeof( data.transponderPasswordA0[0] ), errorPresent, true );
            readValue( subNode, "phgAESKey", data.phgAESKey, sizeof( data.phgAESKey ) / sizeof( data.phgAESKey[0] ), errorPresent );
            readValue( subNode, "transponderPasswordFlag", data.transponderPasswordFlag, errorPresent );
            readValue( subNode, "transponderPasswordB0", data.transponderPasswordB0, sizeof( data.transponderPasswordB0 ) / sizeof( data.transponderPasswordB0[0] ), errorPresent, true );
            readValue( subNode, "transponderPasswordA1", data.transponderPasswordA1, sizeof( data.transponderPasswordA1 ) / sizeof( data.transponderPasswordA1[0] ), errorPresent, true );
            readValue( subNode, "transponderPasswordB1", data.transponderPasswordB1, sizeof( data.transponderPasswordB1 ) / sizeof( data.transponderPasswordB1[0] ), errorPresent, true );
            readValue( subNode, "transponderPasswordA2", data.transponderPasswordA2, sizeof( data.transponderPasswordA2 ) / sizeof( data.transponderPasswordA2[0] ), errorPresent, true );
            readValue( subNode, "transponderPasswordB2", data.transponderPasswordB2, sizeof( data.transponderPasswordB2 ) / sizeof( data.transponderPasswordB2[0] ), errorPresent, true );
            data.transponderSectorTrailer = static_cast<int32_t>(Utils::fromHexString_uint( readValue( subNode, "transponderSectorTrailer", dummy, errorPresent ), int32_t() ));
            readValue( subNode, "transponderInitialValue", data.transponderInitialValue, sizeof( data.transponderInitialValue ) / sizeof( data.transponderInitialValue[0] ), errorPresent );
            readValue( subNode, "memoryUsage", data.memoryUsage, errorPresent );
            readValue( subNode, "entranceOnline", data.entranceOnline, errorPresent );
            readValue( subNode, "entranceOnlineTimeout", data.entranceOnlineTimeout, errorPresent );
            readValue( subNode, "fileName", data.fileName, sizeof( data.fileName ) / sizeof( data.fileName[0] ), errorPresent, true );
            readValue( subNode, "fileTimestamp", data.fileTimestamp, sizeof( data.fileTimestamp ) / sizeof( data.fileTimestamp[0] ), errorPresent, true );
            readValue( subNode, "passwordSetup", data.passwordSetup, sizeof( data.passwordSetup ) / sizeof( data.passwordSetup[0] ), errorPresent );
            readValue( subNode, "userPasswordEnable", data.userPasswordEnable, errorPresent );
            readValue( subNode, "escJumpToMainmenu", data.escJumpToMainmenu, errorPresent );
            readValue( subNode, "ledPerFieldFunction", data.ledPerFieldFunction, errorPresent );
            readValue( subNode, "dataWriteOKTime", data.dataWriteOKTime, errorPresent );
            readValue( subNode, "dataWriteOKFontType", data.dataWriteOKFontType, errorPresent );
            readValue( subNode, "powerActiv", data.powerActiv, errorPresent );
            readValue( subNode, "powerClockTimerCount", data.powerClockTimerCount, errorPresent );
            readValue( subNode, "spsQuery", data.spsQuery, errorPresent );
            readValue( subNode, "dataOnCard", data.dataOnCard, errorPresent );
            readValue( subNode, "acExtension", data.acExtension, errorPresent );
            readValue( subNode, "acTagTimeToClose", data.acTagTimeToClose, errorPresent );
            readValue( subNode, "acAlarmTimeout", data.acAlarmTimeout, errorPresent );
            readValue( subNode, "batteryLoadOptimization", data.batteryLoadOptimization, errorPresent );
            readValue( subNode, "showSmallWatch", data.showSmallWatch, errorPresent );
            readValue( subNode, "dontDimBacklightForCharge", data.dontDimBacklightForCharge, errorPresent );
            readValue( subNode, "flagsEnergie", data.flagsEnergie, errorPresent );
            readValue( subNode, "displayOffAfter", data.displayOffAfter, errorPresent );
            readValue( subNode, "flagsShowSetupNote", data.flagsShowSetupNote, errorPresent );
            readValue( subNode, "phgCryptUseFlags", data.phgCryptUseFlags, errorPresent );
            readValue( subNode, "readerPollingFlags", data.readerPollingFlags, errorPresent );

            if ( !errorPresent )
            {
                auto subNode2 = subNode->FirstChildElement( "recordPriority" );
                errorPresent = (subNode2 == nullptr);
                if ( !errorPresent )
                {
                    tinyxml2::XMLElement *subItem{};
                    constexpr int itemCount = sizeof( data.recordPriorityColumn ) / sizeof( data.recordPriorityColumn[0] );
                    for ( size_t i = 0; i < itemCount && !errorPresent; ++i )
                    {
                        readValue( subNode2, subItem, "recordPriorityColumn", data.recordPriorityColumn[i], errorPresent );
                        if ( subItem == nullptr )
                        {
                            // No further element was found.
                            errorPresent = true;
                            break;
                        }
                    }
                }
            }

            readValue( subNode, "cameraUse", data.cameraUse, errorPresent );
            readValue( subNode, "cameraGrab", data.cameraGrab, errorPresent );
            readValue( subNode, "memoryUsageExtData", data.memoryUsageExtData, errorPresent );
            readValue( subNode, "dateTimeFormatSettings", data.dateTimeFormatSettings, errorPresent );
            readValue( subNode, "useSignature", data.useSignature, errorPresent );
            if ( m_minStudioVersion >= "04.03.15.00" )
            {
                readValue( subNode, "timerTimeBaseDivisor", data.timerTimeBaseDivisor, errorPresent );
            }
            else
            {
                data.timerTimeBaseDivisor = 1;
            }

            if ( !errorPresent )
            {
                // Update parent node for sub-structure.
                auto oldParent = setParentNode( subNode );

                errorPresent |= !readContent( FileXML::ContentType::Fingerprint, data.fingerprint );
                errorPresent |= !readContent( FileXML::ContentType::TimeboyAdaption, data.adaption );
                errorPresent |= !readContent( FileXML::ContentType::GPRS, data.gprs );
                errorPresent |= !readContent( FileXML::ContentType::EnergieOption, data.energieOption );

                if ( !errorPresent )
                {
                    auto subNode2 = subNode->FirstChildElement( "timers" );

                    // Update parent node for sub-structure.
                    auto oldParent2 = setParentNode( subNode2 );
                    for ( size_t i = 0; i < sizeof( data.timer ) / sizeof( data.timer[0] ) && !errorPresent; ++i )
                    {
                        tinyxml2::XMLElement *subItem{};
                        errorPresent |= !readContent( ContentType::Timer, data.timer[i], subItem );
                        if ( subItem == nullptr )
                        {
                            // No further element was found.
                            errorPresent = true;
                            break;
                        }
                    }
                    // Reset parent node because the end of the structure is reached.
                    setParentNode( oldParent2 );
                }

                if ( !errorPresent )
                {
                    errorPresent |= !readContent( ContentType::GPS, data.gps );
                    errorPresent |= !readContent( ContentType::ProtectionPassword, data.protectionPassword );
                }

                // Reset parent node because the end of the structure is reached.
                setParentNode( oldParent );
            }
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructExtendedFlash>( ContentType type, const StructExtendedFlash &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "check", Utils::toHexString( data.check, true ) );
        writeValue( subNode, "size", data.size );
        writeValue( subNode, "flags", Utils::toHexString( data.flags, true ) );
    }
}
template<>
bool FileXML::readContent<StructExtendedFlash>( ContentType type, StructExtendedFlash &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            readValue( subNode, "check", data.check, errorPresent );
            readValue( subNode, "size", data.size, errorPresent );
            readValue( subNode, "flags", data.flags, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}

template<>
void FileXML::writeContent<StructSystemMessage>( ContentType type, const StructSystemMessage &data )
{
    const std::string &nodeName = nodeNameByType( type );
    assert( (m_parent != nullptr) && !nodeName.empty() );
    if ( (m_parent != nullptr) && !nodeName.empty() )
    {
        auto subNode = m_xmlDoc->NewElement( nodeName.c_str() );
        m_parent->LinkEndChild( subNode );

        writeValue( subNode, "type", static_cast<uint8_t>(data.type) );
        writeValue( subNode, "id", data.id.param1 );
        writeValue( subNode, "flags", data.flags );
        writeValue( subNode, "invokeInputChain", data.invokeInputChain );
    }
}
template<>
bool FileXML::readContent<StructSystemMessage>( ContentType type, StructSystemMessage &data, tinyxml2::XMLElement *&lastItem )
{
    const std::string &nodeName = nodeNameByType( type );
    bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
    assert( !errorPresent );
    if ( !errorPresent )
    {
        auto subNode = (lastItem == nullptr) ? m_parent->FirstChildElement( nodeName.c_str() ) : lastItem->NextSiblingElement( nodeName.c_str() );
        errorPresent = (subNode == nullptr);
        if ( !errorPresent )
        {
            uint8_t tmp{};
            data.type = static_cast<EnumSystemMessageType>(readValue( subNode, "type", tmp, errorPresent ));
            readValue( subNode, "id", data.id.param1, errorPresent );
            readValue( subNode, "flags", data.flags, errorPresent );
            readValue( subNode, "invokeInputChain", data.invokeInputChain, errorPresent );
        }
        lastItem = subNode;
    }
    return !errorPresent;
}


std::string FileXML::tagToStruct( const std::string &tag )
{
    return "Struct" + boost::to_upper_copy( tag.substr( 0, 1 ) ) + tag.substr( 1 );
}

const tinyxml2::XMLNode *FileXML::findStructCompound( const std::string &name ) const
{
    // Cancel processing if the doxygen tag file was not loaded correctly or does not exist.
    if ( m_pDoxygenRoot == nullptr )
    {
        return nullptr;
    }

    // find correct compound definition
    const tinyxml2::XMLElement *element = m_pDoxygenRoot->FirstChildElement( "compound" );
    do
    {
        if ( element != nullptr )
        {
            // check kind - struct
            auto attribute = element->Attribute( "kind" );
            if ( (attribute != nullptr) && (std::string( attribute ).compare( "struct" ) == 0) )
            {
                auto nameElement = element->FirstChildElement( "name" );
                if ( (nameElement != nullptr) && (name.compare( nameElement->GetText() ) == 0) )
                {
                    return element;
                }
            }
            element = element->NextSiblingElement( "compound" );
        }
    }
    while ( element != nullptr );

    return nullptr;
}


void FileXML::writeFileNames( const std::array<std::string, 20> &listFileNames, const std::array<std::string, 20> &listEntrance2FileNames )
{
    auto root = m_xmlDoc->RootElement();
    auto node = root->FirstChildElement( "fileLists" );
    if ( node != nullptr )
    {
        node->DeleteChildren();
    }
    else
    {
        node = m_xmlDoc->NewElement( "fileLists" );
        root->LinkEndChild( node );
    }

    auto subNode = m_xmlDoc->NewElement( "listFiles" );
    node->LinkEndChild( subNode );

    for ( const auto fileName : listFileNames )
    {
        writeValue<std::string>( subNode, "listFile", fileName );
    }

    subNode = m_xmlDoc->NewElement( "entranceFiles" );
    node->LinkEndChild( subNode );
    for ( const auto fileName : listEntrance2FileNames )
    {
        writeValue<std::string>( subNode, "entranceFile", fileName );
    }
}
bool FileXML::readFileNames( std::array<std::string, 20> &listFileNames, std::array<std::string, 20> &listEntrance2FileNames )
{
    auto root = m_xmlDoc->RootElement();
    auto parent = (root != nullptr) ? root->FirstChildElement( "fileLists" ) : nullptr;
    auto node = (parent != nullptr) ? parent->FirstChildElement( "listFiles" ) : nullptr;

    bool errorPresent = ((parent == nullptr) || (node == nullptr));
    if ( !errorPresent )
    {
        tinyxml2::XMLElement *lastItem{};
        for ( size_t i = 0; i < listFileNames.size(); ++i )
        {
            readValue( node, lastItem, "listFile", listFileNames[i], errorPresent );
            if ( errorPresent )
            {
                break;
            }
        }
    }

    if ( !errorPresent )
    {
        node = (parent != nullptr) ? parent->FirstChildElement( "entranceFiles" ) : nullptr;

        tinyxml2::XMLElement *lastItem{};
        for ( size_t i = 0; i < listEntrance2FileNames.size(); ++i )
        {
            readValue( node, lastItem, "entranceFile", listEntrance2FileNames[i], errorPresent );
            if ( errorPresent )
            {
                break;
            }
        }
    }

    return !errorPresent;
}
